SystemOrganization addCategory: #'System-Support'!

!Behavior methodsFor: '*System-Support' stamp: 'Alexandre Bergel 11/21/2009 11:49'!
allCallsOn
	"Answer a SortedCollection of all the methods that refer to me by name or as part of an association in a global dict."


	^ (self  systemNavigation allReferencesTo:  (self environment associationAt: self theNonMetaClass name)), (self  systemNavigation allCallsOn:  self theNonMetaClass name)	! !

!Behavior methodsFor: '*System-Support' stamp: 'dvf 8/23/2003 12:43'!
allCallsOn: aSymbol
	"Answer a SortedCollection of all the methods that call on aSymbol."


	^ self  systemNavigation allCallsOn: aSymbol from: self .
	! !

!Behavior methodsFor: '*System-Support' stamp: 'stephane.ducasse 10/12/2008 21:02'!
allUnsentMessages
	"Answer an array of all the messages defined by the receiver that are not sent anywhere in the system."

	^ SystemNavigation default allUnsentMessagesIn: self selectors! !

IdentityDictionary subclass: #SystemDictionary
	instanceVariableNames: 'cachedClassNames'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SystemDictionary commentStamp: '<historical>' prior: 0!
I represent a special dictionary that supports protocol for asking questions about the structure of the system. Other than class names, I contain (print this)...
	Smalltalk keys select: [:k | ((Smalltalk at: k) isKindOf: Class) not]
			thenCollect: [:k | k -> (Smalltalk at: k) class]
!

!SystemDictionary class methodsFor: 'instance creation' stamp: 'StephaneDucasse 10/3/2010 15:02'!
withOrganizer: anOrganizer
	| dictionary |
	dictionary := self new.
	dictionary organization: anOrganizer. "update too the back pointer"
	^dictionary.
	! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'sd 4/17/2003 21:15'!
allClasses  
	"Return all the class defines in the Smalltalk SystemDictionary"
	"Smalltalk allClasses"

	^ self classNames collect: [:name | self at: name]! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'al 1/12/2006 23:59'!
allClassesAndTraits
	"Return all the classes and traits defined in the Smalltalk SystemDictionary"

	^ self classNames , self traitNames collect: [:each | self at: each]! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'al 1/13/2006 00:15'!
allClassesAndTraitsDo: aBlock
	^self allClassesAndTraits do: aBlock! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'sd 4/17/2003 21:18'!
allClassesDo: aBlock
	"Evaluate the argument, aBlock, for each class in the system."

	(self classNames collect: [:name | self at: name]) do: aBlock! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'al 2/23/2006 21:39'!
allTraits
	"Return all traits defined in the Smalltalk SystemDictionary"

	^ self traitNames collect: [:each | self at: each]! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'StephaneDucasse 3/13/2010 15:36'!
allTraitsDo: aBlock
	"Evaluate the argument, aBlock, for each trait in the system."

	(self traitNames collect: [:name | self at: name]) do: aBlock! !

!SystemDictionary methodsFor: 'dictionary access' stamp: 'md 3/2/2006 22:01'!
associationOrUndeclaredAt: key 
	"return an association or install in undeclared.  Used for mating up ImageSegments."

	^ self associationAt: key ifAbsent: [
		Undeclared at: key put: nil.
		Undeclared associationAt: key]! !

!SystemDictionary methodsFor: 'dictionary access' stamp: ''!
at: aKey put: anObject 
	"Override from Dictionary to check Undeclared and fix up
	references to undeclared variables."
	| index element |
	(self includesKey: aKey) ifFalse: 
		[self declare: aKey from: Undeclared.
		self flushClassNameCache].
	super at: aKey put: anObject.
	^ anObject! !

!SystemDictionary methodsFor: 'accessing' stamp: 'StephaneDucasse 3/13/2010 17:27'!
bindingOf: varName
	"SystemDictionaries includes symbols only"
	
	^ super bindingOf: varName asSymbol! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'al 1/12/2006 23:57'!
classNamed: className 
	^self classOrTraitNamed: className.! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'StephaneDucasse 4/27/2010 11:09'!
classNames
	"Answer a SortedCollection of all class names."
	| names |
	cachedClassNames == nil ifTrue:
		[names := OrderedCollection new: self size.
		self keysAndValuesDo: [:key :value|
			"The key == value name test below addresses two separate issues:
				1) Obsolete classes, where key = #Foo and value name = 'AnObsoleteFoo'
				2) Aliases, i.e., Smalltalk at: #OtherName put: aClass"
				((value isKindOf: Class) and: [key == value name])
					ifTrue: [names add: key]].
	cachedClassNames := names asSortedCollection].
	^ cachedClassNames! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'lr 3/14/2010 21:13'!
classOrTraitNamed: aString
	"aString is either a class or trait name or a class or trait name followed by ' class' or 'classTrait' respectively.
	Answer the class or metaclass it names."

	| meta baseName baseClass |
	(aString endsWith: ' class')
		ifTrue: [ 
			meta := true.
			baseName := aString copyFrom: 1 to: aString size - 6 ]
		ifFalse: [ 
			(aString endsWith: ' classTrait')
				ifTrue: [ 
					meta := true.
					baseName := aString copyFrom: 1 to: aString size - 11 ]
				ifFalse: [ 
					meta := false.
					baseName := aString ] ].
	baseClass := Smalltalk globals at: baseName asSymbol ifAbsent: [ ^ nil ].
	meta
		ifTrue: [ ^ baseClass classSide ]
		ifFalse: [ ^ baseClass ]! !

!SystemDictionary methodsFor: 'accessing' stamp: 'StephaneDucasse 3/13/2010 15:38'!
environment
	"for conversion from SmalltalkImage to SystemDictionary"
	
	^ self ! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'StephaneDucasse 7/6/2010 21:41'!
flushClassNameCache
	"Smalltalk flushClassNameCache"
	"Force recomputation of the cached list of class names."

	cachedClassNames := nil! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'StephaneDucasse 3/13/2010 16:32'!
forgetClass: aClass logged: aBool 
	"Delete the class, aClass, from the system.
	Note that this doesn't do everything required to dispose of a class - to do that use Class>>removeFromSystem."

	aBool ifTrue: [SystemChangeNotifier uniqueInstance classRemoved: aClass fromCategory: aClass category].		
	self organization removeElement: aClass name.
	Smalltalk removeFromStartUpList: aClass.
	Smalltalk removeFromShutDownList: aClass.
	self removeKey: aClass name ifAbsent: [].
	self flushClassNameCache! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'StephaneDucasse 8/19/2010 22:32'!
hasClassNamed: aString
	"Answer whether there is a class of the given name, but don't intern aString if it's not alrady interned."

	Symbol hasInterned: aString ifTrue: 
		[:aSymbol | ^ (self at: aSymbol ifAbsent: [nil]) isKindOf: Class].
	^ false! !

!SystemDictionary methodsFor: 'deprecated' stamp: 'StephaneDucasse 3/20/2010 22:58'!
hasSpecialSelector: aLiteral ifTrueSetByte: aBlock

	self deprecated: 'Use Smalltalk instead' on: '20 March 2010' in: 'Pharo1.1'.
	^ Smalltalk hasSpecialSelector: aLiteral ifTrueSetByte: aBlock! !

!SystemDictionary methodsFor: 'objects from disk' stamp: 'tk 9/28/2000 15:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to Smalltalk instead."

	dp := DiskProxy global: #Smalltalk selector: #yourself
			args: #().
	refStrm replace: self with: dp.
	^ dp! !

!SystemDictionary methodsFor: 'accessing' stamp: 'StephaneDucasse 10/3/2010 14:57'!
organization
	"Return the organizer for the receiver"
	
	^ self at: #SystemOrganization! !

!SystemDictionary methodsFor: 'accessing' stamp: 'StephaneDucasse 10/3/2010 14:59'!
organization: anOrganization
	"Return the organizer for the receiver"
	
	anOrganization environment: self.
	^self at: #SystemOrganization put: anOrganization! !

!SystemDictionary methodsFor: 'accessing' stamp: 'nice 1/5/2010 15:59'!
poolUsers
	"Answer a dictionary of pool name -> classes that refer to it.
	Also includes any globally know dictionaries (such as
	Smalltalk, Undeclared etc) which although not strictly
	accurate is potentially useful information"
	"Smalltalk poolUsers"
	| poolUsers |
	poolUsers := Dictionary new.
	self keys
		do: [:k | | pool refs | "yes, using isKindOf: is tacky but for reflective code like
			this it is very useful. If you really object you can:-
			a) go boil your head.
			b) provide a better answer.
			your choice."
			(((pool := self at: k) isKindOf: Dictionary)
					or: [pool isKindOf: SharedPool class])
				ifTrue: [refs := self systemNavigation allClasses
								select: [:c | c sharedPools identityIncludes: pool]
								thenCollect: [:c | c name].
					refs
						add: (self systemNavigation
								allReferencesTo: (self associationAt: k)).
					poolUsers at: k put: refs]].
	^ poolUsers! !

!SystemDictionary methodsFor: 'printing' stamp: 'sma 6/1/2000 09:53'!
printElementsOn: aStream
	aStream nextPutAll:'(lots of globals)'! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'sw 9/5/97 18:30'!
removeClassNamed: aName
	"Invoked from fileouts:  if there is currently a class in the system named aName, then remove it.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: aName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Removal of class named ', aName, ' ignored because ', aName, ' does not exist.'.
			^ self].

	oldClass removeFromSystem! !

!SystemDictionary methodsFor: 'dictionary access' stamp: 'StephaneDucasse 7/6/2010 21:43'!
removeKey: key ifAbsent: aBlock
	"Remove key (and its associated value) from the receiver. If key is not in
	the receiver, answer the result of evaluating aBlock. Otherwise, answer
	the value externally named by key."
	
	self flushClassNameCache.
	^super removeKey: key ifAbsent: aBlock! !

!SystemDictionary methodsFor: 'renaming' stamp: 'StephaneDucasse 3/13/2010 16:38'!
renameClass: aClass as: newName 
	"Rename the class, aClass, to have the title newName."

	^self renameClass: aClass from: aClass name to: newName! !

!SystemDictionary methodsFor: 'renaming' stamp: 'StephaneDucasse 3/13/2010 16:38'!
renameClass: aClass from: oldName 
	"Rename the class, aClass, to have the title newName."

	^self renameClass: aClass from: oldName to: aClass name! !

!SystemDictionary methodsFor: 'renaming' stamp: 'StephaneDucasse 10/6/2010 19:55'!
renameClass: aClass from: oldName to: newName
	"Rename the class, aClass, to have the title newName."

	| oldref category |
	category := self organization categoryOfElement: oldName.
	self organization classify: newName under: category.
	self organization removeElement: oldName.
	oldref := self associationAt: oldName.
	self removeKey: oldName.
	oldref key: newName.
	self add: oldref.  "Old association preserves old refs"
	Smalltalk renamedClass: aClass from: oldName to: newName.
	self flushClassNameCache.
	SystemChangeNotifier uniqueInstance classRenamed: aClass from: oldName to: newName inCategory: category! !

!SystemDictionary methodsFor: 'renaming' stamp: 'sw 10/28/96'!
renameClassNamed: oldName as: newName
	"Invoked from fileouts:  if there is currently a class in the system named oldName, then rename it to newName.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: oldName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Class-rename for ', oldName, ' ignored because ', oldName, ' does not exist.'.
			^ self].

	oldClass rename: newName! !

!SystemDictionary methodsFor: 'objects from disk' stamp: 'tk 3/7/2000 18:40'!
storeDataOn: aDataStream
	"I don't get stored.  Use a DiskProxy"

	self error: 'use a DiskProxy to store me'! !

!SystemDictionary methodsFor: 'class and trait names' stamp: 'MarcusDenker 3/21/2010 20:11'!
traitNames
	"Answer a SortedCollection of all traits (not including class-traits) names."
	| names |
	names := OrderedCollection new.
	self do: 
		[:cl | ((cl isKindOf: Trait)
			and: [(cl name beginsWith: 'AnObsolete') not])
				ifTrue: [names add: cl name]].
	^ names! !

!SystemDictionary methodsFor: 'copying' stamp: 'tk 10/20/2000 11:35'!
veryDeepCopyWith: deepCopier
	"Return self.  I can't be copied.  Do not record me."! !

Categorizer subclass: #SystemOrganizer
	instanceVariableNames: 'environment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SystemOrganizer commentStamp: '<historical>' prior: 0!
My instances provide an organization for the classes in the system, just as a ClassOrganizer organizes the messages within a class. The only difference is the methods for fileIn/Out.!

!SystemOrganizer class methodsFor: 'cleanup' stamp: 'StephaneDucasse 3/9/2010 22:23'!
cleanUp: agressive
	"Remove empty categories when cleaning aggressively"

	agressive ifTrue: [SystemOrganization removeEmptyCategories].
! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'ab 2/10/2005 16:33'!
addCategory: newCategory
	| r |
	r := super addCategory: newCategory.
	SystemChangeNotifier uniqueInstance classCategoryAdded: newCategory.
	^ r! !

!SystemOrganizer methodsFor: 'remove' stamp: 'di 3/3/2001 16:07'!
categoriesMatching: matchString
	"Return all matching categories"
	^ self categories select: [:c | matchString match: c]! !

!SystemOrganizer methodsFor: 'utils' stamp: 'sd 9/12/2010 19:09'!
classesInCategory: category
	^ (self listAtCategoryNamed: category) collect: [ :className | self environment at: className ] ! !

!SystemOrganizer methodsFor: 'utils' stamp: 'sd 9/12/2010 19:10'!
commentInventory: categoryName
	"SystemOrganization commentInventory: 'Morphic*'"

	| classes commentedClasses |
	classes := OrderedCollection new.
	self categories
		withIndexCollect: [ :cat :idx | 
			(categoryName match: cat)
				ifTrue: [ classes addAll: (self listAtCategoryNumber: idx) ]
				ifFalse: [ nil ] ].
	commentedClasses := classes select: [ :catCls | (self environment at: catCls) hasComment ].
	^ 'There are ' , classes size asString , ' classes in ' , categoryName , ' of which ' , commentedClasses size asString
		, ' have comments and ' , (classes size - commentedClasses size) asString , ' do not yet have comments.'! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'StephaneDucasse 10/3/2010 14:45'!
environment
	 ^ environment ifNil: [ environment := Smalltalk globals]! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'NicoPaez 10/1/2010 15:55'!
environment: aSystemDictionary
	 environment := aSystemDictionary.
! !

!SystemOrganizer methodsFor: 'filein/out' stamp: 'PeterHugossonMiller 9/3/2009 11:32'!
fileOut  
	"SystemOrganization fileOut"

	| internalStream |
	internalStream := (String new: 30000) writeStream.
	internalStream nextPutAll: 'SystemOrganization changeFromCategorySpecs: #('; cr;
		print: SystemOrganization;  "ends with a cr"
		nextPutAll: ')!!'; cr.

	FileStream writeSourceCodeFrom: internalStream baseName: (FileDirectory default nextNameFor: 'SystemOrganization' extension: 'st') isSt: true
! !

!SystemOrganizer methodsFor: 'filein/out' stamp: 'PeterHugossonMiller 9/3/2009 11:32'!
fileOutCategory: category 
	"Store on the file named category (a string) concatenated with '.st' all the 
	classes associated with the category."
	
	| internalStream |
	internalStream := (String new: 1000) writeStream.
	self fileOutCategory: category on: internalStream initializing: true.
	^ FileStream writeSourceCodeFrom: internalStream baseName: category isSt: true ! !

!SystemOrganizer methodsFor: 'filein/out' stamp: 'ar 12/22/1999 17:28'!
fileOutCategory: category on: aFileStream 
	"Store on the file associated with aFileStream, all the classes associated 
	with the category and any requested shared pools."
	^self fileOutCategory: category on: aFileStream initializing: true! !

!SystemOrganizer methodsFor: 'filein/out' stamp: 'StephaneDucasse 3/13/2010 17:13'!
fileOutCategory: category on: aFileStream initializing: aBool
	"Store on the file associated with aFileStream, all the traits and classes associated 
	with the category and any requested shared pools in the right order."

	| first poolSet tempClass classes traits |
	traits := self orderedTraitsIn: category.
	classes := self superclassOrder: category.
	poolSet := Set new.
	classes do:  [:class | class sharedPools do: [:eachPool | poolSet add: eachPool]].
	poolSet size > 0 ifTrue: [
		tempClass := Class new.
		tempClass shouldFileOutPools ifTrue: [
			poolSet := poolSet select: [:aPool |
				tempClass shouldFileOutPool: (Smalltalk globals keyAtIdentityValue: aPool)].
			poolSet do: [:aPool | tempClass fileOutPool: aPool onFileStream: aFileStream]]].
	first := true.
	traits, classes do: [:each | 
		first
			ifTrue: [first := false]
			ifFalse: [aFileStream cr; nextPut: Character newPage; cr].
		each
			fileOutOn: aFileStream
			moveSource: false
			toFile: 0
			initializing: false].
	aBool ifTrue: [classes do: [:cls | cls fileOutInitializerOn: aFileStream]].! !

!SystemOrganizer methodsFor: 'private' stamp: 'rw 7/31/2003 17:23'!
ifClassOrganizerDo: aBlock
	"Do nothing, since this is not a class organizer"! !

!SystemOrganizer methodsFor: 'filein/out' stamp: 'tk 9/28/2000 15:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a path to me in the other system instead."

self == SystemOrganization ifTrue: [
	dp := DiskProxy global: #SystemOrganization selector: #yourself args: #().
	refStrm replace: self with: dp.
	^ dp].
^ self
! !

!SystemOrganizer methodsFor: 'utils' stamp: 'sd 9/12/2010 19:13'!
orderedTraitsIn: category
	"Answer an OrderedCollection containing references to the traits in the 
	category whose name is the argument, category (a string). The traits 
	are ordered so they can be filed in." 

	| behaviors traits |
	behaviors := (self listAtCategoryNamed: category asSymbol) collect: [ :title | self environment at: title ].
	traits := behaviors reject: [ :each | each isBehavior ].
	traits := traits
		asSortedCollection: [ :t1 :t2 | (t2 traitComposition allTraits includes: t1) or: [ (t1 traitComposition allTraits includes: t2) not ] ].
	^ traits asArray! !

!SystemOrganizer methodsFor: 'remove' stamp: 'di 3/3/2001 16:08'!
removeCategoriesMatching: matchString
	"Remove all matching categories with their classes"
	(self categoriesMatching: matchString) do:
		[:c | self removeSystemCategory: c]! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'ab 2/10/2005 16:36'!
removeCategory: cat 
	| r |
	r := super removeCategory: cat.
	SystemChangeNotifier uniqueInstance classCategoryRemoved: cat.
	^ r! !

!SystemOrganizer methodsFor: 'utils' stamp: 'sd 9/12/2010 19:13'!
removeMissingClasses
	"Remove any class names that are no longer in the Smalltalk dictionary. Used for cleaning up after garbage collecting user-generated classes."

	"SystemOrganization removeMissingClasses"
 
	elementArray copy
		do: [ :el | 
			(self environment includesKey: el)
				ifFalse: [ self removeElement: el ] ]! !

!SystemOrganizer methodsFor: 'remove' stamp: 'al 4/25/2004 11:38'!
removeSystemCategory: category
	"remove all the classes and traits associated with the category"

	 (self orderedTraitsIn: category) , (self superclassOrder: category)
		reverseDo: [:each | each removeFromSystem].

	self removeCategory: category.
! !

!SystemOrganizer methodsFor: 'accessing' stamp: 'ab 2/10/2005 16:47'!
renameCategory: oldCatString toBe: newCatString
	| r |
	r := super renameCategory: oldCatString toBe: newCatString.
	SystemChangeNotifier uniqueInstance 
		classCategoryRenamedFrom: oldCatString to: newCatString.
	^ r! !

!SystemOrganizer methodsFor: 'utils' stamp: 'sd 9/12/2010 19:14'!
superclassOrder: category
	"Answer an OrderedCollection containing references to the classes in the 
	category whose name is the argument, category (a string). The classes 
	are ordered with superclasses first so they can be filed in."

	| behaviors classes |
	behaviors := (self listAtCategoryNamed: category asSymbol) collect: [ :title | self environment at: title ].
	classes := behaviors select: [ :each | each isBehavior ].
	^ ChangeSet superclassOrder: classes ! !

!SystemOrganizer methodsFor: 'query' stamp: 'sd 9/12/2010 19:15'!
uncommentedClassesIn: categoryName
	"SystemOrganization uncommentedClassesIn: 'Morphic*'"

	| classes |
	classes := OrderedCollection new.
	self categories
		withIndexCollect: [ :cat :idx | 
			(categoryName match: cat)
				ifTrue: [ classes addAll: (self listAtCategoryNumber: idx) ]
				ifFalse: [ nil ] ].
	^ (classes collect: [ :clsName | self environment at: clsName ] thenSelect: [ :cls | cls hasComment not ]) asArray! !

Model subclass: #Project
	instanceVariableNames: 'world changeSet transcript parentProject displayDepth projectParameters isolatedHead'
	classVariableNames: 'AllProjects CurrentProject GoalFreePercent GoalNotMoreThan UIProcess'
	poolDictionaries: ''
	category: 'System-Support'!

!Project commentStamp: 'tk 12/2/2004 12:38' prior: 0!
A Project stores the state of a complete Squeak desktop, including
the windows, and the currently active changeSet.  A project knows who
its parent project is.  When you change projects, whether by entering
or exiting, the screen state of the project being exited is saved in
that project.

A project is retained by its view in the parent world.  It is
effectively named by the name of its changeSet, which can be changed
either by renaming in a changeSorter, or by editing the label of its
view from the parent project.

As the site of major context switch, Projects are the locus of
swapping between the old MVC and the new Morphic worlds.  The
distinction is based on whether the variable 'world' contains a
WorldMorph or a ControlManager.

Saving and Loading
Projects may be stored on the disk in external format.  (Project
named: 'xxx') exportSegment, or choose 'store project on file...'.
Projects may be loaded from a server and stored back.  Storing on a
server never overwrites;  it always makes a new version.  A project
remembers the url of where it lives in urlList.  The list is length
one, for now.  The url may point to a local disk instead of a server.
All projects that the user looks at, are cached in the Squeaklet
folder.  Sorted by server.  The cache holds the most recent version
only.

When a project is loaded into Squeak, its objects are converted to
the current version.  There are three levels of conversion.  First,
each object is converted from raw bits to an object in its old
format.  Then it is sent some or all of these messages:
	comeFullyUpOnReload: smartRefStream  		Used to
re-discover an object that already exists in this image, such as a
resource, global variable, Character, or Symbol.  (sent to objects in
outPointers)
	convertToCurrentVersion: varDict refStream: smartRefStrm
		fill in fields that have been added to a class since
the object was stored.  Used to set the extra inst var to a default
value.  Or, return a new object of a different class.  (sent to
objects that changed instance variables)
	fixUponLoad: aProject refStream: smartRefStrm
	change the object due to conventions that have changed on the
project level.  (sent to all objects in the incoming project)

Here is the calling sequence for storing out a Project:
Project saveAs
Project storeOnServer
Project storeOnServerWithProgressInfo
Project storeOnServerInnards
Project exportSegmentFileName:directory:
Project exportSegmentWithChangeSet:fileName:directory:
ImageSegment writeForExportWithSources:inDirectory:changeSet:
---------
Isolation (not used any more)
When you accept a method, the entire system feels the change, except
projects that are "isolated".  In an isolated project, all new global
variables (including new classes) arestored in the project-local
environment, and all changes to preexisting classes are revoked when
you leave the project.  When you enter another project, that
project's changes are invoked.  Invocation and revocation are handled
efficiently by swapping pointers.  To make a project be isolated,
choose 'isolate changes of this project' from the 'changes...'
section of the screen menu.  You can use an isolated project for
making dangerous change to a system, and you can get out if it
crashes.  A foreign application can have the separate environment it
wants.  Also, you can freeze part of the system for a demo that you
don't want to disturb.  An isolated project shares methods with all
subprojects inside it, unless they are isolated themselves.   Each
isolated project is the head of a tree of projects with which it
shares all methods.

You may 'assert' all changes ever made in the current project to take
effect above this project.  This amounts to exporting all the globals
in the current environment, and zapping the revocation lists to that
the current state of the world will remain in force upon exit from
this project.

[Later: A project may be 'frozen'.  Asserts do not apply to it after
that.  (Great for demos.)  You should be informed when an assert was
blocked in a frozen project.]

Class definitions are layered by the isolation mechanism.  You are
only allowed to change the shape of a class in projects that lie
within its isolation scope.  All versions of the methods are
recompiled, in all projects.  If you remove an inst var that is in
use in an isolated project, it will become an Undeclared global.  It
is best not to remove an inst var when it is being used in another
isolated project. [If we recompile them all, why can't we diagnose
the problem before allowing the change??]

Senders and Implementors do not see versions of a method in isolated
projects.  [again, we might want to make this possible at a cost].
When you ask for versions of a method, you will not get the history
in other isolated projects.

Moving methods and classes between changeSets, and merging changeSets
has no effect on which methods are in force.  But, when you look at a
changeSet from a different isolated project, the methods will contain
code that is not in force.  A changeSet is just a list of method
names, and does not keep separate copies of any code.

When finer grained assertion is needed, use the method (aProject
assertClass: aClass from: thisProject warn: warnConflicts).

How isolated changes work: The first time a class changes, store its
MethodDictionary object.  Keep parallel arrays of associations to
Classes and MethodDictionaries.  Traverse these and install them when
you enter an "ioslated project".  When you leave, store this
project's own MethodDictionaries there.
	To do an assert, we must discover which methods changed here,
and which changed only in the project we are asserting into.  There
is one copy of the 'virgin' method dictionaries in the system.  It is
always being temporarily stored by the currently inForce isolated
project.

isolatedHead - true for the top project, and for each isolated
project.  false or nil for any subproject that shares all methods
with its parent project.

inForce -  true if my methods are installed now.  false if I am
dormant. [is this equivalent to self == Project Current?]

classArray - list of associations to classes

methodDictArray - the method dictionaries of those classes before we
started changing methods.  They hang onto the original
compiledMethods.  (If this project is dormant, it contains the method
dictionaries of those classes as they will be here, in this project).

orgArray - the class organizations of the classes in classArray.

UsingIsolation (class variable) - No longer used.

When you want to save a project in export format from within that
very project, it gets tricky.  We set two flags in parentProject,
exit to it, and let parentProject write the project.
ProjectViewMorph in parentProject checks in its step method, does the
store, clears the flags, and reenters the subProject.

!

!Project class methodsFor: 'constants' stamp: ''!
current
	"Answer the project that is currently being used."

	^CurrentProject! !

!Project class methodsFor: 'utilities' stamp: 'nice 4/4/2010 22:27'!
interruptName: labelString preemptedProcess: theInterruptedProcess
	"Create a Notifier on the active scheduling process with the given label."
	| preemptedProcess projectProcess |
	ActiveHand ifNotNil:[ActiveHand interrupted].
	ActiveWorld := World. "reinstall active globals"
	ActiveHand := World primaryHand.
	ActiveHand interrupted. "make sure this one's interrupted too"
	ActiveEvent := nil.

	projectProcess := self uiProcess.	"we still need the accessor for a while"
	preemptedProcess := theInterruptedProcess ifNil: [Processor preemptedProcess].
	"Only debug preempted process if its priority is >= projectProcess' priority"
	preemptedProcess priority < projectProcess priority 
		ifTrue:[preemptedProcess := projectProcess].
	preemptedProcess suspend.
	ToolSet interrupt: preemptedProcess label: labelString! !

!Project class methodsFor: 'initialization' stamp: 'tak 8/3/2005 18:35'!
localeChanged
	self current localeChanged! !

!Project class methodsFor: 'utilities' stamp: 'RAA 6/3/2000 18:28'!
resumeProcess: aProcess
	"Adopt aProcess as the project process -- probably because of proceeding from a debugger"

	UIProcess := aProcess.
	UIProcess resume! !

!Project class methodsFor: 'utilities' stamp: 'RAA 6/3/2000 18:49'!
spawnNewProcess

	UIProcess := [
		[World doOneCycle.  Processor yield.  false] whileFalse: [].
	] newProcess priority: Processor userSchedulingPriority.
	UIProcess resume! !

!Project class methodsFor: 'utilities' stamp: 'RAA 6/3/2000 18:49'!
spawnNewProcessAndTerminateOld: terminate

	self spawnNewProcess.
	terminate
		ifTrue: [Processor terminateActive]
		ifFalse: [Processor activeProcess suspend]! !

!Project class methodsFor: 'instance creation' stamp: 'RAA 9/27/2000 13:41'!
uiProcess

	^ UIProcess! !

!Project methodsFor: 'accessing' stamp: 'tk 4/5/2000 16:29'!
changeSet
	^ changeSet! !

!Project methodsFor: 'initialization' stamp: 'AlainPlantec 1/7/2010 22:34'!
initialize
	super initialize.
	changeSet := ChangeSet new.
	transcript := TranscriptStream new.
	displayDepth := Display depth.
	parentProject := CurrentProject.
	isolatedHead := false.
! !

!Project methodsFor: 'project parameters' stamp: 'sw 4/24/2001 11:58'!
initializeProjectParameters
	"Initialize the project parameters."

	projectParameters := IdentityDictionary new.
	^ projectParameters! !

!Project methodsFor: 'accessing' stamp: 'RAA 6/3/2000 10:23'!
isCurrentProject

	^self == CurrentProject! !

!Project methodsFor: 'accessing' stamp: 'tk 6/26/1998 22:17'!
labelString
	^ changeSet name! !

!Project methodsFor: 'language' stamp: 'tak 8/4/2005 16:30'!
localeChanged
	"Set the project's natural language as indicated"
	self projectParameterAt: #localeID put: LocaleID current.
	self updateLocaleDependents! !

!Project methodsFor: 'language' stamp: 'StephaneDucasse 11/6/2009 21:06'!
localeID
	"Answer the natural language for the project"

	^ LocaleID current! !

!Project methodsFor: 'accessing' stamp: 'di 7/19/1999 15:06'!
name
	changeSet == nil ifTrue: [^ 'no name'].
	^ changeSet name! !

!Project methodsFor: 'project parameters' stamp: 'sw 2/16/2001 22:35'!
noteThatParameter: prefSymbol justChangedTo: aBoolean
	"Provides a hook so that a user's toggling of a project parameter might precipitate some immediate action"

! !

!Project methodsFor: 'project parameters' stamp: 'ar 5/25/2000 23:23'!
parameterAt: aSymbol
	^self parameterAt: aSymbol ifAbsent:[nil]! !

!Project methodsFor: 'project parameters' stamp: 'ar 5/25/2000 23:23'!
parameterAt: aSymbol ifAbsent: aBlock
	projectParameters ifNil:[^aBlock value].
	^projectParameters at: aSymbol ifAbsent: aBlock! !

!Project methodsFor: 'accessing' stamp: 'di 3/6/98 10:22'!
parent
	^ parentProject! !

!Project methodsFor: 'accessing' stamp: ''!
projectChangeSet
	^ changeSet! !

!Project methodsFor: 'project parameters' stamp: 'sw 2/17/2001 21:36'!
projectParameterAt: aSymbol put: aValue
	"Set the given project parameter to the given value"

	self projectParameters at: aSymbol put: aValue.
	self noteThatParameter: aSymbol justChangedTo: aValue.
	^ aValue! !

!Project methodsFor: 'project parameters' stamp: 'sw 4/22/1999 15:14'!
projectParameters
	^ projectParameters ifNil: [self initializeProjectParameters]! !

!Project methodsFor: 'initialization' stamp: 'StephaneDucasse 3/5/2010 14:27'!
setChangeSet: aChangeSet

	changeSet := aChangeSet
! !

!Project methodsFor: 'file in/out' stamp: 'tk 8/21/1999 07:31'!
storeDataOn: aDataStream
	"Write me out.  All references to other projects are weak references.  They only go out if they are written for another reason."
	| cntInstVars cntIndexedVars localInstVars offset |

	cntInstVars := self class instSize.
	cntIndexedVars := self basicSize.
	localInstVars := Project instVarNames.
	offset := Project superclass instSize.
	aDataStream
		beginInstance: self class
		size: cntInstVars + cntIndexedVars.
	1 to: cntInstVars do:
		[:ii | 
		(ii between: offset+1 and: offset + localInstVars size) 
			ifTrue: [(#('parentProject' 'previousProject' 'nextProject') includes: 
				(localInstVars at: ii-offset)) 
					ifTrue: [aDataStream nextPutWeak: (self instVarAt: ii)]
								"owner only written if in our tree"
					ifFalse: [aDataStream nextPut: (self instVarAt: ii)]]
			ifFalse: [aDataStream nextPut: (self instVarAt: ii)]].

	1 to: cntIndexedVars do:
		[:i | aDataStream nextPut: (self basicAt: i)]! !

!Project methodsFor: 'language' stamp: 'AlainPlantec 10/17/2009 20:32'!
updateLocaleDependents
	"Set the project's natural language as indicated"
	MenuIcons initializeTranslations.
	LanguageEnvironment localeChanged.
! !

!Project methodsFor: 'accessing' stamp: ''!
world
	^ world! !

Object subclass: #Author
	instanceVariableNames: 'fullName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

Author class
	instanceVariableNames: 'uniqueInstance'!

!Author commentStamp: 'MiguelCoba 7/25/2009 01:09' prior: 0!
I am responsible for the full name used to identify the current code author.!

Author class
	instanceVariableNames: 'uniqueInstance'!

!Author class methodsFor: 'utilities' stamp: 'JB 2/12/2010 15:20'!
changeStamp 
	"Answer a string to be pasted into source code to mark who changed it and when."
	^ Author fullName , ' ' , Date today mmddyyyy, ' ',
		((String streamContents: [:s | Time now print24: true on: s]) copyFrom: 1 to: 5)! !

!Author class methodsFor: 'utilities' stamp: 'JB 2/12/2010 15:20'!
changeStampPerSe
	"Answer a string to be pasted into source code to mark who changed it and when."

	^ (Author fullNamePerSe ifNil: ['.']) , ' ' , Date today mmddyyyy, ' ',
		((String streamContents: [:s | Time now print24: true on: s]) 
			copyFrom: 1 to: 5)! !

!Author class methodsFor: 'utilities' stamp: 'JB 2/12/2010 15:23'!
fixStamp: changeStamp 
	| parts |
	parts := changeStamp findTokens: ' '.
	(parts size > 0 and: [parts last first isLetter]) ifTrue:
		["Put initials first in all time stamps..."
		^ String streamContents:
				[:s | s nextPutAll: parts last.
				parts allButLast do: [:p | s space; nextPutAll: p]]].
	^ changeStamp! !

!Author class methodsFor: 'convenience' stamp: 'MiguelCoba 7/25/2009 01:58'!
fullName
	^ Author uniqueInstance fullName! !

!Author class methodsFor: 'convenience' stamp: 'MiguelCoba 7/25/2009 01:58'!
fullName: aString
	^ Author uniqueInstance fullName: aString! !

!Author class methodsFor: 'convenience' stamp: 'MiguelCoba 7/25/2009 01:58'!
fullNamePerSe
	^ Author uniqueInstance fullNamePerSe! !

!Author class methodsFor: 'instance creation' stamp: 'on 5/10/2008 13:05'!
new
	self error: 'Author is a singleton -- send uniqueInstance instead'! !

!Author class methodsFor: 'convenience' stamp: 'MiguelCoba 7/25/2009 01:10'!
requestFullName
	^ Author uniqueInstance requestFullName! !

!Author class methodsFor: 'instance creation' stamp: 'on 5/10/2008 13:10'!
reset
	^ uniqueInstance := nil! !

!Author class methodsFor: 'instance creation' stamp: 'on 5/10/2008 13:09'!
uniqueInstance
	^ uniqueInstance ifNil: [ uniqueInstance := super new ]! !

!Author class methodsFor: 'instance creation' stamp: 'on 5/10/2008 13:27'!
uniqueInstance: anInstance
	"Needed by AuthorTest to restore saved instance"
	^ uniqueInstance := anInstance! !

!Author class methodsFor: 'testing-support' stamp: 'oscar.nierstrasz 10/18/2009 18:31'!
useAuthor: aString during: aBlock
	^ self uniqueInstance useAuthor: aString during: aBlock! !

!Author methodsFor: 'accessing' stamp: 'MarcusDenker 2/21/2010 15:31'!
fullName
	"Answer the full name to be used to identify the current code author."

	[fullName isEmptyOrNil] whileTrue: [
		self requestFullName.
		(fullName isNil or:[fullName = 'FirstnameLastname' translated]) ifTrue: [fullName := nil. ^'Anonymous']].
	^ fullName! !

!Author methodsFor: 'accessing' stamp: 'MiguelCoba 7/25/2009 00:55'!
fullName: aString
	fullName := aString.! !

!Author methodsFor: 'accessing' stamp: 'MiguelCoba 7/25/2009 00:58'!
fullNamePerSe

	"Answer the currently-prevailing author full name, such as it is, empty or not"

	^ fullName! !

!Author methodsFor: 'compatibility' stamp: 'StephaneDucasse 5/13/2010 10:28'!
ifUnknownAuthorUse: aString during: aBlock
	"If no author is set use the name aString while executing aBlock."
	"for compatibility with 1.0"
	
	| previous |
	fullName isEmptyOrNil ifFalse: [ ^ aBlock value ].
	fullName := aString.
	^ aBlock ensure: [ self reset ]! !

!Author methodsFor: 'initialization' stamp: 'MiguelCoba 7/25/2009 03:07'!
initialize
	super initialize.
	fullName := ''.! !

!Author methodsFor: 'ui-requests' stamp: 'MiguelCoba 5/30/2010 22:19'!
messagePrompt
	^
'Please type your full name. 
It will be used to sign the changes you make to the image.
Avoid spaces, accents, dashes, underscore and similar characters. E.g.

Vincent van Gogh -> VincentVanGogh
Miguel Cob -> MiguelCoba
'! !

!Author methodsFor: 'ui-requests' stamp: 'MiguelCoba 7/25/2009 01:54'!
requestFullName
	| initialAnswer |
	initialAnswer := fullName isEmptyOrNil
						ifTrue: ['FirstnameLastname' translated]
						ifFalse: [fullName].
	fullName := UIManager default 
			request: self messagePrompt
			initialAnswer: initialAnswer! !

!Author methodsFor: 'compatibility' stamp: 'tg 6/5/2010 19:20'!
reset
	fullName := ''! !

!Author methodsFor: 'testing-support' stamp: 'oscar.nierstrasz 10/18/2009 18:27'!
useAuthor: aString during: aBlock
	| previous |
	previous := fullName.
	fullName := aString.
	^ aBlock ensure: [ fullName := previous ]! !

Object subclass: #AutoStart
	instanceVariableNames: 'parameters'
	classVariableNames: 'Active InstalledLaunchers UpdateFromServerAtStartup'
	poolDictionaries: ''
	category: 'System-Support'!

!AutoStart class methodsFor: 'private' stamp: 'mir 9/7/2004 13:34'!
active
	^ Active == true! !

!AutoStart class methodsFor: 'private' stamp: 'mir 9/7/2004 13:36'!
active: aBoolean
	Active := aBoolean! !

!AutoStart class methodsFor: 'accessing' stamp: 'mir 7/28/1999 17:47'!
addLauncher: launcher
	self installedLaunchers add: launcher! !

!AutoStart class methodsFor: 'accessing' stamp: ''!
addLauncherFirst: launcher
	self installedLaunchers addFirst: launcher! !

!AutoStart class methodsFor: 'updating' stamp: 'MarcusDenker 2/23/2010 16:45'!
checkForUpdates
	| availableUpdate updateServer |
	World 
		ifNotNil: [
			World install.
			ActiveHand position: 100@100].
	^self processUpdates.
	! !

!AutoStart class methodsFor: 'initialization' stamp: 'mir 7/28/1999 17:44'!
deinstall
	"AutoStart deinstall"

	Smalltalk removeFromStartUpList: AutoStart.
	InstalledLaunchers := nil! !

!AutoStart class methodsFor: 'initialization' stamp: 'mir 9/30/2004 15:05'!
initialize
	"AutoStart initialize"
	"Order: ExternalSettings, SecurityManager, AutoStart"
	Smalltalk addToStartUpList: AutoStart after: SecurityManager.
	Smalltalk addToShutDownList: AutoStart after: SecurityManager.! !

!AutoStart class methodsFor: 'private' stamp: 'mir 7/28/1999 17:43'!
installedLaunchers
	InstalledLaunchers ifNil: [
		InstalledLaunchers := OrderedCollection new].
	^InstalledLaunchers! !

!AutoStart class methodsFor: 'updating' stamp: 'AlainPlantec 1/7/2010 21:37'!
processUpdates
	"Process update files from a well-known update server.  This method is called at system startup time,   Only if the UpdateFromServerAtStartup class variable is true is the actual update processing undertaken automatically"
	| choice |
	(self updateFromServerAtStartup) ifTrue:
		[choice := UIManager default chooseFrom: #('Yes, Update' 'No, Not now' 'Don''t ask again')
			title: 'Shall I look for new code\updates on the server?' withCRs.
		choice = 1 ifTrue: [
			Utilities updateFromServer].
		choice = 3 ifTrue: [
			self updateFromServerAtStartup: false.
			self inform: 'Remember to save you image to make this setting permant.']].
	^false! !

!AutoStart class methodsFor: 'accessing' stamp: 'mir 7/28/1999 17:47'!
removeLauncher: launcher
	self installedLaunchers remove: launcher ifAbsent: []! !

!AutoStart class methodsFor: 'initialization' stamp: 'mir 9/30/2004 15:06'!
shutDown: quitting
	self active: false! !

!AutoStart class methodsFor: 'initialization' stamp: 'pavel.krivanek 9/2/2010 21:49'!
startUp: resuming
	"The image is either being newly started (resuming is true), or it's just been snapshotted.
	If this has just been a snapshot, skip all the startup stuff."

	| startupParameters launchers ws |
	self active
		ifTrue: [ ^ self ].
	self active: true.
	resuming
		ifFalse: [ ^ self ].
	self checkForUpdates
		ifTrue: [ ^ self ].
	startupParameters := SmalltalkImage current extractParameters.
	launchers := self installedLaunchers collect: [ :launcher | launcher new ].
	launchers do: [ :launcher | launcher parameters: startupParameters ].
	ws := Smalltalk globals at: #WorldState ifAbsent: [ nil ].
	(ws notNil and: [Smalltalk globals includesKey: #World])
		ifTrue: [
			launchers do: [:launcher |
				ws addDeferredUIMessage: [launcher startUp]] ]
		ifFalse: [
			launchers do: [:launcher |
				launcher startUp]].
! !

!AutoStart class methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 22:50'!
updateFromServerAtStartup
	^ UpdateFromServerAtStartup ifNil: [UpdateFromServerAtStartup := false]! !

!AutoStart class methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 22:50'!
updateFromServerAtStartup: aBoolean
	UpdateFromServerAtStartup := aBoolean! !

Object subclass: #Beeper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

Beeper class
	instanceVariableNames: 'default'!

!Beeper commentStamp: 'gk 2/26/2004 22:44' prior: 0!
Beeper provides simple audio (or in some other way) feedback to the user.

The recommended use is "Beeper beep" to give the user the equivalence of a beep. If you want to force the beep to use the primitive in the VM for beeping, then use "Beeper beepPrimitive". In either case, if sounds are disabled there will be no beep.

The actual beeping, when you use "Beeper beep", is done by sending a #play message to a registered playable object. You can register your own playable object by invoking the class side method #setDefault: passing in an object that responds to the #play message.

The default playable object is an instance of Beeper itself which implements #play on the instance side. That implementation delegates the playing of the beep to the default SoundService.

Note that #play is introduced as a common interface between AbstractSound and Beeper.
This way we can register instances of AbstractSound as playable entities, for example:

	Beeper setDefault: (SampledSound new
						setSamples: self coffeeCupClink
						samplingRate: 12000).

Then "Beeper beep" will play the coffeeCup sound.!

Beeper class
	instanceVariableNames: 'default'!

!Beeper class methodsFor: 'beeping' stamp: 'pavel.krivanek 3/11/2009 07:39'!
beep
	"The preferred way of producing an audible feedback.
	The default playable entity (an instance of Beeper)
	also uses the pluggable SoundService
	mechanism, so it will use the primitive beep only
	if there is no other sound mechanism available."
	
	self default 
		ifNil: [self beepPrimitive]
		ifNotNil: [ self default play].
! !

!Beeper class methodsFor: 'beeping' stamp: 'AlainPlantec 12/3/2009 09:05'!
beepPrimitive
	"Make a primitive beep. Only use this if
	you want to force this to be a primitive beep.
	Otherwise use Beeper class>>beep
	since this method bypasses the current
	registered playable entity."

	SoundSettings soundEnabled 
		ifTrue: [self primitiveBeep]! !

!Beeper class methodsFor: 'customize' stamp: 'gk 2/22/2004 17:51'!
clearDefault
	"Clear the default playable.
	Will be lazily initialized in Beeper class >>default."

	default := nil! !

!Beeper class methodsFor: 'customize' stamp: 'gk 2/22/2004 17:55'!
default
	"When the default is not defined it is
	initialized using #newDefault."

	default isNil 
		ifTrue: [default := self newDefault ].
	^ default! !

!Beeper class methodsFor: 'customize' stamp: 'gk 2/24/2004 22:12'!
newDefault
	"Subclasses may override me to provide a default beep.
	This base implementation returns an instance of Beeper
	which uses the pluggable sound service."

	^ self new! !

!Beeper class methodsFor: 'private' stamp: 'AlainPlantec 1/7/2010 21:37'!
primitiveBeep
	"Make a primitive beep. Not to be called directly.
	It is much better to use Beeper class>>beep
	or Beeper class>>beepPrimitive
	since this method bypasses the current
	registered playable entity and does not
	check SoundSettings class>>soundEnabled."

	<primitive: 140>
	self primitiveFailed! !

!Beeper class methodsFor: 'customize' stamp: 'gk 2/22/2004 17:54'!
setDefault: aPlayableEntity
	"Set the playable entity used when making a beep.
	The playable entity should implement the message #play."

	default := aPlayableEntity! !

!Beeper methodsFor: 'play interface' stamp: 'gk 2/24/2004 23:25'!
play
	"This is how the default Beeper makes a beep,
	by sending beep to the default sound service.
	The sound system will check if sounds are enabled."

	SoundService default beep! !

Object subclass: #DefaultExternalDropHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!DefaultExternalDropHandler commentStamp: 'dgd 4/5/2004 19:07' prior: 0!
An alternative default handler that uses the file-list services to process files.
!

!DefaultExternalDropHandler class methodsFor: 'initialization' stamp: 'dgd 4/5/2004 19:10'!
initialize
	"initialize the receiver"
	ExternalDropHandler defaultHandler: self new! !

!DefaultExternalDropHandler class methodsFor: 'initialization' stamp: 'dgd 4/5/2004 19:09'!
unload
	"initialize the receiver"
	ExternalDropHandler defaultHandler: nil! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: 'alain.plantec 2/8/2009 22:02'!
chooseServiceFrom: aCollection
	"private - choose a service from aCollection asking the user if  
	needed"
	aCollection size = 1
		ifTrue: [^ aCollection anyOne].
	^ UIManager default chooseFrom: (aCollection collect: [:each | each label]) values: aCollection.
! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: 'StephaneDucasse 9/9/2010 09:57'!
servicesForFileNamed: aString 
	"private - answer a collection of file-services for the file named  
	aString"
	| allServices |
	allServices := FileServices itemsForFile: aString.
	^ allServices
		reject: [:svc | self unwantedSelectors includes: svc selector]! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: 'dgd 4/5/2004 19:23'!
unwantedSelectors
	"private - answer a collection well known unwanted selectors "
	^ #(#removeLineFeeds: #addFileToNewZip: #compressFile: #putUpdate: )! !

Object subclass: #ExternalDropHandler
	instanceVariableNames: 'action type extension'
	classVariableNames: 'DefaultHandler RegisteredHandlers'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalDropHandler class methodsFor: 'private' stamp: 'MarcusDenker 12/12/2009 00:05'!
chooseServiceFrom: aCollection
	"private - choose a service from aCollection asking the user if  
	needed"
	aCollection size = 1 ifTrue: [^ aCollection anyOne].
	^ UIManager default chooseFrom: (aCollection collect: [:each | 
			each label]) values: aCollection.
! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'mir 1/10/2002 17:23'!
defaultGZipHandler
	^ExternalDropHandler
		type: nil
		extension: 'gz'
		action: [:stream :pasteUp :event |
			stream viewGZipContents]! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 1/10/2002 17:17'!
defaultHandler
	DefaultHandler ifNil: [DefaultHandler := ExternalDropHandler type: nil extension: nil action: [:dropStream | dropStream edit]].
	^DefaultHandler! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 1/10/2002 16:54'!
defaultHandler: externalDropHandler
	DefaultHandler := externalDropHandler! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'MarcusDenker 2/15/2010 22:40'!
defaultImageHandler
	
	^ExternalDropHandler
		type: 'image/'
		extension: nil
		action: [:stream :pasteUp :event | | image sketch |
			stream binary.
			image := Form fromBinaryStream: ((RWBinaryOrTextStream with: stream contents) reset).
			sketch := World drawingClass withForm: image.
			pasteUp addMorph: sketch centeredNear: event position.
			image := sketch := nil]! !

!ExternalDropHandler class methodsFor: 'initialization' stamp: 'StephaneDucasse 11/6/2009 17:43'!
initialize
	"ExternalDropHandler initialize"

	self resetRegisteredHandlers.
	self
		registerHandler: self defaultImageHandler;
		registerHandler: self defaultGZipHandler.
	! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 8/24/2004 15:37'!
lookupExternalDropHandler: stream

	| types extension serviceHandler |
	types := stream mimeTypes.

	types ifNotNil: [
		self registeredHandlers do: [:handler | 
			(handler matchesTypes: types)
				ifTrue: [^handler]]].

	extension := FileDirectory extensionFor: stream name.
	self registeredHandlers do: [:handler | 
		(handler matchesExtension: extension)
				ifTrue: [^handler]].
	serviceHandler := self lookupServiceBasedHandler: stream.
	^serviceHandler
		ifNil: [self defaultHandler]! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'StephaneDucasse 9/9/2010 14:35'!
lookupServiceBasedHandler: dropStream
	"the file was just droped, let's do our job"
	| fileName services theOne |
	fileName := dropStream name.

	services := (FileList itemsForFile: fileName)
		reject: [:svc | self unwantedSelectors includes: svc selector].

	"no service, default behavior"
	services isEmpty
		ifTrue: [^nil].

	theOne := self chooseServiceFrom: services.
	^theOne
		ifNotNil: [ExternalDropHandler type: nil extension: nil action: [:stream | theOne performServiceFor: stream]]! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: 'mir 1/10/2002 17:19'!
registerHandler: aHandler
	self registeredHandlers add: aHandler! !

!ExternalDropHandler class methodsFor: 'initialization' stamp: 'nk 6/12/2004 16:15'!
registerStandardExternalDropHandlers
	"ExternalDropHandler registerStandardExternalDropHandlers"

	self registeredHandlers add: (
		ExternalDropHandler
			type: 'image/'
			extension: nil
			action: [:stream :pasteUp :event |
				pasteUp addMorph: (World drawingClass withForm: (Form fromBinaryStream: stream binary)) centeredNear: event position])! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'mir 1/10/2002 15:57'!
registeredHandlers
	RegisteredHandlers ifNil: [RegisteredHandlers := OrderedCollection new].
	^RegisteredHandlers! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'mir 1/10/2002 15:57'!
resetRegisteredHandlers
	RegisteredHandlers := nil! !

!ExternalDropHandler class methodsFor: 'instance creation' stamp: 'mir 1/10/2002 17:16'!
type: aType extension: anExtension action: anAction 
	^self new type: aType extension: anExtension action: anAction ! !

!ExternalDropHandler class methodsFor: 'private' stamp: 'mir 8/24/2004 15:28'!
unwantedSelectors
	"private - answer a collection well known unwanted selectors "
	^ #(#removeLineFeeds: #addFileToNewZip: #compressFile: #putUpdate: )! !

!ExternalDropHandler methodsFor: 'accessing' stamp: 'mir 1/10/2002 15:54'!
extension
	^extension! !

!ExternalDropHandler methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 6/28/2010 15:01'!
handle: dropStream in: pasteUp dropEvent: anEvent
	
	^action cull: dropStream 
			cull: pasteUp 
			cull: anEvent
	! !

!ExternalDropHandler methodsFor: 'testing' stamp: 'spfa 5/25/2004 13:38'!
matchesExtension: aExtension
	(self extension isNil or: [aExtension isNil])
		ifTrue: [^false].
	FileDirectory activeDirectoryClass isCaseSensitive
		ifTrue: [^extension = aExtension]
		ifFalse: [^extension sameAs: aExtension]! !

!ExternalDropHandler methodsFor: 'testing' stamp: 'mir 1/10/2002 16:35'!
matchesTypes: types
	(self type isNil or: [types isNil])
		ifTrue: [^false].
	^types anySatisfy: [:mimeType | mimeType beginsWith: self type]! !

!ExternalDropHandler methodsFor: 'accessing' stamp: 'mir 1/10/2002 15:54'!
type
	^type! !

!ExternalDropHandler methodsFor: 'initialize' stamp: 'mir 1/10/2002 17:17'!
type: aType extension: anExtension action: anAction 
	action := anAction.
	type := aType.
	extension := anExtension! !

Object subclass: #ExternalSemaphoreTable
	instanceVariableNames: ''
	classVariableNames: 'ProtectTable'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalSemaphoreTable commentStamp: '<historical>' prior: 0!
By John M McIntosh johnmci@smalltalkconsulting.com
This class was written to mange the external semaphore table. When I was writing a Socket test server I discovered various race conditions on the access to the externalSemaphore table. This new class uses class side methods to restrict access using a mutex semaphore. It seemed cleaner to deligate the reponsibility here versus adding more code and another class variable to SystemDictionary 

Note that in Smalltalk recreateSpecialObjectsArray we still directly play with the table.!

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: 'JMM 6/6/2000 20:36'!
clearExternalObjects
	"Clear the array of objects that have been registered for use in non-Smalltalk code."

	ProtectTable critical: [Smalltalk specialObjectsArray at: 39 put: Array new].
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: 'JMM 6/6/2000 21:01'!
externalObjects
	^ProtectTable critical: [Smalltalk specialObjectsArray at: 39].! !

!ExternalSemaphoreTable class methodsFor: 'initialize' stamp: 'JMM 6/6/2000 20:32'!
initialize
	ProtectTable := Semaphore forMutualExclusion! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: 'JMM 6/6/2000 20:44'!
registerExternalObject: anObject
	^ ProtectTable critical: [self safelyRegisterExternalObject: anObject]
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: 'JMM 6/6/2000 20:57'!
safelyRegisterExternalObject: anObject
	"Register the given object in the external objects array and return its index. If it is already there, just return its index."

	| objects firstEmptyIndex obj sz newObjects |
	objects := Smalltalk specialObjectsArray at: 39.

	"find the first empty slot"
	firstEmptyIndex := 0.
	1 to: objects size do: [:i |
		obj := objects at: i.
		obj == anObject ifTrue: [^ i].  "object already there, just return its index"
		(obj == nil and: [firstEmptyIndex = 0]) ifTrue: [firstEmptyIndex := i]].

	"if no empty slots, expand the array"
	firstEmptyIndex = 0 ifTrue: [
		sz := objects size.
		newObjects := objects species new: sz + 20.  "grow linearly"
		newObjects replaceFrom: 1 to: sz with: objects startingAt: 1.
		firstEmptyIndex := sz + 1.
		Smalltalk specialObjectsArray at: 39 put: newObjects.
		objects := newObjects].

	objects at: firstEmptyIndex put: anObject.
	^ firstEmptyIndex
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: 'JMM 6/6/2000 20:59'!
safelyUnregisterExternalObject: anObject
	"Unregister the given object in the external objects array. Do nothing if it isn't registered.
	JMM change to return if we clear the element, since it should only appear once in the array"

	| objects |
	anObject ifNil: [^ self].
	objects := Smalltalk specialObjectsArray at: 39.
	1 to: objects size do: [:i |
		(objects at: i) == anObject ifTrue: 
		[objects at: i put: nil.
		^self]].
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: 'JMM 6/6/2000 20:45'!
unregisterExternalObject: anObject
	ProtectTable critical: [self safelyUnregisterExternalObject: anObject]
! !

Object subclass: #ExternalSettings
	instanceVariableNames: ''
	classVariableNames: 'RegisteredClients'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalSettings commentStamp: '<historical>' prior: 0!
ExternalSettings manages settings kept externally, e.g. files.
Objects can register themselves as clients to be notified at startup time to read their settings.

Eventually all the preferences should be managed through this mechanism.
!

!ExternalSettings class methodsFor: 'accessing' stamp: 'sw 1/25/2002 12:39'!
assuredPreferenceDirectory
	"Answer the preference directory, creating it if necessary"

	|  prefDir |
	prefDir := self preferenceDirectory.
	prefDir
		ifNil:
			[prefDir := FileDirectory default directoryNamed: self preferenceDirectoryName.
			prefDir assureExistence].
	^ prefDir! !

!ExternalSettings class methodsFor: 'initialization' stamp: 'ar 8/23/2001 22:56'!
initialize
	"ExternalSettings initialize"
	"Order: ExternalSettings, SecurityManager, AutoStart"
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'dc 5/30/2008 10:17'!
parseServerEntryArgsFrom: stream 
	"Args are in the form <argName>: <argValueString> delimited by end of line.
	It's not a very robust format and should be replaced by something like XML later.
	But it avoids evaluating the entries for security reasons."
	| entries lineStream entryName entryValue |
	entries := Dictionary new.
	stream skipSeparators.
	[ stream atEnd ] whileFalse: 
		[ lineStream := stream nextLine readStream.
		entryName := lineStream upTo: $:.
		lineStream skipSeparators.
		entryValue := lineStream upToEnd.
		(entryName isEmptyOrNil or: [ entryValue isEmptyOrNil ]) ifFalse: 
			[ entries 
				at: entryName
				put: entryValue withoutTrailingBlanks ].
		stream skipSeparators ].
	^ entries! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'StephaneDucasse 3/17/2010 20:56'!
preferenceDirectory
	| prefDirName path |
	prefDirName := self preferenceDirectoryName.
	path := Smalltalk vmPath.
	^(FileDirectory default directoryExists: prefDirName)
		ifTrue: [FileDirectory default directoryNamed: prefDirName]
		ifFalse: [
			((FileDirectory on: path) directoryExists: prefDirName)
				ifTrue: [(FileDirectory on: path) directoryNamed: prefDirName]
				ifFalse: [nil]]
! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'mir 11/16/2001 13:33'!
preferenceDirectoryName
	^'prefs'! !

!ExternalSettings class methodsFor: 'accessing' stamp: 'mir 6/25/2001 18:45'!
registerClient: anObject
	"Register anObject as a settings client to be notified on startup."

	self registeredClients add: anObject! !

!ExternalSettings class methodsFor: 'private' stamp: 'mir 6/25/2001 18:46'!
registeredClients
	RegisteredClients ifNil: [RegisteredClients := Set new].
	^RegisteredClients! !

!ExternalSettings class methodsFor: 'initialization' stamp: 'mir 8/22/2001 15:17'!
shutDown
	"Look for external defs and load them."
	"ExternalSettings shutDown"

	self registeredClients do: [:client | 
		client releaseExternalSettings]! !

!ExternalSettings class methodsFor: 'initialization' stamp: 'mir 11/16/2001 13:29'!
startUp
	"Look for external defs and load them."
	"ExternalSettings startUp"

	| prefDir |
	prefDir := self preferenceDirectory.
	prefDir
		ifNil: [^self].
	self registeredClients do: [:client | 
		client fetchExternalSettingsIn: prefDir]! !

Object subclass: #LRUCache
	instanceVariableNames: 'size factory calls hits values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!LRUCache commentStamp: '<historical>' prior: 0!
I'm a cache of values, given a key I return a Value from the cache or from the factory!

!LRUCache class methodsFor: 'instance creation' stamp: 'dgd 3/26/2003 22:29'!
size: aNumber factory: aBlock 
	"answer an instance of the receiver"
	^ self new initializeSize: aNumber factory: aBlock! !

!LRUCache class methodsFor: 'testing' stamp: 'dgd 3/26/2003 22:22'!
test
	" 
	LRUCache test 
	"
	| c |
	c := LRUCache
				size: 5
				factory: [:key | key * 2].
	c at: 1.
	c at: 2.
	c at: 3.
	c at: 4.
	c at: 1.
	c at: 5.
	c at: 6.
	c at: 7.
	c at: 8.
	c at: 1.
	^ c! !

!LRUCache class methodsFor: 'testing' stamp: 'dgd 3/26/2003 22:22'!
test2
	" 
	LRUCache test2.  
	Time millisecondsToRun:[LRUCache test2]. 
	MessageTally spyOn:[LRUCache test2].  
	"
	| c |
	c := LRUCache
				size: 600
				factory: [:key | key * 2].
	1
		to: 6000
		do: [:each | c at: each].
	^ c! !

!LRUCache methodsFor: 'accessing' stamp: 'pmm 3/13/2010 11:31'!
at: aKey 
	"answer the object for aKey, if not present in the cache creates it.
	Clone the factory block before calling in case of multiple processes!!"
	
	| element keyHash |
	calls := calls + 1.
	keyHash := aKey hash.
	1
		to: size
		do: [:index | 
			element := values at: index.
			(keyHash
						= (element at: 2)
					and: [aKey
							= (element at: 1)])
				ifTrue: ["Found!!"
					hits := hits + 1.
					values
						replaceFrom: 2
						to: index
						with: (values first: index - 1).
					values at: 1 put: element.
					^ element at: 3]].
	"Not found!!"
	element := {aKey. keyHash. factory shallowCopy value: aKey}.
	values
		replaceFrom: 2
		to: size
		with: values allButLast.
	values at: 1 put: element.
	^ element at: 3! !

!LRUCache methodsFor: 'initialization' stamp: 'dgd 3/28/2003 19:42'!
initializeSize: aNumber factory: aBlock 
	"initialize the receiver's size and factory"
	size := aNumber.
	values := Array new: aNumber withAll: {nil. nil. nil}.
	factory := aBlock.
	calls := 0.
	hits := 0! !

!LRUCache methodsFor: 'printing' stamp: 'StephaneDucasse 3/13/2010 13:26'!
printOn: aStream 
	"Append to the argument, aStream, a sequence of characters 
	that identifies the receiver."
	aStream nextPutAll: self class name;
		 nextPutAll: ' size:';
		 nextPutAll: size asString;
		 nextPutAll: ', calls:';
		 nextPutAll: calls asString;
		 nextPutAll: ', hits:';
		 nextPutAll: hits asString;
		 nextPutAll: ', ratio:'.
	calls = 0
		ifTrue: [aStream nextPutAll: 0 asFloat asString]
		ifFalse: [aStream nextPutAll: (hits / calls) asFloat asString]! !

Object subclass: #MethodReference
	instanceVariableNames: 'classSymbol classIsMeta methodSymbol stringVersion category package'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!MethodReference commentStamp: 'tlk 5/9/2006 18:43' prior: 0!
A MethodReference is is a lightweight proxy for a CompiledMethod.  Has methods for pointed to the CompileMethod's source statements, byte codes. Is heavily used my Tools.

Instance Variables
	classIsMeta:		     Boolean class vs. instance
	classSymbol:		Symbol for method's class (without class keyword if meta)
	methodSymbol:		Symbol for method's selector
	stringVersion:		'Class>>selector:' format

!

!MethodReference class methodsFor: 'instance creation' stamp: 'ab 2/6/2005 16:22'!
class: aClass selector: aSelector
	^ self new setStandardClass: aClass methodSymbol: aSelector! !

!MethodReference class methodsFor: 'Package' stamp: 'stephane.ducasse 7/10/2009 09:27'!
class: aClass selector: aSelector package: aPackage

	^ (self class: aClass selector: aSelector ) package: aPackage ; yourself! !

!MethodReference methodsFor: 'comparisons' stamp: 'RAA 5/28/2001 11:56'!
<= anotherMethodReference

	classSymbol < anotherMethodReference classSymbol ifTrue: [^true].
	classSymbol > anotherMethodReference classSymbol ifTrue: [^false].
	classIsMeta = anotherMethodReference classIsMeta ifFalse: [^classIsMeta not].
	^methodSymbol <= anotherMethodReference methodSymbol
! !

!MethodReference methodsFor: 'comparisons' stamp: 'StephaneDucasse 4/27/2010 11:46'!
= anotherMethodReference 
	"Answer whether the receiver and the argument represent the 
	same object."
	^ self species == anotherMethodReference species
		and: [(self classSymbol = anotherMethodReference classSymbol)
				and: [(self classIsMeta = anotherMethodReference classIsMeta)
						and: [self methodSymbol = anotherMethodReference methodSymbol]]]! !

!MethodReference methodsFor: 'accessing' stamp: 'lr 3/14/2010 21:13'!
actualClass
	| actualClass traitName |
	('*classTrait' match: classSymbol)
		ifTrue: [ 
			traitName := classSymbol copyUpTo: Character space.
			^ Smalltalk globals at: traitName asSymbol ifAbsent: [ nil ] ].
	actualClass := Smalltalk globals at: classSymbol ifAbsent: [ ^ nil ].
	classIsMeta
		ifTrue: [ ^ actualClass classSide ].
	^ actualClass! !

!MethodReference methodsFor: 'queries' stamp: 'StephaneDucasse 5/31/2010 21:59'!
asStringOrText
	| text |
	self actualClass ifNil: [ ^stringVersion ].
	text := Text fromString: self stringVersion,' {',self category asString,'} '.
	self isLocalSelector ifFalse: [ text addAttribute: TextEmphasis italic ].
	^ text! !

!MethodReference methodsFor: 'queries' stamp: 'stephane.ducasse 10/12/2008 20:01'!
category
	^ category ifNil: [category := self actualClass organization categoryOfElement: methodSymbol]! !

!MethodReference methodsFor: 'string version' stamp: 'stephane.ducasse 10/12/2008 20:08'!
category: aString
	
	 category := aString! !

!MethodReference methodsFor: 'queries' stamp: 'RAA 5/28/2001 08:11'!
classIsMeta

	^classIsMeta! !

!MethodReference methodsFor: 'queries' stamp: 'RAA 5/28/2001 08:10'!
classSymbol

	^classSymbol! !

!MethodReference methodsFor: 'comparisons' stamp: 'dgd 3/8/2003 11:54'!
hash
	"Answer a SmallInteger whose value is related to the receiver's  
	identity."
	^ (self species hash bitXor: self classSymbol hash)
		bitXor: self methodSymbol hash! !

!MethodReference methodsFor: 'queries' stamp: 'StephaneDucasse 2/6/2010 16:57'!
isValid
	"Answer whether the receiver represents a current selector or Comment"

	| aClass |
	methodSymbol isDoIt ifTrue: [^ false].
	(aClass := self actualClass) ifNil: [^ false].
	^ (aClass includesSelector: methodSymbol) or:
		[methodSymbol == #Comment]! !

!MethodReference methodsFor: 'accessing' stamp: 'stephane.ducasse 6/2/2009 16:02'!
methodClass

	^ self actualClass ! !

!MethodReference methodsFor: 'queries' stamp: 'RAA 5/28/2001 08:10'!
methodSymbol

	^methodSymbol! !

!MethodReference methodsFor: 'accessing' stamp: 'stephane.ducasse 7/8/2009 11:26'!
package
	"we will have to change that in the future since we will not rely on * convention"
	
	^ package! !

!MethodReference methodsFor: 'accessing' stamp: 'stephane.ducasse 7/8/2009 11:27'!
package: aPackage
	
	package := aPackage! !

!MethodReference methodsFor: 'queries' stamp: 'stephane.ducasse 10/27/2008 20:40'!
printOn: aStream
	"Print the receiver on a stream"

	super printOn: aStream.
	aStream nextPutAll: ' ', self actualClass name, ' >> #', methodSymbol! !

!MethodReference methodsFor: 'accessing' stamp: 'stephane.ducasse 7/22/2009 11:47'!
selector

	^ methodSymbol! !

!MethodReference methodsFor: 'setting' stamp: 'RAA 5/28/2001 08:06'!
setClass: aClass methodSymbol: methodSym stringVersion: aString 

	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := methodSym.
	stringVersion := aString.! !

!MethodReference methodsFor: 'setting' stamp: 'RAA 5/28/2001 07:34'!
setClassAndSelectorIn: csBlock

	^csBlock value: self actualClass value: methodSymbol! !

!MethodReference methodsFor: 'setting' stamp: 'RAA 5/28/2001 06:04'!
setClassSymbol: classSym classIsMeta: isMeta methodSymbol: methodSym stringVersion: aString 

	classSymbol := classSym.
	classIsMeta := isMeta.
	methodSymbol := methodSym.
	stringVersion := aString.! !

!MethodReference methodsFor: 'setting' stamp: 'RAA 5/28/2001 11:34'!
setStandardClass: aClass methodSymbol: methodSym

	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := methodSym.
	stringVersion := aClass name , ' ' , methodSym.! !

!MethodReference methodsFor: 'queries' stamp: 'sr 6/4/2004 01:55'!
sourceString
	^ (self actualClass sourceCodeAt: self methodSymbol) asString! !

!MethodReference methodsFor: 'string version' stamp: 'JB 2/18/2010 10:26'!
stringVersion	
	^stringVersion.! !

!MethodReference methodsFor: 'string version' stamp: 'RAA 5/29/2001 14:44'!
stringVersion: aString

	stringVersion := aString! !

!MethodReference methodsFor: 'string version' stamp: 'StephaneDucasse 12/22/2009 11:07'!
stringVersionDefault

	^classSymbol, (classIsMeta ifTrue: [ ' class ' ] ifFalse: [' ']), methodSymbol  ! !

!Object methodsFor: '*System-Support' stamp: 'dvf 8/23/2003 12:27'!
systemNavigation

	^ SystemNavigation default! !

Object subclass: #SecurityManager
	instanceVariableNames: 'privateKeyPair trustedKeys keysFileName securityChecksEnabled showSecurityStatus startInUntrustedDirectory warnAboutInsecureContent automaticKeyGeneration'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Support'!

!SecurityManager class methodsFor: 'accessing' stamp: 'nk 7/30/2004 21:50'!
default
	^Default ifNil: [Default := self new]! !

!SecurityManager class methodsFor: 'initialization' stamp: 'nk 7/30/2004 21:50'!
initialize
	"SecurityManager initialize"

	"Order: ExternalSettings, SecurityManager, AutoStart"

	Default := self new.
	Smalltalk addToStartUpList: self after: ExternalSettings.
	Smalltalk addToShutDownList: self! !

!SecurityManager class methodsFor: 'initialization' stamp: 'ar 2/6/2001 16:46'!
shutDown
	self default shutDown.! !

!SecurityManager class methodsFor: 'initialization' stamp: 'ar 2/6/2001 17:12'!
startUp
	self default startUp.! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:19'!
addTrustedKey: aPublicKey
	"Add a public key to the list of trusted keys"
	trustedKeys := (trustedKeys copyWithout: aPublicKey) copyWith: aPublicKey.! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/20/2009 00:07'!
automaticKeyGeneration
	^ automaticKeyGeneration ifNil: [automaticKeyGeneration := false]! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:14'!
canWriteImage
	"SecurityManager default canWriteImage"
	"Primitive. Return true if the right to write an image hasn't been revoked."
	<primitive: 'primitiveCanWriteImage' module: 'SecurityPlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'security operations' stamp: 'ar 2/6/2001 16:14'!
disableFileAccess
	"SecurityManager default disableFileAccess"
	"Primitive. Disable unlimited access to files.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableFileAccess' module: 'FilePlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: 'ar 2/6/2001 16:15'!
disableImageWrite
	"SecurityManager default disableImageWrite"
	"Primitive. Disable writing to an image file.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableImageWrite' module: 'SecurityPlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: 'ar 2/6/2001 16:15'!
disableSocketAccess
	"SecurityManage default disableSocketAccess"
	"Primitive. Disable access to sockets.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableSocketAccess' module: 'SocketPlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: 'AlainPlantec 12/17/2009 20:28'!
enterRestrictedMode
	"Some insecure contents was encountered. Close all doors and proceed."
	self isInRestrictedMode ifTrue:[^true].
	self securityChecksEnabled ifFalse:[^true]. "it's been your choice..."
	self warnAboutInsecureContent ifTrue:[
		( UIManager default chooseFrom: #('Load it anyways' 'Do not load it')
			title: 
'You are about to load some insecure content.
If you continue, access to files as well as
some other capabilities will be limited.')
			 = 1 ifFalse:[
				"user doesn't really want it"
				^false.
			].
	].
	"here goes the actual restriction"
	self flushSecurityKeys.
	self disableFileAccess.
	self disableImageWrite.
	"self disableSocketAccess."
	FileDirectory setDefaultDirectory: self untrustedUserDirectory.
	^true
! !

!SecurityManager methodsFor: 'security operations' stamp: 'sd 1/30/2004 15:22'!
fileInObjectAndCode: aStream
	| trusted |
	trusted := self positionToSecureContentsOf: aStream.
	trusted ifFalse:[self enterRestrictedMode ifFalse:[
		aStream close.
		^nil]].
	^aStream fileInObjectAndCode! !

!SecurityManager methodsFor: 'initialization' stamp: 'ar 2/6/2001 16:24'!
flushSecurityKey: aKey
	"Flush a security key"
	| n |
	n := aKey first.
	1 to: n basicSize do:[:i| n basicAt: i put: 0].
	n := aKey second.
	1 to: n basicSize do:[:i| n basicAt: i put: 0].
! !

!SecurityManager methodsFor: 'initialization' stamp: 'ar 2/6/2001 16:23'!
flushSecurityKeys
	"Flush all keys"
	privateKeyPair ifNotNil:[
		self flushSecurityKey: privateKeyPair first.
		self flushSecurityKey: privateKeyPair last.
	].
	privateKeyPair := nil.
	trustedKeys do:[:key| self flushSecurityKey: key].
	trustedKeys := #().! !

!SecurityManager methodsFor: 'private' stamp: 'adrian_lienhard 7/18/2009 16:01'!
generateKeyPairInBackground
	"SecurityManager default generateKeyPairInBackground"
	"Silently generate a key set on the local machine while running in the background."
	| guesstimate startTime |
	guesstimate := [ 10 benchmark ] timeToRun * 150.
	startTime := Time millisecondClockValue.
	privateKeyPair := nil.
	[ self generateLocalKeyPair ] fork.
	UIManager default informUserDuring: 
		[ :bar | 
		[ privateKeyPair == nil ] whileTrue: 
			[ bar value: 'Initializing security system (' , ((Time millisecondClockValue - startTime) * 100 // guesstimate) printString , '%)'.
			(Delay forSeconds: 1) wait ] ]! !

!SecurityManager methodsFor: 'private' stamp: 'simon.denier 6/11/2010 17:35'!
generateLocalKeyPair
	"SecurityManager default generateLocalKeyPair"
	"Generate a key set on the local machine."
	| dsa |
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandomFromString: 
		Time millisecondClockValue printString, 
		Date today printString, 
		OSPlatform platformName printString.
	privateKeyPair := dsa generateKeySet.
	self storeSecurityKeys.! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:16'!
hasFileAccess
	"SecurityManager default hasFileAccess"
	"Return true if the right to access arbitrary files hasn't been revoked"
	<primitive: 'primitiveHasFileAccess' module: 'FilePlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:16'!
hasSocketAccess
	"SecurityManager default hasSocketAccess"
	"Return true if the right to access sockets hasn't been revoked"
	<primitive: 'primitiveHasSocketAccess' module: 'SocketPlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'initialization' stamp: 'adrian_lienhard 7/18/2009 16:01'!
initialize
	super initialize.
	privateKeyPair := nil.
	trustedKeys := #().
	keysFileName := 'Pharo.keys'.! !

!SecurityManager methodsFor: 'testing' stamp: 'ar 2/6/2001 16:13'!
isInRestrictedMode
	"Return true if we're in restricted mode"
	^(self canWriteImage 
		or:[self hasFileAccess 
		"or:[self hasSocketAccess]"]) not! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:17'!
keysFileName
	^keysFileName! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:17'!
keysFileName: aFileName
	keysFileName := aFileName! !

!SecurityManager methodsFor: 'filein/out' stamp: 'AlexandreBergel 7/30/2008 13:41'!
loadSecurityKeys
	"SecurityManager default loadSecurityKeys"
	"Load the keys file for the current user"
	| fd loc file keys |
	self isInRestrictedMode ifTrue: [ ^ self ]. "no point in even trying"
	loc := self secureUserDirectory. "where to get it from"
	loc last = FileDirectory pathNameDelimiter ifFalse:[
		loc := loc copyWith: FileDirectory pathNameDelimiter.
	].

	fd := FileDirectory on: loc.
	file := [fd readOnlyFileNamed: keysFileName] 
				on: FileDoesNotExistException 
				do:[:ex| ^ self "no keys file"].
				
	keys := Object readFrom: file.
	privateKeyPair := keys first.
	trustedKeys := keys last.
	file close.! !

!SecurityManager methodsFor: 'security operations' stamp: 'RAA 3/2/2002 14:33'!
positionToSecureContentsOf: aStream
	| bytes trusted part1 part2 sig hash dsa okay pos |
	aStream binary.
	pos := aStream position.
	bytes := aStream next: 4.
	bytes = 'SPRJ' asByteArray ifFalse:[
		"was not signed"
		aStream position: pos.
		^false].
	part1 := (aStream nextInto: (LargePositiveInteger basicNew: 20)) normalize.
	part2 := (aStream nextInto: (LargePositiveInteger basicNew: 20)) normalize.
	sig := Array with: part1 with: part2.
	hash := SecureHashAlgorithm new hashStream: aStream.
	dsa := DigitalSignatureAlgorithm new.
	trusted := self trustedKeys.
	okay := (trusted detect:[:key| dsa verifySignature: sig ofMessageHash: hash publicKey: key]
			ifNone:[nil]) notNil.
	aStream position: pos+44.
	^okay! !

!SecurityManager methodsFor: 'accessing' stamp: 'tak 3/15/2005 00:45'!
primSecureUserDirectory
	<primitive: 'primitiveGetSecureUserDirectory' module: 'SecurityPlugin'>
	^ nil! !

!SecurityManager methodsFor: 'private' stamp: 'tak 12/17/2004 14:15'!
primUntrustedUserDirectory
	"Primitive. Return the untrusted user directory that is the root directory for files that are visible even in restricted mode."
	<primitive: 'primitiveGetUntrustedUserDirectory' module: 'SecurityPlugin'>
	^ nil! !

!SecurityManager methodsFor: 'private' stamp: 'mir 11/10/2003 16:14'!
printStateOn: stream
 	"Print the current state of myself onto stream.
 	Used to gather information in the debug log."
 
 	stream
 		nextPutAll: 'SecurityManager state:'; cr;
 		nextPutAll: 'Restricted: '; nextPutAll: self isInRestrictedMode asString; cr;
 		nextPutAll: 'FileAccess: '; nextPutAll: self hasFileAccess asString; cr;
 		nextPutAll: 'SocketAccess: '; nextPutAll: self hasSocketAccess asString; cr;
 		nextPutAll: 'Working Dir '; nextPutAll: FileDirectory default pathName asString; cr;
 		nextPutAll: 'Trusted Dir '; nextPutAll: self secureUserDirectory asString; cr;
 		nextPutAll: 'Untrusted Dir '; nextPutAll: self untrustedUserDirectory asString; cr;
 		cr! !

!SecurityManager methodsFor: 'accessing' stamp: 'tak 3/15/2005 00:46'!
secureUserDirectory
	"SecurityManager default secureUserDirectory"
	| dir |
	dir := self primSecureUserDirectory.
	^ dir
		ifNil: [FileDirectory default pathName]
		ifNotNil: [(FilePath pathName: dir isEncoded: true) asSqueakPathName]! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:25'!
securityChecksEnabled
	^ securityChecksEnabled ifNil: [securityChecksEnabled := false]! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:27'!
securityChecksEnabled: aBoolean
	securityChecksEnabled := aBoolean! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:25'!
showSecurityStatus
	^ showSecurityStatus ifNil: [showSecurityStatus := false]! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:27'!
showSecurityStatus: aBoolean
	showSecurityStatus := aBoolean! !

!SecurityManager methodsFor: 'initialization' stamp: 'ar 2/6/2001 16:22'!
shutDown
	"Flush existing keys"
	self flushSecurityKeys.! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:20'!
signingKey
	"Return the key used for signing projects"
	^privateKeyPair ifNotNil:[privateKeyPair first]! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:26'!
startInUntrustedDirectory
	^ startInUntrustedDirectory ifNil: [startInUntrustedDirectory := false]! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:27'!
startInUntrustedDirectory: aBoolean
	startInUntrustedDirectory := aBoolean! !

!SecurityManager methodsFor: 'initialization' stamp: 'AlainPlantec 12/20/2009 00:07'!
startUp
	"Attempt to load existing keys"
	self loadSecurityKeys.
	(privateKeyPair == nil 
		and:[self isInRestrictedMode not
		and:[self automaticKeyGeneration]]) ifTrue:[
			self generateKeyPairInBackground.
	].! !

!SecurityManager methodsFor: 'filein/out' stamp: 'sw 1/25/2002 12:41'!
storeSecurityKeys
	"Store the keys file for the current user"
	"SecurityManager default storeSecurityKeys"

	| fd loc file |
	self isInRestrictedMode ifTrue:[^self]. "no point in even trying"
	loc := self secureUserDirectory. "where to put it"
	loc last = FileDirectory pathNameDelimiter ifFalse:
		[loc := loc copyWith: FileDirectory pathNameDelimiter].
	fd := FileDirectory on: loc.
	fd assureExistence.
	fd deleteFileNamed: self keysFileName ifAbsent:[].
	file := fd newFileNamed: self keysFileName.
	{privateKeyPair. trustedKeys} storeOn: file.
	file close! !

!SecurityManager methodsFor: 'accessing' stamp: 'ar 2/6/2001 16:42'!
trustedKeys
	"Return an array of trusted public keys for verifying some project"
	privateKeyPair ifNil:[^trustedKeys].
	^{privateKeyPair second}, trustedKeys! !

!SecurityManager methodsFor: 'accessing' stamp: 'tak 12/17/2004 14:19'!
untrustedUserDirectory
	"SecurityManager default untrustedUserDirectory"
	| dir |
	dir := self primUntrustedUserDirectory.
	^ dir
		ifNil: [FileDirectory default pathName]
		ifNotNil: [(FilePath pathName: dir isEncoded: true) asSqueakPathName]! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:26'!
warnAboutInsecureContent
	^ warnAboutInsecureContent ifNil: [warnAboutInsecureContent := false]! !

!SecurityManager methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 20:26'!
warnAboutInsecureContent: aBoolean
	warnAboutInsecureContent := aBoolean! !

Object subclass: #SmalltalkImage
	instanceVariableNames: 'globals'
	classVariableNames: 'CheckChangesFileAvailability CheckSourcesFileAvailability EndianCache LastImageName LastQuitLogPosition LastStats LowSpaceProcess LowSpaceSemaphore MemoryHogs ShutDownList SourceFileVersionString SpecialSelectors StartUpList StartupStamp WordSize'
	poolDictionaries: ''
	category: 'System-Support'!

!SmalltalkImage commentStamp: 'StephaneDucasse 3/9/2010 16:02' prior: 0!
I represent the SmalltalkImage and partly the VM. Using my current instance you can 
	- get the value of some VM parameters, system arguments, vm profiling,
	endianess status, external objects,....

	- save the image, manage sources

As you will notice browsing my code I'm a fat class having still too much responsibility.
But this is life. sd-2 July 2003

PS: if someone wants to split me go ahead.

--------------------------------------------------------------------------------------------------------
We now have a common cleanUp protocol:

       MyClass class>>cleanUp: aggressive

This (class-side) method is called with an argument that indicates whether to do aggressive, i.e., potentially 
destructive cleanup. Doing aggressive cleanup will delete projects, change sets, uni-classes and more. Gentle cleanup is expected to only flush transient caches. If your class only has gentle cleanup you can simply implement

       MyClass class>>cleanUp

By default #cleanUp: delegates to this method (similar to startUp/startUp: and shutDown/shutDown:). You can
run cleanup by executing:

       Smalltalk cleanUp. "gently"
       Smalltalk cleanUp: true. "aggressive"!

!SmalltalkImage class methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 14:27'!
checkChangesFileAvailability
	^ CheckChangesFileAvailability ifNil: [CheckChangesFileAvailability := true]! !

!SmalltalkImage class methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 14:28'!
checkChangesFileAvailability: aBoolean
	CheckChangesFileAvailability := aBoolean! !

!SmalltalkImage class methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 14:28'!
checkSourcesFileAvailability
	^ CheckSourcesFileAvailability ifNil: [CheckSourcesFileAvailability := true]! !

!SmalltalkImage class methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 14:28'!
checkSourcesFileAvailability: aBoolean
	CheckSourcesFileAvailability := aBoolean! !

!SmalltalkImage class methodsFor: 'cleanup' stamp: 'StephaneDucasse 3/21/2010 15:25'!
cleanUp
	"Flush caches"
	
	"should probably distribute cleaning to systemDictionary"
	
	Smalltalk globals flushClassNameCache.
	Smalltalk cleanUpUndoCommands.
	Undeclared removeUnreferencedKeys! !

!SmalltalkImage class methodsFor: 'instance creation' stamp: 'MarcusDenker 4/1/2010 11:10'!
current
	^Smalltalk! !

!SmalltalkImage class methodsFor: 'initialization' stamp: 'StephaneDucasse 3/13/2010 19:43'!
howToGetClassesNotRegisteredInStartups

	| classes startClassNames stopClassNames |
	classes := ((SystemNavigation default allCallsOn:
#'addToStartUpList:') collect: [ :e | e methodClass ]),
((SystemNavigation default allCallsOn: #'addToStartUpList:after:')
collect: [ :e | e methodClass ]).
startClassNames := classes collect: [ :n | n instanceSide name ].
(Smalltalk class classVarNamed: 'StartUpList') do: [ :s |
	startClassNames remove: s ifAbsent: [] ].


classes := ((SystemNavigation default allCallsOn:
#'addToShutDownList:') collect: [ :e | e methodClass ]),
((SystemNavigation default allCallsOn: #'addToShutDownList:after:')
collect: [ :e | e methodClass ]).
stopClassNames := classes collect: [ :n | n instanceSide name ].
(Smalltalk class classVarNamed: 'ShutDownList') do: [ :s |
	stopClassNames remove: s ifAbsent: [] ].

((startClassNames, stopClassNames) asSet asSortedCollection remove: #InputEventSensor ; yourself)
	"do: [:each | (Smalltalk at: each) initialize] does not work because of dependencies"! !

!SmalltalkImage class methodsFor: 'initialization' stamp: 'StephaneDucasse 3/13/2010 19:49'!
initialize
	"SmalltalkImage initialize"

	self initializeStartUpList.
	self initializeShutDownList.
	Smalltalk addToStartUpList: SmalltalkImage! !

!SmalltalkImage class methodsFor: 'initialization' stamp: 'lr 3/14/2010 21:13'!
initializeShutDownList
	"SmalltalkImage initializeShutDownList"

	| oldList |
	oldList := ShutDownList.
	ShutDownList := OrderedCollection new.	"These get processed from the bottom up..."
	#(#Delay #DisplayScreen #InputEventFetcher #Form #PasteUpMorph #StrikeFont #Color #FileDirectory #SoundPlayer #HttpUrl #Password)
		do: [ :clsName | Smalltalk globals at: clsName ifPresent: [ :cls | Smalltalk addToShutDownList: cls ] ].
	oldList
		ifNotNil: [ 
			oldList
				reverseDo: [ :className | Smalltalk globals at: className ifPresent: [ :theClass | Smalltalk addToShutDownList: theClass ] ] ]! !

!SmalltalkImage class methodsFor: 'initialization' stamp: 'lr 3/14/2010 21:13'!
initializeStartUpList
	"SmalltalkImage initializeStartUpList"

	| oldList |
	oldList := StartUpList.
	StartUpList := OrderedCollection new.	"These get processed from the top down..."
	#(#Delay #DisplayScreen #Cursor #InputEventFetcher #ProcessorScheduler #LanguageEnvironment #FileDirectory #NaturalLanguageTranslator #ShortIntegerArray #ShortRunArray #CrLfFileStream)
		do: [ :clsName | Smalltalk globals at: clsName ifPresent: [ :cls | Smalltalk addToStartUpList: cls ] ].
	oldList
		ifNotNil: [ 
			oldList
				do: [ :className | Smalltalk globals at: className ifPresent: [ :theClass | Smalltalk addToStartUpList: theClass ] ] ].
	#(#PasteUpMorph) do: [ :clsName | Smalltalk globals at: clsName ifPresent: [ :cls | Smalltalk addToStartUpList: cls ] ]! !

!SmalltalkImage class methodsFor: 'instance creation' stamp: 'sd 9/30/2003 13:39'!
new

	self error: 'Use current'.! !

!SmalltalkImage class methodsFor: 'initialization' stamp: 'AdrianLienhard 9/1/2010 20:56'!
startUp
	SystemChangeNotifier uniqueInstance
		notify: Smalltalk
		ofAllSystemChangesUsing: #event:! !

!SmalltalkImage class methodsFor: 'initialization' stamp: 'lr 3/14/2010 21:13'!
temporaryFixStartUpList
	"SmalltalkImage temporaryFixStartUpList"

	#(#CPUWatcher #Clipboard #CommandHistory #ExternalSettings #FT2Handle #FreeTypeCache #SecurityManager #FreeTypeFontProvider #FreeTypeSettings #SmalltalkImage #HostSystemMenus #HostWindowProxy #InternetConfiguration #Locale #LogicalFont #MCMethodDefinition #MenuIcons #MultiByteFileStream #OSPlatform #ProcessBrowser #AutoStart #Symbol #SystemDictionary #TTCFont #UITheme #UUIDGenerator #WeakArray)
		do: [ :each | (Smalltalk globals at: each) initialize ]! !

!SmalltalkImage class methodsFor: 'initialize-release' stamp: 'StephaneDucasse 3/17/2010 21:29'!
wordSize
	^ WordSize ifNil: [WordSize := [Smalltalk vmParameterAt: 40] on: Error do: [4]]! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'nice 1/16/2010 12:57'!
abandonSources
	"Smalltalk abandonSources"
	"Replaces every method by a copy with the 4-byte source pointer 
	 replaced by a string of all arg and temp names, followed by its
	 length. These names can then be used to inform the decompiler."
	"wod 11/3/1998: zap the organization before rather than after
	 condensing changes."
	"eem 7/1/2009 13:59 update for the closure schematic temp names regime"
	| oldMethods newMethods bTotal bCount |
	(self confirm: 'This method will preserve most temp names
(up to about 15k characters of temporaries)
while allowing the sources file to be discarded.
-- CAUTION --
If you have backed up your system and
are prepared to face the consequences of
abandoning source code files, choose Yes.
If you have any doubts, you may choose No
to back out with no harm done.')
			== true
		ifFalse: [^ self inform: 'Okay - no harm done'].
	oldMethods := OrderedCollection new: CompiledMethod instanceCount.
	newMethods := OrderedCollection new: CompiledMethod instanceCount.
	bTotal := 0.
	bCount := 0.
	self systemNavigation allBehaviorsDo: [:b | bTotal := bTotal + 1].
	'Saving temp names for better decompilation...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: bTotal
		during:
			[:bar |
			self systemNavigation allBehaviorsDo:
				[:cl |  "for test: (Array with: Arc with: Arc class) do:"
				bar value: (bCount := bCount + 1).
				cl selectorsAndMethodsDo:
					[:selector :m |
					| oldCodeString methodNode |
					m fileIndex > 0 ifTrue:
						[oldCodeString := cl sourceCodeAt: selector.
						methodNode := cl compilerClass new
											parse: oldCodeString
											in: cl
											notifying: nil.
						oldMethods addLast: m.
						newMethods addLast: (m copyWithTempsFromMethodNode: methodNode)]]]].
	oldMethods asArray elementsExchangeIdentityWith: newMethods asArray.
	self systemNavigation allBehaviorsDo: [:b | b zapOrganization].
	self condenseChanges.
	SmalltalkImage checkSourcesFileAvailability: false! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'StephaneDucasse 3/17/2010 21:21'!
abandonTempNames
	"Replaces every method by a copy with no source pointer or
	encoded temp names."
	"Smalltalk abandonTempNames"
	| continue oldMethods newMethods n |
	continue := self confirm: '-- CAUTION --
If you have backed up your system and
are prepared to face the consequences of
abandoning all source code, hit Yes.
If you have any doubts, hit No,
to back out with no harm done.'.
	continue
		ifFalse: [^ self inform: 'Okay - no harm done'].
	self garbageCollect.
	oldMethods := OrderedCollection new.
	newMethods := OrderedCollection new.
	n := 0.
	'Removing temp names to save space...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: [:bar | self systemNavigation
				allBehaviorsDo: [:cl | cl selectorsAndMethodsDo: [:sel :m | 
							bar value: (n := n + 1).
							oldMethods addLast: m.
							newMethods
								addLast: (m copyWithTrailerBytes: #(0 ))]]].
	oldMethods asArray elementsExchangeIdentityWith: newMethods asArray.
	Smalltalk closeSourceFiles.
	self flag: #shouldUseAEnsureBlockToBeSureThatTheFileIsClosed.
	"sd: 17 April 2003"
	SmalltalkImage checkChangesFileAvailability: false.
	SmalltalkImage checkSourcesFileAvailability: false! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: 'StephaneDucasse 3/13/2010 15:10'!
add: aClass toList: startUpOrShutDownList after: predecessor
	"Add the name of aClass to the startUp or shutDown list.
	Add it after the name of predecessor, or at the end if predecessor is nil."

	| name earlierName |
	name := aClass name.
	(self at: name ifAbsent: [nil]) == aClass ifFalse:
		[self error: name , ' cannot be found in Smalltalk dictionary.'].
	predecessor == nil
		ifTrue: ["No-op if alredy in the list."
				(startUpOrShutDownList includes: name) ifFalse:
					[startUpOrShutDownList == StartUpList
						ifTrue: ["Add to end of startUp list"
								startUpOrShutDownList addLast: name]
						ifFalse: ["Add to front of shutDown list"
								startUpOrShutDownList addFirst: name]]]
		ifFalse: ["Add after predecessor, moving it if already there."
				earlierName := predecessor name.
				(self at: earlierName) == predecessor ifFalse:
					[self error: earlierName , ' cannot be found in Smalltalk dictionary.'].
				(startUpOrShutDownList includes: earlierName) ifFalse:
					[self error: earlierName , ' cannot be found in the list.'].
				startUpOrShutDownList remove: name ifAbsent:[].
				startUpOrShutDownList add: name after: earlierName]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/4/1999 15:38'!
addToShutDownList: aClass
	"This will add a ref to this class at the BEGINNING of the shutDown list."

	self addToShutDownList: aClass after: nil! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:04'!
addToShutDownList: aClass after: predecessor

	self add: aClass toList: ShutDownList after: predecessor! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/4/1999 15:37'!
addToStartUpList: aClass
	"This will add a ref to this class at the END of the startUp list."

	self addToStartUpList: aClass after: nil! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:04'!
addToStartUpList: aClass after: predecessor

	self add: aClass toList: StartUpList after: predecessor! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/20/2010 13:58'!
allClasses  
	"Return all the class defines in the Smalltalk SmalltalkImage"
	"Smalltalk allClasses"
	"May be we could discard this one in the future"
	^ self classNames collect: [:name | self at: name]! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'al 1/12/2006 23:59'!
allClassesAndTraits
	"Return all the classes and traits defined in the Smalltalk SmalltalkImage"

	^ self classNames , self traitNames collect: [:each | self at: each]! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'al 1/13/2006 00:15'!
allClassesAndTraitsDo: aBlock
	^self allClassesAndTraits do: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'sd 4/17/2003 21:18'!
allClassesDo: aBlock
	"Evaluate the argument, aBlock, for each class in the system."

	(self classNames collect: [:name | self at: name]) do: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'al 2/23/2006 21:39'!
allTraits
	"Return all traits defined in the Smalltalk SmalltalkImage"

	^ self traitNames collect: [:each | self at: each]! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 6/30/2010 22:51'!
argumentAt: i
	"Answer the i-th argument of the command line, or nil if not so many argument."
	
	^self getSystemAttribute: 2 + i! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'GermanArduino 8/2/2010 18:58'!
arguments
	"Answer an array with all the command line arguments.
	This does not include imagePath, documentPath nor any option."
	
	"Smalltalk commandLine arguments"
	
	^Array streamContents: [:str |
		| arg i |
		i := 1.
		[i > 998 or: [(arg := self argumentAt: i) == nil]]
			whileFalse:
				[str nextPut: arg.
				i := i + 1]].! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:54'!
associationAt: aKey
	"DO NOT DEPRECATE - used by ImageSegments"

	^globals associationAt: aKey! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:56'!
associationAt: key ifAbsent: aBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals associationAt: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 17:14'!
associationDeclareAt: aKey

	^ globals associationDeclareAt: aKey! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 17:17'!
associationOrUndeclaredAt: key 
	"return an association or install in undeclared.  Used for mating up ImageSegments."

	^ globals associationOrUndeclaredAt: key! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'NS 1/16/2004 15:34'!
assureStartupStampLogged
	"If there is a startup stamp not yet actually logged to disk, do it now."
	| changesFile |
	StartupStamp ifNil: [^ self].
	(SourceFiles isNil or: [(changesFile := SourceFiles at: 2) == nil]) ifTrue: [^ self].
	changesFile isReadOnly ifTrue:[^self].
	changesFile setToEnd; cr; cr.
	changesFile nextChunkPut: StartupStamp asString; cr.
	StartupStamp := nil.
	self forceChangesToDisk.! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:42'!
at: aKey
	"Answer the global associated with the key."

	^globals at: aKey! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:42'!
at: key ifAbsent: aBlock
	"Answer the global associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	^globals at: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:42'!
at: aKey ifAbsentPut: aBlock 
	"Return the global at the given key. 
	If key is not included in the receiver store the result 
	of evaluating aBlock as new value."

	^globals at: aKey ifAbsentPut: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:42'!
at: key ifPresent: aBlock
	"Lookup the given key in the globals. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	^globals at: key ifPresent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:42'!
at: aKey put: anObject 
	"Set the global at key to be anObject.  If key is not found, create a
	new entry for key and set is value to anObject. Answer anObject."

	^globals at: aKey put: anObject! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 17:19'!
bindingOf: varName
	"Answer the binding of some variable resolved in the scope of the receiver"

	^globals bindingOf: varName! !

!SmalltalkImage methodsFor: 'vm' stamp: 'StephaneDucasse 6/30/2010 22:50'!
buildDate			
	"Return a String reflecting the build date of the VM"
	"Smalltalk buildDate"

	^self getSystemAttribute: 1006! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
bytesLeft
	"Answer the number of bytes of space available. Does a full garbage collection."

	^ self garbageCollect
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'VeronicaUquillas 6/11/2010 13:01'!
bytesLeft: aBool
	"Return the amount of available space. If aBool is true, include possibly available swap space. If aBool is false, include possibly available physical memory. For a report on the largest free block currently availabe within Pharo memory but not counting extra memory use #primBytesLeft."
	<primitive: 112>
	^self primBytesLeft! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 2/25/2001 18:00'!
bytesLeftString
	"Return a string describing the amount of memory available"
	| availInternal availPhysical availTotal |
	self garbageCollect.
	availInternal := self primBytesLeft.
	availPhysical := self bytesLeft: false.
	availTotal := self bytesLeft: true.
	(availTotal > (availInternal + 10000)) "compensate for mini allocations inbetween"
		ifFalse:[^availInternal asStringWithCommas, ' bytes available'].
	^String streamContents:[:s|
		s nextPutAll: availInternal asStringWithCommas, 	' bytes (internal) '; cr.
		s nextPutAll: availPhysical asStringWithCommas,	' bytes (physical) '; cr.
		s nextPutAll: availTotal asStringWithCommas, 	' bytes (total)     '].! !

!SmalltalkImage methodsFor: 'endian' stamp: 'StephaneDucasse 2/3/2010 22:19'!
calcEndianness
	| bytes word blt |
	"What endian-ness is the current hardware?  The String '1234' will be stored into a machine word.  On BigEndian machines (the Mac), $1 will be the high byte if the word.  On LittleEndian machines (the PC), $4 will be the high byte."
	"SmalltalkImage current endianness"

	bytes := #[0 0 0 0].  "(1 2 3 4) or (4 3 2 1)"
	word := WordArray with: 16r01020304.
	blt := (BitBlt toForm: (Form new hackBits: bytes)) 
				sourceForm: (Form new hackBits: word).
	blt combinationRule: Form over.  "store"
	blt sourceY: 0; destY: 0; height: 1; width: 4.
	blt sourceX: 0; destX: 0.
	blt copyBits.  "paste the word into the bytes"
	bytes first = 1 ifTrue: [^ #big].
	bytes first = 4 ifTrue: [^ #little].
	self error: 'Ted is confused'.! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'yo 7/2/2004 13:24'!
changeImageNameTo: aString
	
	self imageName: aString asSqueakPathName.
	LastImageName := self imageName! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'bf 1/12/2006 19:43'!
changesName
	"Answer the name for the changes file corresponding to the image file name."
	"Smalltalk changesName"

	| imName |
	imName := FileDirectory baseNameFor: self imageName.
	^ imName, FileDirectory dot, 'changes'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'simon.denier 6/11/2010 17:34'!
checkAndOpenSourcesAndChanges
	"Open the changes and sources files and install them in SourceFiles. 
	Inform the user of problems regarding write permissions or CR/CRLF mixups."
	"Note: SourcesName and imageName are full paths; 
	changesName is a  local name."
	| sources changes msg wmsg |
	msg := 'Pharo cannot locate &fileRef.

Please check that the file is named properly and is in the
same directory as this image.'.
	wmsg := 'Pharo cannot write to &fileRef.

Please check that you have write permission for this file.

You won''t be able to save this image correctly until you fix this.'.

	sources := FileDirectory openSources: self sourcesName forImage: LastImageName.
	changes := FileDirectory openChanges: self changesName forImage: LastImageName.

	((sources == nil or: [sources atEnd])
			and: [self checkSourcesFileAvailability])
		ifTrue: [OSPlatform isMacOS
				ifTrue: [msg := msg , '
Make sure the sources file is not an Alias.'].
self inform: (msg copyReplaceAll: '&fileRef' with: 'the sources file named ' , self sourcesName)].

	(changes == nil
			and: [self checkChangesFileAvailability])
		ifTrue: [self inform: (msg copyReplaceAll: '&fileRef' with: 'the changes file named ' , self sourcesName)].

	(self checkChangesFileAvailability and: [changes notNil])
		ifTrue: [changes isReadOnly
				ifTrue: [self inform: (wmsg copyReplaceAll: '&fileRef' with: 'the changes file named ' , self changesName)].

			((changes next: 200)
					includesSubString: String crlf)
				ifTrue: [self inform: 'The changes file named ' , self changesName , '
has been injured by an unpacking utility.  Crs were changed to CrLfs.
Please set the preferences in your decompressing program to 
"do not convert text files" and unpack the system again.']].

	SourceFiles := Array with: sources with: changes! !

!SmalltalkImage methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 14:29'!
checkChangesFileAvailability
	^ self class checkChangesFileAvailability! !

!SmalltalkImage methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 14:30'!
checkSourcesFileAvailability
	^ self class checkSourcesFileAvailability! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 8/19/2010 22:25'!
classNamed: className 
	^self globals classOrTraitNamed: className.! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'MarcusDenker 3/18/2010 19:24'!
classNames
	^globals classNames! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'lr 3/14/2010 21:13'!
classOrTraitNamed: aString
	"aString is either a class or trait name or a class or trait name followed by ' class' or 'classTrait' respectively.
	Answer the class or metaclass it names."

	| meta baseName baseClass |
	(aString endsWith: ' class')
		ifTrue: [ 
			meta := true.
			baseName := aString copyFrom: 1 to: aString size - 6 ]
		ifFalse: [ 
			(aString endsWith: ' classTrait')
				ifTrue: [ 
					meta := true.
					baseName := aString copyFrom: 1 to: aString size - 11 ]
				ifFalse: [ 
					meta := false.
					baseName := aString ] ].
	baseClass := Smalltalk globals at: baseName asSymbol ifAbsent: [ ^ nil ].
	meta
		ifTrue: [ ^ baseClass classSide ]
		ifFalse: [ ^ baseClass ]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
cleanOutUndeclared 
	Undeclared removeUnreferencedKeys! !

!SmalltalkImage methodsFor: 'cleanup' stamp: 'StephaneDucasse 3/9/2010 16:07'!
cleanUp	
	"Smalltalk cleanUp"
	"Gently clean up"
	
	^self cleanUp: false! !

!SmalltalkImage methodsFor: 'cleanup' stamp: 'StephaneDucasse 3/9/2010 16:05'!
cleanUp: aggressive
	"Clean up. When aggressive is true, this will destroy change sets, etc."
	"Smalltalk cleanUp: false"
	"Smalltalk cleanUp: true"

	^self cleanUp: aggressive except: #()! !

!SmalltalkImage methodsFor: 'cleanup' stamp: 'StephaneDucasse 8/10/2010 21:59'!
cleanUp: aggressive except: exclusions
	"Clean up. When aggressive is true, this will destroy resources, change sets, etc.
	Leave out any classes specifically listed in exclusions."

	"Smalltalk cleanUp: true except: {Point . ChangeSet}"

	^self cleanUp: aggressive except: exclusions confirming: aggressive! !

!SmalltalkImage methodsFor: 'cleanup' stamp: 'StephaneDucasse 8/10/2010 22:01'!
cleanUp: aggressive except: exclusions confirming: aBool
	"Clean up. When aggressive is true, this will destroy resources, change sets, etc.
	Leave out any classes specifically listed in exclusions."

	"Smalltalk cleanUp: true except: {Point . ChangeSet} 
		- will ask for confirmation"
	"SmalltalkImage current cleanUp: true except: {} confirming: false 
		- will not ask for confirmation"

	| classes |
	aBool ifTrue:[
		"Give the user a chance to bail"
		(self confirm: 'Cleanup will destroy projects, change sets and more.
Are you sure you want to proceed?') ifFalse: [^self]].

	"Find all classes implementing #cleanUp or cleanUp:"
	classes := Smalltalk allClasses select: [:aClass| 
		(aClass class includesSelector: #cleanUp) 
			or: [aClass class includesSelector: #cleanUp:]].

	"Leave out the classes in the exclusion set"
	classes := classes reject: [:aClass| exclusions includes: aClass].

	"Arrange classes in superclass order, superclasses before subclasses.
	This will ensure that specific cleanup (like MethodDictionary compaction)
	will run after generic superclass cleanup (HashedCollection rehashing).
	Otherwise generic superclass cleanup might undo specific one (in this
	case rehashing will undo a good bit of MD compaction)."
	classes := Class superclassOrder: classes.

	"Run the cleanup code"
	classes 
		do:[:aClass| aClass cleanUp: aggressive]
		displayingProgress: [:aClass| 'Cleaning up in ', aClass name].! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'nice 1/5/2010 15:59'!
cleanUpUndoCommands
	"Smalltalk cleanUpUndoCommands"  "<== print this to get classes involved"

	| classes |
	classes := Bag new.
	'Ferreting out obsolete undo commands'
		displayProgressAt: Sensor cursorPoint
		from: 0 to: Morph withAllSubclasses size
		during:
	[:bar | | i | i := 0.
	Morph withAllSubclassesDo:
		[:c | bar value: (i := i+1).
		c allInstancesDo:
			[:m | | p | (p := m otherProperties) ifNotNil:
				[p keys do:
					[:k | (p at: k) class == Command ifTrue:
						[classes add: c name.
						m removeProperty: k]]]]]].
	^ classes! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'JMM 6/6/2000 20:36'!
clearExternalObjects
	"Clear the array of objects that have been registered for use in non-Smalltalk code."
	"Smalltalk clearExternalObjects"

	ExternalSemaphoreTable clearExternalObjects
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'sd 11/16/2003 13:12'!
closeSourceFiles
	"Shut down the source files if appropriate.  1/29/96 sw: changed so that the closing and nilification only take place if the entry was a FileStream, thus allowing stringified sources to remain in the saved image file"

	1 to: 2 do: [:i |
		((SourceFiles at: i) isKindOf: FileStream)
			ifTrue:
				[(SourceFiles at: i) close.
				SourceFiles at: i put: nil]]! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'GermanArduino 8/2/2010 19:02'!
commandLine
	"Answer the object to query about command line."
	
	^self! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'sd 9/29/2004 18:30'!
compactClassesArray
	"Smalltalk compactClassesArray"
	"Return the array of 31 classes whose instances may be
	represented compactly"
	^ self specialObjectsArray at: 29! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'StephaneDucasse 5/9/2010 18:29'!
condenseChanges
	"Move all the changes onto a compacted sources file."
	"Smalltalk condenseChanges"
	| f oldChanges |
	f := FileStream fileNamed: 'ST80.temp'.
	f header; timeStamp.
	'Condensing Changes File...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: self classNames size + self traitNames size
		during: [:bar | | count | 
			count := 0.
			self
				allClassesAndTraitsDo: [:classOrTrait | 
					bar value: (count := count + 1).
					classOrTrait moveChangesTo: f.
					classOrTrait putClassCommentToCondensedChangesFile: f.
					classOrTrait classSide moveChangesTo: f]].
	self lastQuitLogPosition: f position.
	f trailer; close.
	CompiledMethod cleanUp.
	
	oldChanges := SourceFiles at: 2.
	oldChanges close.
	FileDirectory default deleteFileNamed: oldChanges name , '.old';
		 rename: oldChanges name toBe: oldChanges name , '.old';
		 rename: f name toBe: oldChanges name.
	self setMacFileInfoOn: oldChanges name.
	SourceFiles
		at: 2
		put: (FileStream oldFileNamed: oldChanges name)! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'StephaneDucasse 5/9/2010 18:29'!
condenseSources	
	"Move all the changes onto a compacted sources file."
	"Smalltalk condenseSources"

	| f dir newVersionString |
	self fixUpProblemsWithAllCategory.
	"The above removes any concrete, spurious '-- all --' categories, which mess up the process."
	dir := FileDirectory default.
	newVersionString := UIManager default request: 'Please designate the version
for the new source code file...' initialAnswer: Smalltalk  sourceFileVersionString.
	newVersionString ifNil: [^ self].
	newVersionString = self sourceFileVersionString ifTrue:
		[^ self error: 'The new source file must not be the same as the old.'].
	Smalltalk sourceFileVersionString: newVersionString.

	"Write all sources with fileIndex 1"
	f := FileStream newFileNamed: Smalltalk sourcesName.
	f header; timeStamp.
'Condensing Sources File...'
	displayProgressAt: Sensor cursorPoint
	from: 0 to: self classNames size + self traitNames size
	during:
		[:bar | | count | count := 0.
		Smalltalk allClassesAndTraitsDo:
			[:classOrTrait | bar value: (count := count + 1).
			classOrTrait fileOutOn: f moveSource: true toFile: 1]].
	f trailer; close.
	
	CompiledMethod cleanUp.
	
	"Make a new empty changes file"
	self closeSourceFiles.
	dir rename: Smalltalk changesName
		toBe: Smalltalk changesName , '.old'.
	(FileStream newFileNamed: Smalltalk changesName)
		header; timeStamp; close.
	self lastQuitLogPosition: 0.

	self setMacFileInfoOn: SmalltalkImage current changesName.
	self setMacFileInfoOn: SmalltalkImage current sourcesName.
	self openSourceFiles.
	self inform: 'Source files have been rewritten!!
Check that all is well,
and then save/quit.'! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
createStackOverflow
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; createStackOverflow"

	self createStackOverflow.  "infinite recursion"! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'nk 7/29/2004 10:03'!
datedVersion
	"Answer the version of this release."

	^SystemVersion current datedVersion! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:56'!
do: aBlock
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals do: aBlock
	
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 6/30/2010 22:50'!
documentPath
	"Answer the absolute path of the document passed to the vm or nil if none."
	"Smalltalk commandLine documentPath"
	
	^self getSystemAttribute: 2! !

!SmalltalkImage methodsFor: 'endian' stamp: 'yo 2/18/2004 18:24'!
endianness

	EndianCache ifNil: [EndianCache := self calcEndianness].
	^ EndianCache.
! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'StephaneDucasse 3/13/2010 15:43'!
environment
	"For conversion from Smalltalk to SystemDictionary"

	^globals! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'StephaneDucasse 8/19/2010 22:20'!
event: anEvent
	"Hook for SystemChangeNotifier"

	(anEvent isRemoved and: [anEvent itemKind = SystemChangeNotifier classKind]) ifTrue: [
		anEvent item acceptsLoggingOfCompilation 
			ifTrue: [self logChange: 'Smalltalk globals removeClassNamed: #' , anEvent item name].
	].
	anEvent isDoIt 
		ifTrue: [self logChange: anEvent item].
	(anEvent isRemoved and: [anEvent itemKind = SystemChangeNotifier methodKind]) ifTrue: [
		anEvent itemClass acceptsLoggingOfCompilation 
			ifTrue: [self logChange: anEvent itemClass name , ' removeSelector: #' , anEvent itemSelector]].! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: ''!
exitToDebugger
	"Primitive. Enter the machine language debugger, if one exists. Essential.
	See Object documentation whatIsAPrimitive."

	<primitive: 114>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'JMM 6/6/2000 21:01'!
externalObjects
	"Return an array of objects that have been registered for use in non-Smalltalk code. Smalltalk objects should be referrenced by external code only via indirection through this array, thus allowing the objects to move during compaction. This array can be cleared when the VM re-starts, since variables in external code do not survive snapshots. Note that external code should not attempt to access a Smalltalk object, even via this mechanism, while garbage collection is in progress."
	"Smalltalk externalObjects"

	^ ExternalSemaphoreTable externalObjects
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'StephaneDucasse 3/21/2010 15:33'!
externalizeSources   
	"Write the sources and changes streams onto external files."
 	"Smalltalk externalizeSources"
	"the logic of this method is complex because it uses changesName and self changesName
	may be this is normal - sd"
	
	| sourcesName changesName aFile |
	sourcesName := self sourcesName.
	(FileDirectory default fileExists: sourcesName)
		ifTrue: [^ self inform:
'Sorry, you must first move or remove the
file named ', sourcesName].
	changesName := self changesName.
	(FileDirectory default fileExists: changesName)
		ifTrue: [^ self inform:
'Sorry, you must first move or remove the
file named ', changesName].

	aFile :=  FileStream newFileNamed: sourcesName.
	aFile nextPutAll: SourceFiles first originalContents.
	aFile close.
	self setMacFileInfoOn: sourcesName.
	SourceFiles at: 1 put: (FileStream readOnlyFileNamed: sourcesName).

	aFile := FileStream newFileNamed: self changesName.
	aFile nextPutAll: SourceFiles last contents.
	aFile close.
	"On Mac, set the file type and creator (noop on other platforms)"
	FileDirectory default
		setMacFileNamed: Smalltalk changesName
		type: 'STch'
		creator: 'FAST'.
	SourceFiles at: 2 put: (FileStream oldFileNamed: changesName).

	self inform: 'Sources successfully externalized'.
! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'sd 6/27/2003 23:27'!
extraVMMemory
	"Answer the current setting of the 'extraVMMemory' VM parameter. See the comment in extraVMMemory: for details."

	^ self vmParameterAt: 23
! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'sd 6/27/2003 23:27'!
extraVMMemory: extraBytesToReserve
	"Request that the given amount of extra memory be reserved for use by the virtual machine to leave extra C heap space available for things like plugins, network and file buffers, and so on. This request is stored when the image is saved and honored when the image is next started up. Answer the previous value of this parameter."

	extraBytesToReserve < 0
		ifTrue: [self error: 'VM memory reservation must be non-negative'].
	^ self vmParameterAt: 23 put: extraBytesToReserve
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'AdrienBarreau 6/29/2010 16:22'!
extractMinusParameters
	"Returns a dictionary matching parameters beginning by a minus character and the list of the next parameters.
	Here is an example: 
		I suppose I run pharo with: `pharo myImage.image -minus1 'Hello' -minus2 -minus3 'Hello' 'World'`.
		This message will return a citionary with: 
			minus1 --> #(Hello)
			minus2 --> #()
			minus3 --> #('Hello' 'World')"
			
	| args dict str previous previousWithMinus index |
	
	dict := Dictionary new.
	args := OrderedCollection new.
	previous := previousWithMinus := ''.

	str := self getSystemAttribute: (index := 2).
	[ str isNil ] whileFalse: [ args add: str. str := self getSystemAttribute: (index := index + 1) ].

	args ifEmpty: [ ^ dict ].

	previous := previousWithMinus := args detect: [ :each | each first = $- ].
	[ (args first = previous) & (args isEmpty not) ] whileFalse: [ args removeFirst ].
	
	args do: [ :each |
		(each first = $-)
			ifTrue: [ 
				(previous first = $-) ifTrue: [ dict at: previous allButFirst put: #() ].
				previousWithMinus := each ]
			ifFalse: [ dict at: previousWithMinus allButFirst put: ((dict at: previousWithMinus allButFirst ifAbsent: [ #() ]), (Array with: each)) ].
				
		previous := each.
	].

	(previous first = $-) ifTrue: [ dict at: previous allButFirst put: #() ].

	^ dict! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'MarcusDenker 4/1/2010 11:02'!
extractParameters

	| pName value index dict |
	dict := Dictionary new.
	index := 3. "Muss bei 3 starten, da 2 documentName ist"
	[pName := self  getSystemAttribute: index.
	pName isEmptyOrNil] whileFalse:[
		index := index + 1.
		value := self getSystemAttribute: index.
		value ifNil: [value := ''].
 		dict at: pName asUppercase put: value.
		index := index + 1].
	^dict! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'StephaneDucasse 4/27/2010 11:54'!
fixObsoleteReferences
	"SmalltalkImage current fixObsoleteReferences.
	SystemNavigation default obsoleteBehaviors size > 0
		ifTrue: [ SystemNavigation default obsoleteBehaviors inspect.
			self error:'Still have obsolete behaviors. See inspector']"

	| informee obsoleteBindings obsName realName realClass |
	Smalltalk
		garbageCollect;
		garbageCollect.
	CompiledMethod allInstances
		do: [ :method | 
			obsoleteBindings := method literals
				select: [ :literal | literal isVariableBinding 
									and: [ literal value isBehavior  and: [ literal value isObsolete ]]].
			obsoleteBindings
				do: [ :binding | 
					obsName := binding value name.
					Transcript
						show: 'Binding: ';
						show: obsName;
						cr.
					realName := obsName copyReplaceAll: 'AnObsolete' with: ''.
					realClass := Smalltalk globals at: realName asSymbol ifAbsent: [ UndefinedObject ].
					binding key: binding key value: realClass ] ].
	Behavior flushObsoleteSubclasses.
	Smalltalk
		garbageCollect;
		garbageCollect! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'nice 1/5/2010 15:59'!
fixUpProblemsWithAllCategory
	"Moves all methods that are in formally classified a category named '-- all --' into the default 'as yet unclassified' category"

	"Smalltalk fixUpProblemsWithAllCategory"

	| methodCount classCount |
	self flag: #ShouldBeMovedInClassOrganization.
	methodCount := 0.
	classCount := 0.
	self systemNavigation allBehaviorsDo:
		[:aClass | | org any | org := aClass organization.
			(org categories includes: #'-- all --') ifTrue:
				[any := false.
				aClass selectorsDo:
					[:aSelector | | aCategory |
						aCategory := org categoryOfElement: aSelector.
						aCategory = #'-- all --' ifTrue:
							[org classify: aSelector under: ClassOrganizer default suppressIfDefault: false.
							Transcript cr; show: aClass name, ' >> ', aSelector.
							methodCount := methodCount + 1.
							any := true]].
			any ifTrue: [classCount := classCount + 1].
			org removeEmptyCategories]].
	Transcript cr; show: methodCount printString, ' methods in ', classCount printString, ' classes moved from "-- all --" to "as yet unclassified"'
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 8/8/2010 18:17'!
flagInterpretedMethods: aBoolean
	"The Cog VM can be instructed to set the flag bit of CompiledMethods that it executes but will only interpret.  This can be used e.g. to profile startup. See CompiledMethod>>#flag & CompiledMethod>>#clearFlag. This flag persists across snapshots, stored in the image header."
	
	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 2) + (aBoolean ifTrue: [2] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 7/6/2010 21:41'!
flushClassNameCache
	"Smalltalk flushClassNameCache"
	"Force recomputation of the cached list of class names."

	globals flushClassNameCache.! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'ar 2/6/2001 18:42'!
forceChangesToDisk
	"Ensure that the changes file has been fully written to disk by closing and re-opening it. This makes the system more robust in the face of a power failure or hard-reboot."

	| changesFile |
	changesFile := SourceFiles at: 2.
	(changesFile isKindOf: FileStream) ifTrue: [
		changesFile flush.
		SecurityManager default hasFileAccess ifTrue:[
			changesFile close.
			changesFile open: changesFile name forWrite: true].
		changesFile setToEnd.
	].
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'JMM 1/27/2005 13:23'!
forceTenure
	"Primitive. Tell the GC logic to force a tenure on the next increment GC."
	<primitive: 'primitiveForceTenure'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 16:39'!
forgetClass: aClass logged: aBool 
	"Delete the class, aClass, from the system."

	^globals forgetClass: aClass logged: aBool! !

!SmalltalkImage methodsFor: 'modules' stamp: 'dvf 6/30/2005 12:11'!
forgetModule: aString
	"Primitive. If the module named aString is loaded, unloaded. If not, and it is marked an unloadable, unmark it so the VM will try to load it again next time. See comment for #unloadModule:."
	<primitive: 571>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'nice 5/11/2009 23:30'!
fullNameForChangesNamed: aName
	| imgName |
	imgName := self fullNameForImageNamed: aName.
	^FileDirectory fileName: (FileDirectory baseNameFor: imgName) extension: FileDirectory changeSuffix.! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'nice 5/11/2009 23:33'!
fullNameForImageNamed: aName
	| imgDir |
	imgDir := FileDirectory on: self imagePath.
	^FileDirectory fileName: (imgDir fullNameFor: aName) extension: FileDirectory imageSuffix.! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 2/11/2001 02:36'!
garbageCollect
	"Primitive. Reclaims all garbage and answers the number of bytes of available space."
	Object flushDependents.
	Object flushEvents.
	^self primitiveGarbageCollect! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'StephaneDucasse 3/21/2010 12:11'!
garbageCollectAndReport
	"Do a garbage collection, and report results to the user."

	"SmalltalkImage current garbageCollectAndReport"

	| reportString |
	reportString := String
		streamContents: [ :aStream | 
			aStream nextPutAll: self bytesLeftString.
			Smalltalk globals
				at: #Command
				ifPresent: [ :cmdClass | 
					| cc |
					(cc := cmdClass instanceCount) > 0
						ifTrue: [ 
							aStream
								cr;
								nextPutAll:
										'(note: there are ' , cc printString
												,
													' undo record(s) present in your
system; purging them may free up more space.)' ] ] ].
	self inform: reportString! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
garbageCollectMost
	"Primitive. Reclaims recently created garbage (which is usually most of it) fairly quickly and answers the number of bytes of available space."

	<primitive: 131>
	^ self primBytesLeft! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'DamienCassou 9/29/2009 13:11'!
getFileNameFromUser

	| newName |
	newName := UIManager default
		request: 'New File Name?' translated
		initialAnswer: (FileDirectory localNameFor: self imageName).
	newName isEmptyOrNil ifTrue: [^nil].
	((FileDirectory default fileOrDirectoryExists: (self fullNameForImageNamed: newName)) or:
	 [FileDirectory default fileOrDirectoryExists: (self fullNameForChangesNamed: newName)]) ifTrue: [
		(self confirm: ('{1} already exists. Overwrite?' translated format: {newName})) ifFalse: [^nil]].
	^newName
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'VeronicaUquillas 6/11/2010 13:22'!
getSystemAttribute: attributeID 
	"Optional. Answer the string for the system attribute with the given 
	integer ID. Answer nil if the given attribute is not defined on this 
	platform. On platforms that support invoking programs from command 
	lines (e.g., Unix), this mechanism can be used to pass command line 
	arguments to programs written in Pharo.

	By convention, the first command line argument that is not a VM
	configuration option is considered a 'document' to be filed in. Such a
	document can add methods and classes, can contain a serialized object,
	can include code to be executed, or any combination of these.

	Currently defined attributes include: 
	-1000...-1 - command line arguments that specify VM options 
	0 - the full path name for currently executing VM 
	(or, on some platforms, just the path name of the VM's directory) 
	1 - full path name of this image 
	2 - a Pharo document to open, if any 
	3...1000 - command line arguments for Pharo programs 
	1001 - this platform's operating system 
	1002 - operating system version 
	1003 - this platform's processor type
	1004 - vm version"

	<primitive: 149>
	^ nil! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'sd 6/27/2003 23:47'!
getVMParameters	
	"Answer an Array containing the current values of the VM's internal
	parameter/metric registers.  Each value is stored in the array at the
	index corresponding to its VM register.  (See #vmParameterAt: and
	#vmParameterAt:put:.)"
	"SmalltalkImage current getVMParameters"
	
	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'accessing' stamp: 'StephaneDucasse 3/13/2010 15:33'!
globals
	"Answer the global SystemDictionary"
	^globals! !

!SmalltalkImage methodsFor: 'private' stamp: 'StephaneDucasse 3/13/2010 15:33'!
globals: aSystemDictionary
	"Sets the system-wide globals"
	
	globals ifNotNil: [self error: 'Cannot overwrite existing globals'].
	globals := aSystemDictionary! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 4/28/2010 09:41'!
hasBindingThatBeginsWith: aString

	^ self globals hasBindingThatBeginsWith: aString! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 8/19/2010 22:32'!
hasClassNamed: aString
	"Answer whether there is a class of the given name, but don't intern aString if it's not alrady interned."

	^ globals hasClassNamed: aString! !

!SmalltalkImage methodsFor: 'testing' stamp: 'StephaneDucasse 3/17/2010 20:49'!
hasDisplay

	| arg i |
	i := -1.
	[(arg := self getSystemAttribute: i) isNil]
		whileFalse: [(#('-nodisplay' '-headless') includes: arg) ifTrue: [^ false].
			i := i - 1].
	^ true! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'MarcusDenker 3/22/2010 17:01'!
hasSpecialSelector: aLiteral ifTrueSetByte: aBlock
	"Some basic messages such as #+ have their own bytecode for speed.
	To gain some space, the literal array of compiled methods does not contain such message selectors.
	This implies that the cross referencer and other tools have to check if a selector is of the optimized one 
	or not by calling this method. This is an optimization when machines were slow and small and it 
	may be removed in the future to gain simplicity."
	
	1 to: self specialSelectorSize do:
		[:index | 
		(self specialSelectorAt: index) == aLiteral
			ifTrue: [aBlock value: index + 16rAF. ^true]].
	^false! !

!SmalltalkImage methodsFor: 'image' stamp: 'dtl 4/11/2010 11:45'!
image
	"Answer the object to query about the current object memory and execution environment."
	
	^self! !

!SmalltalkImage methodsFor: 'image' stamp: 'dtl 4/11/2010 11:47'!
imageFormatVersion
	"Answer an integer identifying the type of image. The image version number may
	identify the format of the image (e.g. 32 or 64-bit word size) or specific requirements
	of the image (e.g. block closure support required). This invokes an optional primitive
	that may not be available on all virtual machines."

	"Smalltalk image imageFormatVersion"

	<primitive: 'primitiveImageFormatVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveImageFormatVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'yo 7/2/2004 13:24'!
imageName
	"Answer the full path name for the current image."
	"SmalltalkImage current imageName"

	| str |
	str := self primImageName.
	^ (FilePath pathName: str isEncoded: true) asSqueakPathName.
! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'yo 2/24/2005 18:34'!
imageName: newName
	"Set the the full path name for the current image.  All further snapshots will use this."

	| encoded |
	encoded := (FilePath pathName: newName isEncoded: false) asVmPathName.
	self primImageName: encoded.
! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'sd 11/16/2003 13:57'!
imagePath
	"Answer the path for the directory containing the image file."
	"SmalltalkImage current imagePath"

	^ FileDirectory dirPathFor: self imageName
! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:58'!
includes: element
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals includes: element! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:57'!
includesKey: key
	"Answer whether the receiver has a key equal to the argument, key."

	^globals includesKey: key! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
installLowSpaceWatcher
	"Start a process to watch for low-space conditions."
	"Smalltalk installLowSpaceWatcher"

	self primSignalAtBytesLeft: 0.  "disable low-space interrupts"
	LowSpaceProcess == nil ifFalse: [LowSpaceProcess terminate].
	LowSpaceProcess := [self lowSpaceWatcher] newProcess.
	LowSpaceProcess priority: Processor lowIOPriority.
	LowSpaceProcess resume.

! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'StephaneDucasse 3/17/2010 21:50'!
internalizeChangeLog    
		"Smalltalk internalizeChangeLog"
	"Bring the changes file into a memory-resident filestream, for faster access and freedom from external file system.  1/31/96 sw"

	| reply aName aFile |
	reply := self confirm:  'CAUTION -- do not undertake this lightly!!
If you have backed up your system and
are prepared to face the consequences of
the requested internalization of sources,
hit Yes.  If you have any doubts, hit No
to back out with no harm done.'.

	(reply ==  true) ifFalse:
		[^ self inform: 'Okay - abandoned'].

	aName := Smalltalk changesName.
	(aFile := SourceFiles last) == nil ifTrue:
		[(FileDirectory default fileExists: aName)
			ifFalse: [^ self halt: 'Cannot locate ', aName, ' so cannot proceed.'].
		aFile := FileStream readOnlyFileNamed: aName].
	SourceFiles at: 2 put: (ReadWriteStream with: aFile contentsOfEntireFile).

	self inform: 'Okay, changes file internalized'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'StephaneDucasse 3/21/2010 15:34'!
internalizeSources    
		"Smalltalk internalizeSources"
	"Bring the sources and changes files into memory-resident filestreams, for faster access and freedom from file-system interface.  1/29/96 sw"

	| reply aName aFile |
	reply := self confirm:  'CAUTION -- do not undertake this lightly!!
If you have backed up your system and
are prepared to face the consequences of
the requested internalization of sources,
hit Yes.  If you have any doubts, hit No
to back out with no harm done.'.

	(reply ==  true) ifFalse:
		[^ self inform: 'Okay - abandoned'].

	aName := self sourcesName.
	(aFile := SourceFiles first) == nil ifTrue:
		[(FileDirectory default fileExists: aName)
			ifFalse: [^ self halt: 'Cannot locate ', aName, ' so cannot proceed.'].
		aFile := FileStream readOnlyFileNamed: aName].
	SourceFiles at: 1 put: (ReadWriteStream with: aFile contentsOfEntireFile).

	aName := self changesName.
	(aFile := SourceFiles last) == nil ifTrue:
		[(FileDirectory default fileExists: aName)
			ifFalse: [^ self halt: 'Cannot locate ', aName, ' so cannot proceed.'].
		aFile := FileStream readOnlyFileNamed: aName].
	SourceFiles at: 2 put: (ReadWriteStream with: aFile contentsOfEntireFile).

	self inform: 'Okay, sources internalized'! !

!SmalltalkImage methodsFor: 'vm' stamp: 'dtl 4/11/2010 11:38'!
interpreterSourceVersion
	"Answer a string corresponding to the version of the interpreter source.
	This represents the version level of the Smalltalk source code (interpreter
	and various plugins) that is translated to C by a CCodeGenerator, as distinct
	from the external platform source code, typically written in C and managed
	separately for each platform. An optional primitive is invoked that may not
	be available on all virtual machines."

	"Smalltalk vm interpreterSourceVersion"

	<primitive: 'primitiveInterpreterSourceVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveInterpreterSourceVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'endian' stamp: 'sd 6/27/2003 23:25'!
isBigEndian
	^self endianness == #big! !

!SmalltalkImage methodsFor: 'endian' stamp: 'sd 6/27/2003 23:25'!
isLittleEndian
	^self endianness == #little! !

!SmalltalkImage methodsFor: 'system attribute - deprecated' stamp: 'dfsdfsd 6/11/2010 15:53'!
isMacOS

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isMacOS! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:46'!
isRoot: oop
	"Primitive. Answer whether the object is currently a root for youngSpace."
	<primitive: 'primitiveIsRoot'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'GermanArduino 8/2/2010 08:39'!
isRunningCog
	"Returns true if we're running on a Cog VM (JIT or StackInterpreter)
	 (vmParameterAt: 42 is the number of stack pages)"

	^[(self vmParameterAt: 42) > 0] on: Error do:[:ex| ex return: false]! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'GermanArduino 8/2/2010 08:39'!
isRunningCogit
	"Returns true if we're running on the Cog JIT
	 (vmParameterAt: 46 is the size of the machine code zone)"

	^[(self vmParameterAt: 46) > 0] on: Error do:[:ex| ex return: false]! !

!SmalltalkImage methodsFor: 'printing' stamp: 'stephane.ducasse 9/10/2008 21:36'!
isSelfEvaluating
	self == Smalltalk ifTrue: [^true].
	^super isSelfEvaluating! !

!SmalltalkImage methodsFor: 'testing' stamp: 'SeanDeNigris 8/16/2010 12:22'!
isTrunk
	"Trunk image file names use several letter combinations e.g. 'a' for alpha, 'rc' for release candidates"

	^ (self version allButFirst: 5) anySatisfy: [ :char | char isLetter ]! !

!SmalltalkImage methodsFor: 'system attribute - deprecated' stamp: 'dfsdfsd 6/11/2010 15:53'!
isUnix

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isUnix ! !

!SmalltalkImage methodsFor: 'system attribute - deprecated' stamp: 'dfsdfsd 6/11/2010 15:53'!
isWin32

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isWin32 ! !

!SmalltalkImage methodsFor: 'system attribute - deprecated' stamp: 'dfsdfsd 6/11/2010 15:53'!
isWin32CE

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isWin32CE! !

!SmalltalkImage methodsFor: 'system attribute - deprecated' stamp: 'dfsdfsd 6/11/2010 15:53'!
isX11

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isX11! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:47'!
isYoung: oop
	"Primitive. Answer whether the object currently resides in youngSpace."
	<primitive: 'primitiveIsYoung'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:58'!
keyAtIdentityValue: anObject ifAbsent: aBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keyAtIdentityValue: anObject ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'lr 3/17/2010 18:00'!
keyAtValue: anObject
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keyAtValue: anObject! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:58'!
keys
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keys! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'lr 3/17/2010 17:26'!
keysAndValuesDo: aBlock
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keysAndValuesDo: aBlock
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'sd 11/16/2003 14:02'!
lastQuitLogPosition
	^ LastQuitLogPosition! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'sd 11/16/2003 14:03'!
lastQuitLogPosition: aNumber
	"should be only use to ensure the transition from SystemDictionary to SmalltalkImage, then  	be removed"
		
	LastQuitLogPosition := aNumber! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'MarcusDenker 12/2/2009 07:34'!
lastRemoval
	"Smalltalk lastRemoval"
	"Some explicit removals - add unwanted methods keeping
	other methods."
	| oldDicts newDicts |
	#(#abandonSources )
		do: [:each | self class removeSelector: each].
	"Get rid of all unsent methods."
	"Shrink method dictionaries."
	self garbageCollect.
	oldDicts := MethodDictionary allInstances.
	newDicts := Array new: oldDicts size.
	oldDicts
		withIndexDo: [:d :index | newDicts at: index put: d rehashWithoutBecome].
	oldDicts elementsExchangeIdentityWith: newDicts.
	oldDicts := newDicts := nil.
	self
		allClassesDo: [:c | c zapOrganization].
	SystemOrganization := nil.
	ChangeSet current initialize! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'al 10/12/2008 21:04'!
lastUpdateString
	"SmalltalkImage current lastUpdateString"
	^'Latest update: #' translated, SystemVersion current highestUpdate printString! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'al 10/12/2008 21:03'!
licenseString
	^ 'LICENSE
	
Licensed under the MIT License with parts under the Apache License.

Copyright (c) Pharo Project, and Contributors Copyright (c) 1996-2008 Viewpoints
Research Institute, and Contributors Copyright (c) 1996 Apple Computer, Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions: The
above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

You may obtain a copy of the Apache License at
http://www.apache.org/licenses/LICENSE-2.0'! !

!SmalltalkImage methodsFor: 'modules' stamp: 'sd 6/27/2003 23:47'!
listBuiltinModule: index
	"Return the name of the n-th builtin module.
	This list is not sorted!!"
	<primitive: 572>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'modules' stamp: 'PeterHugossonMiller 9/3/2009 11:22'!
listBuiltinModules
	"SmalltalkImage current listBuiltinModules"
	"Return a list of all builtin modules (e.g., plugins). Builtin plugins are those that are 	compiled with the VM directly, as opposed to plugins residing in an external shared library. 	The list will include all builtin plugins regardless of whether they are currently loaded 
	or not. Note that the list returned is not sorted!!"

	| modules index name |
	modules := Array new writeStream.
	index := 1.
	[true] whileTrue:[
		name := self listBuiltinModule: index.
		name ifNil:[^modules contents].
		modules nextPut: name.
		index := index + 1.
	].! !

!SmalltalkImage methodsFor: 'modules' stamp: 'sd 6/27/2003 23:48'!
listLoadedModule: index
	"Return the name of the n-th loaded module.
	This list is not sorted!!"
	<primitive: 573>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'modules' stamp: 'PeterHugossonMiller 9/3/2009 11:22'!
listLoadedModules
	"SmalltalkImage current listLoadedModules"
	"Return a list of all currently loaded modules (e.g., plugins). Loaded modules are those that currently in use (e.g., active). The list returned will contain all currently active modules regardless of whether they're builtin (that is compiled with the VM) or external (e.g., residing in some external shared library). Note that the returned list is not sorted!!"
	| modules index name |
	modules := Array new writeStream.
	index := 1.
	[true] whileTrue:[
		name := self listLoadedModule: index.
		name ifNil:[^modules contents].
		modules nextPut: name.
		index := index + 1.
	].! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'ar 4/10/2005 18:02'!
logChange: aStringOrText 
	"Write the argument, aString, onto the changes file."
	| aString changesFile |
	(SourceFiles isNil or: [(SourceFiles at: 2) == nil]) ifTrue: [^ self].
	self assureStartupStampLogged.

	aString := aStringOrText asString.
	(aString findFirst: [:char | char isSeparator not]) = 0
		ifTrue: [^ self].  "null doits confuse replay"
	(changesFile := SourceFiles at: 2).
	changesFile isReadOnly ifTrue:[^self].
	changesFile setToEnd; cr; cr.
	changesFile nextChunkPut: aString.
		"If want style changes in DoIt, use nextChunkPutWithStyle:, and allow Texts to get here"
	self forceChangesToDisk.! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'StephaneDucasse 3/21/2010 12:17'!
logError: errMsg inContext: aContext to: aFilename
	"Log the error message and a stack trace to the given file."

	"wks - 9-09 - do not delete existing errors.  Note that this could be made
	conditional or left here in this state for those who want it.
	FileDirectory default deleteFileNamed: aFilename ifAbsent: []."
	
	FileStream fileNamed: aFilename do: [:ff |
		ff wantsLineEndConversion: true.
		ff setToEnd.
		ff nextPutAll: 'THERE_BE_DRAGONS_HERE'; cr. 
	  	ff nextPutAll: errMsg; cr.
		aContext errorReportOn: ff.
		"wks 9-09 - write some type of separator"
		ff nextPutAll: (String new: 60 withAll: $- ); cr; cr].
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'di 8/18/2000 16:49'!
lowSpaceThreshold 
	"Return the low space threshold. When the amount of free memory (after garbage collection) falls below this limit, the system is in serious danger of completely exhausting memory and crashing. This limit should be made high enough to allow the user open a debugger to diagnose a problem or to save the image."

	thisContext isPseudoContext
		ifTrue: [^ 400000  "Enough for JIT compiler"]
		ifFalse: [^ 200000  "Enough for interpreter"]! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'StephaneDucasse 3/20/2010 23:17'!
lowSpaceWatcher
	"Wait until the low space semaphore is signalled, then take appropriate
	actions. "
	| free preemptedProcess |
	self garbageCollectMost <= self lowSpaceThreshold
		ifTrue: [self garbageCollect <= self lowSpaceThreshold
				ifTrue: ["free space must be above threshold before
					starting low space watcher"
					^ Beeper beep]].
	self specialObjectsArray at: 23 put: nil.
	"process causing low space will be saved here"
	LowSpaceSemaphore := Semaphore new.
	self primLowSpaceSemaphore: LowSpaceSemaphore.
	self primSignalAtBytesLeft: self lowSpaceThreshold.
	"enable low space interrupts"
	LowSpaceSemaphore wait.
	"wait for a low space condition..."
	self primSignalAtBytesLeft: 0.
	"disable low space interrupts"
	self primLowSpaceSemaphore: nil.
	LowSpaceProcess := nil.
	"The process that was active at the time of the low space interrupt."
	preemptedProcess := Smalltalk specialObjectsArray at: 23.
	Smalltalk specialObjectsArray at: 23 put: nil.
	"Note: user now unprotected until the low space watcher is re-installed"
	self memoryHogs isEmpty
		ifFalse: [free := self bytesLeft.
			self memoryHogs
				do: [:hog | hog freeSomeSpace].
			self bytesLeft > free
				ifTrue: [^ self installLowSpaceWatcher]].
	UIManager default lowSpaceWatcherDefaultAction: preemptedProcess! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'nk 10/28/2000 20:37'!
lowSpaceWatcherProcess
	^LowSpaceProcess! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'sma 4/22/2000 19:03'!
memoryHogs
	"Answer the list of objects to notify with #freeSomeSpace if memory gets full."

	^ MemoryHogs ifNil: [MemoryHogs := OrderedCollection new]! !

!SmalltalkImage methodsFor: 'objects from disk' stamp: 'tk 9/28/2000 15:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to Smalltalk instead."

	dp := DiskProxy global: #Smalltalk selector: #yourself
			args: #().
	refStrm replace: self with: dp.
	^ dp! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
okayToProceedEvenIfSpaceIsLow
	"Return true if either there is enough memory to do so safely or if the user gives permission after being given fair warning."

	self garbageCollectMost > self lowSpaceThreshold ifTrue: [^ true].  "quick"
	self garbageCollect > self lowSpaceThreshold ifTrue: [^ true].  "work harder"

	^ self confirm:
'WARNING: There is not enough space to start the low space watcher.
If you proceed, you will not be warned again, and the system may
run out of memory and crash. If you do proceed, you can start the
low space notifier when more space becomes available simply by
opening and then closing a debugger (e.g., by hitting Cmd-period.)
Do you want to proceed?'
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'AlainPlantec 12/11/2009 14:09'!
openSourceFiles

	self imageName = LastImageName 
		ifFalse: ["Reset the author full name to blank when the image gets moved"
			LastImageName := self imageName.
			Author fullName: ''].
	self checkAndOpenSourcesAndChanges.
	StandardSourceFileArray install! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 6/30/2010 22:51'!
optionAt: i
	"Answer the i-th option of the command line, or nil if not so many options."
	
	^self getSystemAttribute: i negated! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'GermanArduino 8/2/2010 18:56'!
options
	"Answer an array with all the command line options."
	
	"Smalltalk commandLine options"
	
	^Array streamContents: [:str |
		| arg i |
		i := 1.
		[i > 1000 or: [(arg := self optionAt: i) == nil]]
			whileFalse:
				[str nextPut: arg.
				i := i + 1]].! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 3/13/2010 16:39'!
organization
	"Return the organizer for the receiver"
	^globals organization! !

!SmalltalkImage methodsFor: 'os - deprecated' stamp: 'StephaneDucasse 9/3/2010 21:31'!
os
	"Answer the object to query about os."
	
	^ OSPlatform! !

!SmalltalkImage methodsFor: 'os - deprecated' stamp: 'dfsdfsd 6/11/2010 15:49'!
osVersion
	"Return the version number string of the platform we're running on"
	"SmalltalkImage current osVersion"

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform osVersion! !

!SmalltalkImage methodsFor: 'os' stamp: 'StephaneDucasse 8/8/2010 18:16'!
platform
	"Answer the object to query about os."
	
	^ OSPlatform! !

!SmalltalkImage methodsFor: 'os - deprecated' stamp: 'GuillermoPolito 10/6/2010 17:22'!
platformName
	"Return the name of the platform we're running on"

	self deprecated: 'Use Smalltalk os platformName'.
	^self os platformName! !

!SmalltalkImage methodsFor: 'vm' stamp: 'dtl 4/11/2010 11:39'!
platformSourceVersion
	"Answer a string corresponding to the version of the external platform source
	code, typically written in C and managed separately for each platform. This
	invokes an optional primitive that may not be available on all virtual machines."

	"Smalltalk vm platformSourceVersion"

	<primitive: 'primitivePlatformSourceVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitivePlatformSourceVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'os - deprecated' stamp: 'dfsdfsd 6/11/2010 15:50'!
platformSubtype
	"Return the subType of the platform we're running on"

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform platformSubtype ! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'nice 1/5/2010 15:59'!
poolUsers
	"Answer a dictionary of pool name -> classes that refer to it.
	Also includes any globally know dictionaries (such as
	Smalltalk, Undeclared etc) which although not strictly
	accurate is potentially useful information"
	"Smalltalk poolUsers"
	| poolUsers |
	poolUsers := Dictionary new.
	self keys
		do: [:k | | pool refs | "yes, using isKindOf: is tacky but for reflective code like
			this it is very useful. If you really object you can:-
			a) go boil your head.
			b) provide a better answer.
			your choice."
			(((pool := self at: k) isKindOf: Dictionary)
					or: [pool isKindOf: SharedPool class])
				ifTrue: [refs := self systemNavigation allClasses
								select: [:c | c sharedPools identityIncludes: pool]
								thenCollect: [:c | c name].
					refs
						add: (self systemNavigation
								allReferencesTo: (self associationAt: k)).
					poolUsers at: k put: refs]].
	^ poolUsers! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primBytesLeft
	"Primitive. Answer the number of bytes available for new object data.
	Not accurate unless preceded by
		Smalltalk garbageCollectMost (for reasonable accuracy), or
		Smalltalk garbageCollect (for real accuracy).
	See Object documentation whatIsAPrimitive."

	<primitive: 112>
	^ 0! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: 'yo 3/29/2004 09:36'!
primImageName
	"Answer the full path name for the current image."
	"Smalltalk imageName"

	<primitive: 121>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: 'yo 3/29/2004 09:36'!
primImageName: newName
	"Set the the full path name for the current image.  All further snapshots will use this."

	<primitive: 121>
	^ self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primLowSpaceSemaphore: aSemaphore
	"Primitive. Register the given Semaphore to be signalled when the
	number of free bytes drops below some threshold. Disable low-space
	interrupts if the argument is nil."

	<primitive: 124>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primSignalAtBytesLeft: numBytes
	"Tell the interpreter the low-space threshold in bytes. When the free
	space falls below this threshold, the interpreter will signal the low-space
	semaphore, if one has been registered.  Disable low-space interrupts if the
	argument is zero.  Fail if numBytes is not an Integer."

	<primitive: 125>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: 'yo 3/29/2004 09:36'!
primVmPath
	"Answer the path for the directory containing the Smalltalk virtual machine. Return the empty string if this primitive is not implemented."
	"Smalltalk vmPath"

	<primitive: 142>
	^ ''! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 2/11/2001 02:16'!
primitiveGarbageCollect
	"Primitive. Reclaims all garbage and answers the number of bytes of available space."

	<primitive: 130>
	^ self primBytesLeft! !

!SmalltalkImage methodsFor: 'printing' stamp: 'sma 6/1/2000 09:53'!
printElementsOn: aStream
	aStream nextPutAll:'(lots of globals)'! !

!SmalltalkImage methodsFor: 'printing' stamp: 'stephane.ducasse 9/10/2008 21:36'!
printOn: aStream
	self == Smalltalk ifFalse: [^super printOn: aStream].
	aStream nextPutAll: 'Smalltalk'! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 8/8/2010 18:17'!
processHasThreadIdInstVar: aBoolean
	"The threaded VM needs to know if the 4th inst var of Process is threadId which it uses to control process-to-thread binding. This flag persists across snapshots, stored in the image header."
	aBoolean ifTrue: [self assert: (Process instVarNames at: 4) ='threadId'].
	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 1) + (aBoolean ifTrue: [1] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 8/8/2010 18:17'!
processPreemptionYields
	"Answer whether the VM causes a process to yield on process preemption, i.e. to put a preempted process at the back of its run queue.  If the parameter is unavailable (non-Cog VMs) or bit 2 (4) is 0 then preemption yields."

	^(([self vmParameterAt: 48]
			on: Error
			do: [:ex| ^true]) allMask: 4) not! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 8/8/2010 18:17'!
processPreemptionYields: aBoolean
	"The Cog VM can be instructed not to yield on process preemption, i.e. not to put a preempted process at the back of its run queue.  By default preempting a process causes it to yield (Blue Book semantics) which can have unfortunate effects. This flag persists across snapshots, stored in the image header."
	
	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 4) + (aBoolean ifTrue: [0] ifFalse: [4])! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'AdrianLienhard 9/1/2010 20:58'!
processShutDownList: quitting
	"Send #shutDown to each class that needs to wrap up before a snapshot.
	Also void the endianness chace; this can't safely be done on start-up because
	Smalltalk is too late in the start-up sequence."

	EndianCache := nil.
	self send: #shutDown: toClassesNamedIn: ShutDownList with: quitting.
! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ar 11/16/1999 20:12'!
processStartUpList: resuming
	"Send #startUp to each class that needs to run initialization after a snapshot."

	self send: #startUp: toClassesNamedIn: StartUpList with: resuming.
! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
quitPrimitive
	"Primitive. Exit to another operating system on the host machine, if one
	exists. All state changes in the object space since the last snapshot are lost.
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 113>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'bf 1/12/2006 19:20'!
recordStartupStamp
	
	StartupStamp := '----STARTUP----', Time dateAndTimeNow printString, ' as ', self imageName.
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'sw 2/3/2000 15:59'!
recover: nCharacters
	"Schedule an editable text view on the last n characters of changes."
	self writeRecentCharacters: nCharacters toFileNamed: 'st80.recent'! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'Igor.Stasenko 9/21/2010 03:09'!
recreateSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	| newArray |
	newArray := Array new: 56.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (self associationAt: #Processor).
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: Float.
	newArray at: 11 put: MethodContext.
	newArray at: 12 put: BlockContext.
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: (self specialObjectsArray at: 18).
	"(low space Semaphore)"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 #class 0
							#blockCopy: 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it."
	newArray at: 25 put: (self specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers"
	newArray at: 29 put: self compactClassesArray.
	newArray at: 30 put: (self specialObjectsArray at: 30).
	"(delay Semaphore)"
	newArray at: 31 put: (self specialObjectsArray at: 31).
	"(user interrupt Semaphore)"
	"Prototype instances that can be copied for fast initialization"
	newArray at: 32 put: (Float new: 2).
	newArray at: 33 put: (LargePositiveInteger new: 4).
	newArray at: 34 put: Point new.
	newArray at: 35 put: #cannotInterpret:.
	"Note: This must be fixed once we start using context prototypes (yeah, right)"
	"(MethodContext new: CompiledMethod fullFrameSize)."
	newArray at: 36 put: (self specialObjectsArray at: 36). "Is the prototype MethodContext (unused by the VM)"
	newArray at: 37 put: BlockClosure.
	"(BlockContext new: CompiledMethod fullFrameSize)."
	newArray at: 38 put: (self specialObjectsArray at: 38). "Is the prototype BlockContext (unused by the VM)"
	"array of objects referred to by external code"
	newArray at: 39 put: (self specialObjectsArray at: 39).	"preserve external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: nil. "Reserved for a LinkedList instance for overlapped calls in CogMT"
	"finalization Semaphore"
	newArray at: 42 put: ((self specialObjectsArray at: 42) ifNil: [Semaphore new]).
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (self at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (self at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (self at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (self at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (self at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	"newArray at: 51 put: #attemptToAssign:withIndex:."
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move').
	"53 to 55 are for Alien"
	newArray at: 53 put: (self at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallback:stack:registers:jmpbuf:.
	newArray at: 55 put: (self at: #UnsafeAlien ifAbsent: []).

	"Weak reference finalization"
	newArray at: 56 put: (self at: #WeakFinalizationList ifAbsent: []).

	"Now replace the interpreter's reference in one atomic operation"
	self specialObjectsArray become: newArray
	! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'JMM 6/6/2000 20:39'!
registerExternalObject: anObject
	"Register the given object in the external objects array and return its index. If it is already there, just return its index."

	^ExternalSemaphoreTable registerExternalObject: anObject! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'HenrikSperreJohansen 6/28/2010 15:17'!
removeAllLineFeeds
	"Smalltalk removeAllLineFeeds"
	"Scan all methods for source code with lineFeeds.
	Replaces all occurrences of <CR><LF> or <LF> by <CR>.
	When done, offers to display an Inspector containing the message
	names grouped by author initials.
	In this dictionary, the key 'OK' contains the methods that had literals that contained <LF> characters."
	| totalStripped totalOK authors |
	'Scanning sources for LineFeeds.
This will take a few minutes...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: [:bar | | n | 
			n := 0.
			authors := self
						removeAllLineFeedsQuietlyCalling: [(n := n + 1) \\ 100 = 0
								ifTrue: [bar value: n]]].
	totalStripped := authors
				inject: 1
				into: [:sum :set | sum + set size].
	totalOK := (authors at: 'OK') size.
	totalStripped := totalStripped - totalOK.
	Transcript cr; show: totalStripped printString , ' methods stripped of LFs.'.
	Transcript cr; show: totalOK printString , ' methods still correctly contain LFs.'.
	(self confirm: 'Do you want to see the affected methods?')
		ifTrue: [authors inspect]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'HenrikSperreJohansen 6/28/2010 15:17'!
removeAllLineFeedsQuietly
	"Smalltalk removeAllLineFeedsQuietly"
	"Scan all methods for source code with lineFeeds.
	Replaces all occurrences of <CR><LF> or <LF> by <CR>.
	Answer a Dictionary keyed by author name containing sets of affected method names,
	as well as (at the key 'OK') a list of methods that still contain LF characters inside literal strings or characters."
	^self removeAllLineFeedsQuietlyCalling: [].! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'HenrikSperreJohansen 6/28/2010 15:17'!
removeAllLineFeedsQuietlyCalling: aBlock
	"Smalltalk removeAllLineFeedsQuietly"
	"Scan all methods for source code with lineFeeds.
	Replaces all occurrences of <CR><LF> or <LF> by <CR>.
	Answer a Dictionary keyed by author name containing sets of affected method names,
	as well as (at the key 'OK') a list of methods that still contain LF characters inside literal strings or characters.
	Evaluate aBlock for each method so that status can be updated."
	| authors |
	authors := Dictionary new.
	authors at: 'OK' put: Set new.
	self systemNavigation
		allBehaviorsDo: [:cls | cls selectorsAndMethodsDo: [:selector :method |
					| oldStamp newCodeString oldCodeString oldCategory nameString | 
					aBlock cull: cls cull: selector.
					oldCodeString := cls sourceCodeAt: selector.
					(oldCodeString includes: Character lf)
						ifTrue: [
							newCodeString := oldCodeString withSqueakLineEndings.
							nameString := cls name , '>>' , selector.
							((cls compiledMethodAt: selector) hasLiteralSuchThat: [ :lit | lit asString includes: Character lf ])
								ifTrue: [(authors at: 'OK')
										add: nameString]
								ifFalse: [oldStamp := (method timeStamp)
												copy replaceAll: Character cr
												with: Character space.
									(authors
										at: (oldStamp copyFrom: 1 to: (oldStamp findFirst: [ :c | c isAlphaNumeric not ]))
										ifAbsentPut: [Set new])
										add: nameString.
									oldCategory := cls whichCategoryIncludesSelector: selector.
									cls
										compile: newCodeString
										classified: oldCategory
										withStamp: oldStamp
										notifying: nil ]]]].
	^ authors! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 8/19/2010 22:20'!
removeClassNamed: aName
	"Invoked from fileouts:  if there is currently a class in the system named aName, then remove it.  If anything untoward happens, report it in the Transcript.  "

	self globals removeClassNamed: aName! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'sd 9/29/2004 18:15'!
removeEmptyMessageCategories
	"Smalltalk removeEmptyMessageCategories"
	self garbageCollect.
	(ClassOrganizer allInstances copyWith: SystemOrganization)
		do: [:org | org removeEmptyCategories]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:22'!
removeFromShutDownList: aClass

	ShutDownList remove: aClass name ifAbsent: []! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'di 2/3/1999 22:22'!
removeFromStartUpList: aClass

	StartUpList remove: aClass name ifAbsent: []! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:58'!
removeKey: key
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals removeKey: key! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:58'!
removeKey: key ifAbsent: aBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals removeKey: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'StephaneDucasse 8/19/2010 22:36'!
removeSelector: descriptor 
	"Safely remove a selector from a class (or metaclass). If the
	class or the method doesn't exist anymore, never mind and
	answer nil.
	This method should be used instead of 'Class removeSelector:
	#method' to omit global class references."
	| class sel |
	class := self
				at: descriptor first
				ifAbsent: [^ nil].
	(descriptor size > 2 and: [descriptor second == #class])
		ifTrue: 
			[class := class class.
			sel := descriptor third]
		ifFalse: 
			[sel := descriptor second].
	^ class removeSelector: sel! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 3/13/2010 16:39'!
renameClass: aClass as: newName 
	"Rename the class, aClass, to have the title newName."

	^globals renameClass: aClass as: newName! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 3/13/2010 16:39'!
renameClass: aClass from: oldName
	"Rename the class, aClass, to have the title newName."

	^globals renameClass: aClass from: oldName! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'sw 10/28/96'!
renameClassNamed: oldName as: newName
	"Invoked from fileouts:  if there is currently a class in the system named oldName, then rename it to newName.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: oldName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Class-rename for ', oldName, ' ignored because ', oldName, ' does not exist.'.
			^ self].

	oldClass rename: newName! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 3/21/2010 12:43'!
renamedClass: aClass from: oldName to: newName
	"Inform Smalltalk that aClass has been renamed"
	
	(Array with: StartUpList with: ShutDownList) 
		do: [ :list | | index |  
			index := list indexOf: oldName ifAbsent: [0].
			index > 0 ifTrue: [list at: index put: newName]].
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'StephaneDucasse 2/16/2010 13:34'!
reportCPUandRAM
	"Write several text files with useful analysis for profiling purposes.
	Overwrites any existing report.
	SmalltalkImage current reportCPUandRAM
	"	

	| stream |
	"VM statistics (Memory use and GC, mainly)"
	FileStream 
		forceNewFileNamed: 'Log-MemoryStats.txt'
		do: [:aStream | aStream nextPutAll: self vmStatisticsReportString ].
	
	"Process list"
	stream := FileStream 
				forceNewFileNamed: 'Log-ProcessList.txt' 
				do: [ :aStream |  ProcessBrowser new processNameList 
								do: [ :each | aStream nextPutAll: each; cr ]].

"Fork all these, so they run in sequence, as the system is back running"
[
	"Process taking most CPU"
	stream := FileStream forceNewFileNamed: 'Log-ThePig.txt'.
	ProcessBrowser dumpPigStackOn: stream andClose: true.
	
	"Tally of all processes"
	stream := FileStream forceNewFileNamed: 'Log-FullTally.txt'.
	[ | tally |
		tally := MessageTally new.
		tally reportOtherProcesses: true.
		tally spyAllEvery: 1 on: [ (Delay forMilliseconds: 1000) wait ].
		tally report: stream ] ensure: [ stream close ].

	"Memory Analysis"
	stream := FileStream forceNewFileNamed: 'Log-MemoryAnalysis.txt'.
	[ SpaceTally new printSpaceAnalysis: 1 on: stream ]
		ensure: [ stream close ]

] forkNamed: 'CPU usage analysis'! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:48'!
rootTable
	"Primitive. Answer a snapshot of the VMs root table. 
	Keep in mind that the primitive may itself cause GC."
	<primitive: 'primitiveRootTable'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:49'!
rootTableAt: index
	"Primitive. Answer the nth element of the VMs root table"
	<primitive: 'primitiveRootTableAt'>
	^nil! !

!SmalltalkImage methodsFor: 'saving' stamp: 'kph 1/29/2007 00:52'!
saveAs

	"Put up the 'saveAs' prompt, obtain a name, and save the image  under that new name."

	self saveAs: self getFileNameFromUser! !

!SmalltalkImage methodsFor: 'saving' stamp: 'nice 5/12/2009 00:13'!
saveAs: newName
	newName ifNil: [ ^ self ].
	(SourceFiles at: 2) ifNotNil:
		[self closeSourceFiles; "so copying the changes file will always work"
			 saveChangesInFileNamed: (self fullNameForChangesNamed: newName)].
	^self saveImageInFileNamed: (self fullNameForImageNamed: newName)! !

!SmalltalkImage methodsFor: 'saving' stamp: 'MarianoMartinezPeck 1/19/2010 13:58'!
saveAsEmbeddedImage
	"Save the current state of the system as an embedded image"

	| dir newName newImageName oldImageSegDir haveSegs previousFullImageName |
	dir := FileDirectory default.
	previousFullImageName := self imageName.
	newName := UIManager default request: 'Select existing VM file'
				initialAnswer: (FileDirectory localNameFor: '').
	newName isEmptyOrNil ifTrue: [^Smalltalk].
	newName := FileDirectory baseNameFor: newName asFileName.
	newImageName := newName.
	(dir includesKey: newImageName) 
		ifFalse: 
			[^self 
				inform: 'Unable to find name ' , newName , ' Please choose another name.'].
	haveSegs := self areThereImageSegments.

	self logChange: '----SAVEAS (EMBEDDED) ' , newName , '----' 
				, Date dateAndTimeNow printString.
	self imageName: (dir fullNameFor: newImageName) asSqueakPathName.
	LastImageName := self imageName.
	self closeSourceFiles.
	self saveImageSegments: previousFullImageName.

	self 
		snapshot: true
		andQuit: true
		embedded: true! !

!SmalltalkImage methodsFor: 'saving' stamp: 'stephane.ducasse 3/1/2009 08:48'!
saveAsNewVersion
	"Save the image/changes using the next available version number."
	"SmalltalkImage current saveAsNewVersion"
	
	| newName changesName aName anIndex |
	aName := FileDirectory baseNameFor: (FileDirectory default localNameFor: self imageName).
	anIndex := aName lastIndexOf: FileDirectory extensionDelimiter ifAbsent: [nil].
	(anIndex notNil and: [(aName copyFrom: anIndex + 1 to: aName size) isAllDigits])
		ifTrue:
			[aName := aName copyFrom: 1 to: anIndex - 1].

	newName := FileDirectory default nextNameFor: aName extension: FileDirectory imageSuffix.
	changesName := self fullNameForChangesNamed: newName.

	"Check to see if there is a .changes file that would cause a problem if we saved a new .image file with the new version number"
	(FileDirectory default fileOrDirectoryExists: changesName)
		ifTrue:
			[^ self inform:
'There is already .changes file of the desired name,
', newName, '
curiously already present, even though there is
no corresponding .image file.   Please remedy
manually and then repeat your request.'].

	^self saveAs: newName


! !

!SmalltalkImage methodsFor: 'saving' stamp: 'StephaneDucasse 3/21/2010 10:25'!
saveChangesInFileNamed: aString
	| fullChangesName |
	fullChangesName := (FileDirectory default fullNameFor: aString).
	(FileDirectory default directoryNamed:(FileDirectory dirPathFor: fullChangesName )) assureExistence.
	FileDirectory default 
		copyFileWithoutOverwriteConfirmationNamed: Smalltalk changesName 
		toFileNamed: fullChangesName.
	self setMacFileInfoOn: fullChangesName.! !

!SmalltalkImage methodsFor: 'saving' stamp: 'MarianoMartinezPeck 1/19/2010 10:30'!
saveImageInFileNamed: aString
	| fullImageName previousFullImageName |
	fullImageName := (FileDirectory default fullNameFor: aString).
	(FileDirectory default directoryNamed:(FileDirectory dirPathFor: fullImageName )) assureExistence.
	previousFullImageName := self imageName.
	^self
		changeImageNameTo: fullImageName;
		closeSourceFiles;
		openSourceFiles;  "so SNAPSHOT appears in new changes file"
		saveImageSegments: previousFullImageName;
		snapshot: true andQuit: false! !

!SmalltalkImage methodsFor: 'saving' stamp: 'YanniChiu 6/11/2010 21:14'!
saveImageSegments
	self saveImageSegments: self imageName! !

!SmalltalkImage methodsFor: 'saving' stamp: 'lr 3/14/2010 21:13'!
saveImageSegments: previousFullImageName
	"This method is used for example when you save an image with another name. In this case, if there are segments written to files and there are ImageSegment instances, then it copies the file from the previous image location, to the new one. It is like a backup of the segment files when renaming an image. "

	| haveSegs oldImageSegDir theClass newImageSegDir |
	"If the class does not exist or has no instances, it has no sense to backUp the segments."
	self areThereImageSegments
		ifFalse: [ ^ self ].
	theClass := Smalltalk globals at: #ImageSegment.
	oldImageSegDir := theClass segmentDirectoryFrom: previousFullImageName.
	newImageSegDir := theClass segmentDirectory.	"create the folder"
	oldImageSegDir fileNames
		do: [ :theName | 
			| imageSegmentName "copy all segment files" |
			imageSegmentName := oldImageSegDir pathName , FileDirectory slash , theName.
			newImageSegDir copyFileWithoutOverwriteConfirmationNamed: imageSegmentName toFileNamed: theName ]! !

!SmalltalkImage methodsFor: 'saving' stamp: 'sd 11/16/2003 14:12'!
saveSession
	self snapshot: true andQuit: false! !

!SmalltalkImage methodsFor: 'deprecated' stamp: 'StephaneDucasse 3/13/2010 15:58'!
scopeFor: varName from: lower envtAndPathIfFound: envtAndPathBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	(globals includesKey: varName)
		ifTrue: [^ envtAndPathBlock value: self value: String new]
		ifFalse: [^ nil]! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'lr 3/17/2010 17:59'!
select: aBlock
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals select: aBlock
! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'StephaneDucasse 3/22/2010 16:55'!
send: startUpOrShutDown toClassesNamedIn: startUpOrShutDownList with: argument
	"Send the message #startUp: or #shutDown: to each class named in the list.
	The argument indicates if the system is about to quit (for #shutDown:) or if
	the image is resuming (for #startUp:).
	If any name cannot be found, then remove it from the list."

	| removals |
	removals := OrderedCollection new.
	startUpOrShutDownList
		do: [ :name | 
			| class |
			class := self at: name ifAbsent: [ nil ].
			class == nil
				ifTrue: [ removals add: name ]
				ifFalse: [ 
							[ class perform: startUpOrShutDown with: argument ]
								on: Exception
								do: [ :ex | 
									Smalltalk hasDisplay
										ifTrue: [ ex pass ]
										ifFalse: [ 
											globals at: #Console ifPresent: [ :console | console printNl: ex description ].
											self logError: ex printString inContext: thisContext to: 'PharoDebug.log' ] ] ] ].	"Remove any obsolete entries, but after the iteration"
	startUpOrShutDownList removeAll: removals! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'JMM 1/27/2005 13:12'!
setGCBiasToGrow: aNumber
	"Primitive. Indicate that the GC logic should be bias to grow"
	<primitive: 'primitiveSetGCBiasToGrow'>
	^self primitiveFailed
"Example:
	Smalltalk setGCBiasToGrowGCLimit: 16*1024*1024.
	Smalltalk setGCBiasToGrow: 1.
"! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'JMM 1/27/2005 12:27'!
setGCBiasToGrowGCLimit: aNumber
	"Primitive. Indicate that the bias to grow logic should do a GC after aNumber Bytes"
	<primitive: 'primitiveSetGCBiasToGrowGCLimit'>
	^self primitiveFailed
"Example:
	Smalltalk setGCBiasToGrowGCLimit: 16*1024*1024.
"! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'StephaneDucasse 3/21/2010 12:08'!
setGCParameters
	"Adjust the VM's default GC parameters to avoid premature tenuring."

	self vmParameterAt: 5 put: 4000.  "do an incremental GC after this many allocations"
	self vmParameterAt: 6 put: 2000.  "tenure when more than this many objects survive the GC"
	self setGCBiasToGrowGCLimit: 16*1024*1024.
	self  setGCBiasToGrow: 1.
! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'ar 1/18/2005 16:54'!
setGCSemaphore: semaIndex
	"Primitive. Indicate the GC semaphore index to be signaled on GC occurance."
	<primitive: 'primitiveSetGCSemaphore'>
	^self primitiveFailed
"Example:

	| index sema process |
	sema := Semaphore new.
	index := Smalltalk registerExternalObject: sema.
	Smalltalk setGCSemaphore: index.
	process := [
		[[true] whileTrue:[
			sema wait.
			Smalltalk beep.
		]] ensure:[
			Smalltalk setGCSemaphore: 0.
			Smalltalk unregisterExternalObject: sema.
		].
	] fork.
	process inspect.
"! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: 'MPH 10/24/2000 14:27'!
setMacFileInfoOn: aString
	"On Mac, set the file type and creator (noop on other platforms)"
	FileDirectory default
		setMacFileNamed: aString
		type: 'STch'
		creator: 'FAST'.! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'StephaneDucasse 3/17/2010 21:52'!
shutDown
	^ Smalltalk closeSourceFiles! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'gk 2/23/2004 20:51'!
shutDownSound
	"No longer used in the release, but retained for backward compatibility."

	SoundService default shutDown
! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
signalLowSpace
	"Signal the low-space semaphore to alert the user that space is running low."

	LowSpaceSemaphore signal.! !

!SmalltalkImage methodsFor: 'to clean later' stamp: 'StephaneDucasse 3/13/2010 15:56'!
size
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals size! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'sd 11/16/2003 14:12'!
snapshot: save andQuit: quit
	^self snapshot: save andQuit: quit embedded: false! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'StephaneDucasse 3/20/2010 23:20'!
snapshot: save andQuit: quit embedded: embeddedFlag
	"Mark the changes file and close all files as part of #processShutdownList.
	If save is true, save the current state of this Smalltalk in the image file.
	If quit is true, then exit to the outer OS shell.
	The latter part of this method runs when resuming a previously saved image. This resume logic checks for a document file to process when starting up."

	| resuming msg |
	Object flushDependents.
	Object flushEvents.
	(SourceFiles at: 2)
		ifNotNil: [ 
			msg := String
				streamContents: [ :s | 
					s
						nextPutAll: '----';
						nextPutAll:
								(save
										ifTrue: [ 
											quit
												ifTrue: [ 'QUIT' ]
												ifFalse: [ 'SNAPSHOT' ] ]
										ifFalse: [ 
											quit
												ifTrue: [ 'QUIT/NOSAVE' ]
												ifFalse: [ 'NOP' ] ]);
						nextPutAll: '----';
						print: Date dateAndTimeNow;
						space;
						nextPutAll: (FileDirectory default localNameFor: self imageName);
						nextPutAll: ' priorSource: ';
						print: LastQuitLogPosition ].
			self assureStartupStampLogged.
			save
				ifTrue: [ 
					LastQuitLogPosition := (SourceFiles at: 2)
						setToEnd;
						position ].
			self logChange: msg.
			Transcript
				cr;
				show: msg ].
	self processShutDownList: quit.
	Cursor write show.
	save
		ifTrue: [ 
			resuming := embeddedFlag
				ifTrue: [ self snapshotEmbeddedPrimitive ]
				ifFalse: [ self snapshotPrimitive ].	"<-- PC frozen here on image file"
			resuming == false
				ifTrue: [ 
					"Time to reclaim segment files is immediately after a save"
					Smalltalk globals at: #ImageSegment ifPresent: [ :theClass | theClass reclaimObsoleteSegmentFiles ] ]	"guard against failure" ]
		ifFalse: [ resuming := false ].
	quit & (resuming == false)
		ifTrue: [ self quitPrimitive ].
	Cursor normal show.
	self setGCParameters.
	resuming == true
		ifTrue: [ Smalltalk clearExternalObjects ].
	self processStartUpList: resuming == true.
	resuming == true
		ifTrue: [ self recordStartupStamp ].
	UIManager default onSnapshot.	"Now it's time to raise an error"
	resuming == nil
		ifTrue: [ self error: 'Failed to write image file (disk full?)' ].
	^ resuming! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'StephaneDucasse 12/29/2009 22:05'!
snapshotEmbeddedPrimitive
	"This is from experimental code to create single-file applications. 
	It saves a copy of the VM and append an image."
	<primitive: 247>
	^nil "indicates error writing embedded image file"! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: 'ar 7/22/2000 14:34'!
snapshotPrimitive
	"Primitive. Write the current state of the object memory on a file in the
	same format as the Smalltalk-80 release. The file can later be resumed,
	returning you to this exact state. Return normally after writing the file.
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 97>
	^nil "indicates error writing image file"! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'sd 9/24/2003 12:43'!
sourceFileVersionString

	^ SourceFileVersionString! !

!SmalltalkImage methodsFor: 'private' stamp: 'sd 9/24/2003 12:42'!
sourceFileVersionString: aString

	SourceFileVersionString := aString! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'sd 9/24/2003 12:44'!
sourcesName
	"Answer the full path to the version-stable source code"
	^ self vmPath , SourceFileVersionString , FileDirectory dot , 'sources'! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialNargsAt: anInteger 
	"Answer the number of arguments for the special selector at: anInteger."

	^ (self specialObjectsArray at: 24) at: anInteger * 2! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialObjectsArray  "Smalltalk specialObjectsArray at: 1"
	<primitive: 129>
	^ self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectorAt: anInteger 
	"Answer the special message selector from the interleaved specialSelectors array."

	^ (self specialObjectsArray at: 24) at: anInteger * 2 - 1! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectorSize
	"Answer the number of special selectors in the system."

	^ (self specialObjectsArray at: 24) size // 2! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectors
	"Used by SystemTracer only."

	^SpecialSelectors! !

!SmalltalkImage methodsFor: 'objects from disk' stamp: 'tk 3/7/2000 18:40'!
storeDataOn: aDataStream
	"I don't get stored.  Use a DiskProxy"

	self error: 'use a DiskProxy to store me'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'al 10/12/2008 21:05'!
systemInformationString
	^ String cr, SystemVersion current version, String cr, self lastUpdateString, String cr, String cr, String cr, self licenseString! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'nice 1/16/2010 12:58'!
testFormatter
	"Smalltalk testFormatter"

	"Reformats the source for every method in the system, and
	then compiles that source and verifies that it generates
	identical code"

	| badOnes |
	badOnes := OrderedCollection new.
	'Formatting all classes...' 
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: 
			[:bar | | n | 
			n := 0.
			self systemNavigation allBehaviorsDo: 
					[:cls | 
					"Transcript cr; show: cls name."

					cls selectorsAndMethodsDo: 
							[:selector :oldMethod |
							| newCodeString newMethod methodNode | 
							(n := n + 1) \\ 100 = 0 ifTrue: [bar value: n].
							newCodeString := cls prettyPrinterClass 
										format: (cls sourceCodeAt: selector)
										in: cls
										notifying: nil.
							methodNode := cls compilerClass new 
										compile: newCodeString
										in: cls
										notifying: nil
										ifFail: [].
							newMethod := methodNode generate.
							oldMethod = newMethod 
								ifFalse: 
									[Transcript
										cr;
										show: '***' , cls name , ' ' , selector.
									badOnes add: cls name , ' ' , selector]]]].
	self systemNavigation browseMessageList: badOnes asSortedCollection
		name: 'Formatter Discrepancies'! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'nice 1/5/2010 15:59'!
testFormatter2
	"Smalltalk testFormatter2"

	"Reformats the source for every method in the system, and
	then verifies that the order of source tokens is unchanged."

	| badOnes |
	badOnes := OrderedCollection new.
	'Formatting all classes...' 
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: 
			[:bar | | n | 
			n := 0.
			self systemNavigation allBehaviorsDo: 
					[:cls | 
					"Transcript cr; show: cls name."

					cls selectorsDo: 
							[:selector | | newCodeString oldTokens oldCodeString newTokens | 
							(n := n + 1) \\ 100 = 0 ifTrue: [bar value: n].
							oldCodeString := (cls sourceCodeAt: selector) asString.
							newCodeString := cls prettyPrinterClass 
										format: oldCodeString
										in: cls
										notifying: nil.
							oldTokens := oldCodeString findTokens: Character separators.
							newTokens := newCodeString findTokens: Character separators.
							oldTokens = newTokens 
								ifFalse: 
									[Transcript
										cr;
										show: '***' , cls name , ' ' , selector.
									badOnes add: cls name , ' ' , selector]]]].
	self systemNavigation browseMessageList: badOnes asSortedCollection
		name: 'Formatter Discrepancies'! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'sd 7/2/2003 21:45'!
textMarkerForShortReport

	^  'Since last view	'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'StephaneDucasse 3/17/2010 21:02'!
timeStamp: aStream 
	"Writes system version and current time on stream aStream."

	| dateTime |
	dateTime := Time dateAndTimeNow.
	aStream nextPutAll: 'From ', self datedVersion, ' [', self  lastUpdateString, '] on ', (dateTime at: 1) printString,
						' at ', (dateTime at: 2) printString! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: 'StephaneDucasse 3/22/2010 16:56'!
traitNames
	"Answer a SortedCollection of all traits (not including class-traits) names."
	
	^ globals traitNames! !

!SmalltalkImage methodsFor: 'external' stamp: 'sd 6/28/2003 17:38'!
unbindExternalPrimitives
	"Primitive. Force all external primitives to be looked up again afterwards. Since external primitives that have not found are bound for fast failure this method will force the lookup of all primitives again so that after adding some plugin the primitives may be found."
	<primitive: 570>
	"Do nothing if the primitive fails for compatibility with older VMs"
! !

!SmalltalkImage methodsFor: 'modules' stamp: 'sd 6/27/2003 23:49'!
unloadModule: aString
	"Primitive. Unload the given module.
	This primitive is intended for development only since some
	platform do not implement unloading of DLL's accordingly.
	Also, the mechanism for unloading may not be supported
	on all platforms."
	<primitive: 571>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: 'JMM 6/6/2000 20:40'!
unregisterExternalObject: anObject
	"Unregister the given object in the external objects array. Do nothing if it isn't registered."

	ExternalSemaphoreTable unregisterExternalObject: anObject! !

!SmalltalkImage methodsFor: 'shrinking' stamp: 'sd 4/29/2003 19:06'!
unusedClasses
	"Enumerates all classes in the system and returns a list of those that are 
	apparently unused. A class is considered in use if it (a) has subclasses 
	or (b) is referred to by some method or (c) has its name in use as a 
	literal. "
	"Smalltalk unusedClasses asSortedCollection"
	^ self systemNavigation allUnusedClassesWithout: {{}. {}}! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'apb 10/3/2000 16:40'!
useUpMemory
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemory"

	| lst |
	lst := nil.
	[true] whileTrue: [
		lst := Link nextLink: lst.
	].! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'di 8/18/2000 21:15'!
useUpMemoryWithArrays 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithArrays"

	| b |  "First use up most of memory."
	b := String new: self bytesLeft - self lowSpaceThreshold - 100000.
	b := b.  "Avoid unused value warning"
	(1 to: 10000) collect: [:i | Array new: 10000]! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'di 8/18/2000 16:49'!
useUpMemoryWithContexts 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithContexts"

	self useUpMemoryWithContexts! !

!SmalltalkImage methodsFor: 'memory space' stamp: 'di 8/18/2000 16:50'!
useUpMemoryWithTinyObjects 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithTinyObjects"

	| b |  "First use up most of memory."
	b := String new: self bytesLeft - self lowSpaceThreshold - 100000.
	b := b.  "Avoid unused value warning"
	(1 to: 10000) collect: [:i | BitBlt new]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'sd 4/17/2003 21:01'!
verifyChanges		
	"Smalltalk verifyChanges"
	"Recompile all methods in the changes file."
	self systemNavigation allBehaviorsDo: [:class | class recompileChanges].
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'md 5/16/2006 12:34'!
version
	"Answer the version of this release."

	^SystemVersion current version! !

!SmalltalkImage methodsFor: 'vm' stamp: 'dtl 4/11/2010 11:39'!
versionLabel
	"Answer a string corresponding to the version of virtual machine. This
	represents the version level of the Smalltalk source code (interpreter
	and various plugins) that is translated to C by a CCodeGenerator,  in
	addition to the external platform source code, typically written in C and
	managed separately for each platform.
	
	This invokes an optional primitive that may not be available on all virtual
	machines. See also vmVersion, which answers a string identifying the image
	from which virtual machine sources were generated."

	"Smalltalk vm versionLabel"

	<primitive: 'primitiveVMVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveVMVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'copying' stamp: 'tk 10/20/2000 11:35'!
veryDeepCopyWith: deepCopier
	"Return self.  I can't be copied.  Do not record me."! !

!SmalltalkImage methodsFor: 'vm' stamp: 'dtl 4/11/2010 11:15'!
vm
	"Answer the object to query about virtual machine."
	
	^self! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'StephaneDucasse 4/23/2010 16:18'!
vmFullPath
	"Return the full vm path as a string"
	"SmalltalkImage current vmFullPath"

	^self getSystemAttribute: 0! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'tpr 4/27/2005 11:03'!
vmParameterAt: parameterIndex
	"parameterIndex is a positive integer corresponding to one of the VM's internal
	parameter/metric registers.  Answer with the current value of that register.
	Fail if parameterIndex has no corresponding register.
	VM parameters are numbered as follows:
		1	end of old-space (0-based, read-only)
		2	end of young-space (read-only)
		3	end of memory (read-only)
		4	allocationCount (read-only)
		5	allocations between GCs (read-write)
		6	survivor count tenuring threshold (read-write)
		7	full GCs since startup (read-only)
		8	total milliseconds in full GCs since startup (read-only)
		9	incremental GCs since startup (read-only)
		10	total milliseconds in incremental GCs since startup (read-only)
		11	tenures of surving objects since startup (read-only)
		12-20 specific to the translating VM
		21	root table size (read-only)
		22	root table overflows since startup (read-only)
		23	bytes of extra memory to reserve for VM buffers, plugins, etc.

		24	memory threshold above which shrinking object memory (rw)
		25	memory headroom when growing object memory (rw)
		26  interruptChecksEveryNms - force an ioProcessEvents every N milliseconds, in case the image  is not calling getNextEvent often (rw)
		27	number of times mark loop iterated for current IGC/FGC (read-only) includes ALL marking
		28	number of times sweep loop iterated  for current IGC/FGC (read-only)
		29	number of times make forward loop iterated for current IGC/FGC (read-only)
		30	number of times compact move loop iterated for current IGC/FGC (read-only)
		31	number of grow memory requests (read-only)
		32	number of shrink memory requests (read-only)
		33	number of root table entries used for current IGC/FGC (read-only)
		34	number of allocations done before current IGC/FGC (read-only)
		35	number of survivor objects after current IGC/FGC (read-only)
		36  millisecond clock when current IGC/FGC completed (read-only)
		37  number of marked objects for Roots of the world, not including Root Table entries for current IGC/FGC (read-only)
		38  milliseconds taken by current IGC  (read-only)
		39  Number of finalization signals for Weak Objects pending when current IGC/FGC completed (read-only)
		40  VM word size - 4 or 8 (read-only)"

	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'sd 6/27/2003 23:27'!
vmParameterAt: parameterIndex put: newValue
	"parameterIndex is a positive integer corresponding to one of the VM's internal
	parameter/metric registers.  Store newValue (a positive integer) into that
	register and answer with the previous value that was stored there.
	Fail if newValue is out of range, if parameterIndex has no corresponding
	register, or if the corresponding register is read-only."

	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'yo 7/2/2004 13:36'!
vmPath
	"Answer the path for the directory containing the Smalltalk virtual machine. Return the 	empty string if this primitive is not implemented."
	"SmalltalkImage current vmPath"

	^ (FilePath pathName: (self primVmPath) isEncoded: true) asSqueakPathName.
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'StephaneDucasse 7/9/2010 23:13'!
vmStatisticsReportString
	"StringHolderView open: (StringHolder new contents:
		SmalltalkImage current vmStatisticsReportString) label: 'VM Statistics'"

	| params oldSpaceEnd youngSpaceEnd memoryEnd fullGCs fullGCTime incrGCs incrGCTime tenureCount mcMisses mcHits icHits upTime sendCount upTime2 fullGCs2 fullGCTime2 incrGCs2 incrGCTime2 tenureCount2 str |
	params := self getVMParameters.
	oldSpaceEnd			:= params at: 1.
	youngSpaceEnd		:= params at: 2.
	memoryEnd			:= params at: 3.
	fullGCs				:= params at: 7.
	fullGCTime			:= params at: 8.
	incrGCs				:= params at: 9.
	incrGCTime			:= params at: 10.
	tenureCount			:= params at: 11.
	upTime := Time millisecondClockValue.
	str := (String new: 1000) writeStream.
	str	nextPutAll: 'uptime			';
		print: (upTime / 1000 / 60 // 60); nextPut: $h;
		print: (upTime / 1000 / 60 \\ 60) asInteger; nextPut: $m;
		print: (upTime / 1000 \\ 60) asInteger; nextPut: $s; cr.

	str	nextPutAll: 'memory			';
		nextPutAll: memoryEnd asStringWithCommas; nextPutAll: ' bytes'; cr.
	str	nextPutAll:	'	old			';
		nextPutAll: oldSpaceEnd asStringWithCommas; nextPutAll: ' bytes (';
		print: ((oldSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.
	str	nextPutAll: '	young		';
		nextPutAll: (youngSpaceEnd - oldSpaceEnd) asStringWithCommas; nextPutAll: ' bytes (';
		print: ((youngSpaceEnd - oldSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.
	str	nextPutAll: '	used		';
		nextPutAll: youngSpaceEnd asStringWithCommas; nextPutAll: ' bytes (';
		print: ((youngSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.
	str	nextPutAll: '	free		';
		nextPutAll: (memoryEnd - youngSpaceEnd) asStringWithCommas; nextPutAll: ' bytes (';
		print: ((memoryEnd - youngSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.

	str	nextPutAll: 'GCs				';
		nextPutAll: (fullGCs + incrGCs) asStringWithCommas.
	fullGCs + incrGCs > 0 ifTrue: [
		str
			nextPutAll: ' ('; 
			print: ((upTime / (fullGCs + incrGCs)) roundTo: 1); 
			nextPutAll: 'ms between GCs)'
	].
	str cr.
	str	nextPutAll: '	full			';
		print: fullGCs; nextPutAll: ' totalling '; nextPutAll: fullGCTime asStringWithCommas; nextPutAll: 'ms (';
		print: ((fullGCTime / upTime * 100) roundTo: 0.1);
		nextPutAll: '% uptime)'.
	fullGCs = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: ((fullGCTime / fullGCs) roundTo: 0.1); nextPutAll: 'ms'].
	str	cr.
	str	nextPutAll: '	incr		';
		print: incrGCs; nextPutAll: ' totalling '; nextPutAll: incrGCTime asStringWithCommas; nextPutAll: 'ms (';
		print: ((incrGCTime / upTime * 100) roundTo: 0.1);
		nextPutAll: '% uptime), avg '; print: ((incrGCTime / incrGCs) roundTo: 0.1); nextPutAll: 'ms'; cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount asStringWithCommas.
	tenureCount = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: (incrGCs / tenureCount) asInteger; nextPutAll: ' GCs/tenure)'].
	str	cr.

LastStats ifNil: [LastStats := Array new: 6]
ifNotNil: [
	upTime2 := upTime - (LastStats at: 1).
	fullGCs2 := fullGCs - (LastStats at: 2).
	fullGCTime2 := fullGCTime - (LastStats at: 3).
	incrGCs2 := incrGCs - (LastStats at: 4).
	incrGCTime2 := incrGCTime - (LastStats at: 5).
	tenureCount2 := tenureCount - (LastStats at: 6).

	str	nextPutAll: self textMarkerForShortReport ;
		nextPutAll: (fullGCs2 + incrGCs2) asStringWithCommas.
	fullGCs2 + incrGCs2 > 0 ifTrue: [
		str
			nextPutAll: ' ('; 
			print: ((upTime2 / (fullGCs2 + incrGCs2)) roundTo: 1); 
			nextPutAll: 'ms between GCs)'.
	].
	str cr.
	str	nextPutAll: '	uptime		'; print: ((upTime2 / 1000.0) roundTo: 0.1); nextPutAll: 's'; cr.
	str	nextPutAll: '	full			';
		print: fullGCs2; nextPutAll: ' totalling '; nextPutAll: fullGCTime2 asStringWithCommas; nextPutAll: 'ms (';
		print: ((fullGCTime2 / upTime2 * 100) roundTo: 0.1);
		nextPutAll: '% uptime)'.
	fullGCs2 = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: ((fullGCTime2 / fullGCs2) roundTo: 0.1); nextPutAll: 'ms'].
	str	cr.
	str	nextPutAll: '	incr		';
		print: incrGCs2; nextPutAll: ' totalling '; nextPutAll: incrGCTime2 asStringWithCommas; nextPutAll: 'ms (';
		print: ((incrGCTime2 / upTime2 * 100) roundTo: 0.1);
		nextPutAll: '% uptime), avg '.
	incrGCs2 > 0 ifTrue: [
		 str print: ((incrGCTime2 / incrGCs2) roundTo: 0.1); nextPutAll: 'ms'
	].
	str cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount2 asStringWithCommas.
	tenureCount2 = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: (incrGCs2 / tenureCount2) asInteger; nextPutAll: ' GCs/tenure)'].
	str	cr.
].
	LastStats at: 1 put: upTime.
	LastStats at: 2 put: fullGCs.
	LastStats at: 3 put: fullGCTime.
	LastStats at: 4 put: incrGCs.
	LastStats at: 5 put: incrGCTime.
	LastStats at: 6 put: tenureCount.

	^ str contents
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'HenrikSperreJohansen 6/12/2010 02:41'!
vmStatisticsShortString
	"Convenience item for access to recent statistics only"
	"StringHolderView open: (StringHolder new contents: SmalltalkImage current vmStatisticsShortString)
		label: 'VM Recent Statistics'"
	^ self vmStatisticsReportString readStream
		upToAll: 'Since';
		nextLine;
		upToEnd! !

!SmalltalkImage methodsFor: 'system attribute' stamp: 'dtl 8/5/2006 09:47'!
vmVersion	
	"Return a string identifying the interpreter version"
	"SmalltalkImage current vmVersion"

	^self getSystemAttribute: 1004! !

!SmalltalkImage methodsFor: 'os - deprecated' stamp: 'dfsdfsd 6/11/2010 15:50'!
windowSystemName			
	"Return the name of the window system currently being used for display."
	"Smalltalk os windowSystemName"

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform windowSystemName ! !

!SmalltalkImage methodsFor: 'image' stamp: 'dtl 1/4/2010 21:40'!
wordSize
	"Answer the size in bytes of an object pointer or word in the object memory.
	The value does not change for a given image, but may be modified by a SystemTracer
	when converting the image to another format. The value is cached in WordSize to
	avoid the performance overhead of repeatedly consulting the VM."

	"Smalltalk wordSize"

	^ WordSize ifNil: [WordSize := [SmalltalkImage current vmParameterAt: 40] on: Error do: [4]]! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'adrian-lienhard 5/27/2009 21:28'!
writeRecentCharacters: nCharacters toFileNamed: aFilename
	"Schedule an editable text view on the last n characters of changes."
	| changes |
	changes := SourceFiles at: 2.
	changes setToEnd; skip: nCharacters negated.
	(FileStream newFileNamed: aFilename) nextPutAll: (changes next: nCharacters); close; open; edit! !

Object subclass: #SystemNavigation
	instanceVariableNames: 'browserClass hierarchyBrowserClass'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Support'!

!SystemNavigation commentStamp: 'sd 4/15/2003 22:30' prior: 0!
I support the navigation of the system. I act as a facade but as I could require some state
or different way of navigating the system all my behavior are on the instance side!

!SystemNavigation class methodsFor: 'accessing' stamp: 'marcus.denker 9/14/2008 21:05'!
default
	Default ifNil: [Default := self new].
	^Default! !

!SystemNavigation methodsFor: 'message sends' stamp: 'LaurentLaffont 6/27/2010 10:34'!
addSelectorsReferingTo: aSymbol in: class to: sortedSenders special: special byte: byte 
	{class. class class} do: [:behavior| (behavior 
		thoroughWhichSelectorsReferTo: aSymbol
		special: special
		byte: byte) do: [ :sel | 
			sortedSenders add: (MethodReference 
				class: behavior
				selector: sel) ]]! !

!SystemNavigation methodsFor: 'query' stamp: 'StephaneDucasse 2/13/2010 15:32'!
allBehaviorsDo: aBlock 
	"Evaluate the argument, aBlock, for each kind of Behavior in the system 
	(that is, Object and its subclasses and Traits).
	ar 7/15/1999: The code below will not enumerate any obsolete or anonymous behaviors for which the following should be executed:

		Smalltalk allObjectsDo:[:obj| obj isBehavior ifTrue:[aBlock value: obj]].

	but what follows is way faster than enumerating all objects."

	aBlock value: ProtoObject.
	ProtoObject allSubclassesDo:  aBlock.		

	"Classes outside the ProtoObject hierarchy"
	Class subclassesDo: [:aClass |
		(aClass == ProtoObject class
			or: [aClass isMeta not]) ifFalse:
			["Enumerate the non-meta class and its subclasses"
			aBlock value: aClass soleInstance.
			aClass soleInstance allSubclassesDo: aBlock]].
	
	Trait allInstances , ClassTrait allInstances do: [:trait |
		aBlock value: trait]! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:08'!
allCallsOn: aSymbol 
	"Answer a Collection of all the methods that call on the provided symbol even deeply embedded in literal array."
	"self new allCallsOn: #allCallsOn:"
	
	^ self allReferencesTo: aSymbol! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:09'!
allCallsOn: firstLiteral and: secondLiteral
	"Answer a SortedCollection of all the methods that call on both aLiteral 
	and secondLiteral."

	| aCollection firstSpecial secondSpecial firstByte secondByte |
	self flag: #ShouldUseAllCallsOn:. "sd"
	aCollection := SortedCollection new.
	firstSpecial := Smalltalk hasSpecialSelector: firstLiteral ifTrueSetByte: [:b | firstByte := b].
	secondSpecial := Smalltalk hasSpecialSelector: secondLiteral ifTrueSetByte: [:b | secondByte := b].
	self allBehaviorsDo: [:class | | secondArray |
			secondArray := class whichSelectorsReferTo: secondLiteral special: secondSpecial byte: secondByte.
			((class whichSelectorsReferTo: firstLiteral special: firstSpecial byte: firstByte) select: [:aSel | 
				(secondArray includes: aSel)]) do: [:sel | 
					aCollection add: (MethodReference class: class selector: sel)]].
	^aCollection! !

!SystemNavigation methodsFor: 'query' stamp: 'StephaneDucasse 3/13/2010 15:22'!
allCallsOn: aSymbol from: aClass
	"Answer a SortedCollection of all the methods that call on aSymbol."

	| aSortedCollection special byte |
	aSortedCollection := SortedCollection new.
	special := Smalltalk hasSpecialSelector: aSymbol ifTrueSetByte: [:b | byte := b ].
	aClass withAllSubclassesDo: [ :class |
		(class whichSelectorsReferTo: aSymbol special: special byte: byte) do: [:sel |
				aSortedCollection add: (MethodReference class: class selector: sel)]].
	^aSortedCollection! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 8/30/2010 10:42'!
allCallsOn: aSymbol localTo: aClass
	"Answer a Set of MethodReferences for all the methods
	 that call on aSymbol in, above or below the given class."

	| aSet special byte enum |
	aSet := Set new.
	special := Smalltalk hasSpecialSelector: aSymbol ifTrueSetByte: [:b | byte := b].
	enum := [:behavior|
			(behavior whichSelectorsReferTo: aSymbol special: special byte: byte) do:
				[:sel | aSet add: (MethodReference new setStandardClass: behavior  methodSymbol: sel)]].
	aClass theNonMetaClass withAllSuperAndSubclassesDoGently: enum.
		aClass theNonMetaClass class withAllSuperAndSubclassesDoGently: enum.
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 8/30/2010 10:47'!
allCallsOn: aLiteral localToPackage: packageNameOrInfo
	"Answer a Set of MethodReferences for all the methods
	 that call on aSymbol in the given package."

	| aSet special byte |
	aSet := Set new.
	special := Smalltalk hasSpecialSelector: aLiteral ifTrueSetByte: [:b | byte := b].
	Cursor wait showWhile:
		[(self packageInfoFor: packageNameOrInfo) actualMethodsDo:
			[:method |
			((method hasLiteral: aLiteral) or: [special and: [method scanFor: byte]]) ifTrue:
				[((aLiteral isVariableBinding) not
					or: [method literals allButLast includes: aLiteral])
						ifTrue: [aSet add: method methodReference]]].].
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: 'StephaneDucasse 3/20/2010 13:55'!
allClasses
	"currently returns all the classes defined in Smalltalk but could be customized 
	for dealing with environments and in such a case would return on really all the classes"

	^ Smalltalk globals allClasses

	! !

!SystemNavigation methodsFor: 'query' stamp: 'StephaneDucasse 3/20/2010 13:59'!
allClassesAndTraits
	
	^ Smalltalk globals allClassesAndTraits

	! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/17/2003 21:31'!
allClassesDo: aBlock
	"currently returns all the classes defined in Smalltalk but could be customized 
	for dealing with environments and  in such a case would work on really all the classes"

	^ Smalltalk allClassesDo: aBlock

	! !

!SystemNavigation methodsFor: 'query' stamp: 'marcus.denker 9/29/2008 08:27'!
allClassesImplementing: aSelector  
	"Answer an Array of all classes that implement the message aSelector."

	| aCollection |
	aCollection := ReadWriteStream on: Array new.
	self allBehaviorsDo: [:class | (class includesSelector: aSelector) ifTrue: [aCollection nextPut: class]].
	^ aCollection contents! !

!SystemNavigation methodsFor: 'packages' stamp: 'StephaneDucasse 10/17/2010 20:49'!
allClassesInPackageNamed: packageName
	"Returns the list of classes that belongs to a Monticello package."
	
	"It would be probably better to delegate that query to the package itself"
	| classes organization |
	organization := Smalltalk globals organization.
	classes := (organization classesInCategory: packageName) asSet.
	(organization categoriesMatching: packageName, '-*') do: [:category| 
		classes addAll: (organization classesInCategory: category)].
	^classes asArray! !

!SystemNavigation methodsFor: 'query' stamp: 'nice 1/16/2010 13:30'!
allClassesWithUnimplementedCalls
	"Answer an Array of classes that have messages with calls to methods that aren't implemented
	anywhere in the system"
	| all dict |
	dict := Dictionary new.
	all := self systemNavigation allImplementedMessages.
	self systemNavigation allBehaviorsDo: [:cl | 
		cl selectorsAndMethodsDo: [:sel :meth | 
			meth primitive = 0 ifTrue: [
				meth messages do: [:m | 
					(all includes: m) ifFalse: [
						((dict at: cl ifAbsentPut: [ Dictionary new ])
							at: sel ifAbsentPut: [ OrderedCollection new])
								add: m
					]
				]
			]
		]
	].
	^ dict! !

!SystemNavigation methodsFor: 'author' stamp: 'nice 1/5/2010 15:59'!
allContributors
	"SystemNavigation default allContributors"
	
	| bag |
	bag := Bag new.
	self allBehaviorsDo: [ :behavior |
		behavior methodsDo: [ :compiledMethod | | stamp |
			stamp := compiledMethod timeStamp.
			stamp notEmpty ifTrue: [
				bag add: compiledMethod timeStamp substrings first ]]].
	^bag! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/18/2003 10:04'!
allGlobalRefs
	"Answer a set of symbols that may be refs to Global names.  In some sense we should only need the associations, but this will also catch, eg, HTML tag types."

	^ self allGlobalRefsWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'lr 3/14/2010 21:13'!
allGlobalRefsWithout: classesAndMessagesPair
	"Answer a set of symbols that may be refs to Global names. In some  
	sense we should only need the associations, but this will also catch, eg,  
	HTML tag types. This method computes its result in the absence of  
	specified classes and messages."

	"may be a problem if namespaces are introduced as for the moment  
	only Smalltalk is queried. sd 29/4/03"

	| globalRefs absentClasses absentSelectors |
	globalRefs := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	absentSelectors := classesAndMessagesPair second.
	self flag: #shouldBeRewrittenUsingSmalltalkAllClassesDo:.	"sd 29/04/03"
	Cursor execute
		showWhile: [ 
			Smalltalk classNames
				do: [ :cName | 
					((absentClasses includes: cName)
						ifTrue: [ {} ]
						ifFalse: [ 
							{(Smalltalk globals at: cName).
							((Smalltalk globals at: cName) class)} ])
						do: [ :cl | 
							(absentSelectors isEmpty
								ifTrue: [ cl selectors ]
								ifFalse: [ cl selectors copyWithoutAll: absentSelectors ])
								do: [ :sel | 
									"Include all capitalized symbols for good 
									measure"
									(cl compiledMethodAt: sel)
										literalsDo: [ :m | 
											(m isSymbol and: [ m size > 0 and: [ m first canBeGlobalVarInitial ] ])
												ifTrue: [ globalRefs add: m ].
											(m isMemberOf: Array)
												ifTrue: [ 
													m
														do: [ :x | 
															(x isSymbol and: [ x size > 0 and: [ x first canBeGlobalVarInitial ] ])
																ifTrue: [ globalRefs add: x ] ] ].
											m isVariableBinding
												ifTrue: [ m key ifNotNil: [ globalRefs add: m key ] ] ] ] ] ] ].
	^ globalRefs! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 13:30'!
allImplementedMessages
	"Answer a Set of all the messages that are implemented in the system."
	^ self allImplementedMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:09'!
allImplementedMessagesWithout: classesAndMessagesPair
	"Answer a Set of all the messages that are implemented in the system,  
	computed in the absence of the supplied classes and messages. Note this  
	reports messages that are in the absent selectors set."

	| messages absentClasses |
	messages := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	self flag: #shouldBeRewrittenUsingSmalltalkAllClassesDo:.	"sd 29/04/03"
	Smalltalk classNames , Smalltalk traitNames
		do: [ :name | 
			((absentClasses includes: name)
				ifTrue: [ {} ]
				ifFalse: [ 
					{(Smalltalk globals at: name).
					((Smalltalk globals at: name) classSide)} ]) do: [ :each | messages addAll: each selectors ] ].
	^ messages! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:09'!
allImplementorsOf: aSelector 
	"Answer a SortedCollection of all the methods that implement the message 
	aSelector."
	| aCollection |
	aCollection := SortedCollection new.
	self allBehaviorsDo: [:class | 
		(class includesSelector: aSelector)
			ifTrue: [aCollection add: (MethodReference new setStandardClass: class methodSymbol: aSelector)]].
	^ aCollection! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:09'!
allImplementorsOf: aSelector localTo: aClass 
	"Answer a SortedCollection of all the methods that implement the message  
	aSelector in, above, or below the given class."
	| cls aCollection |
	aCollection := SortedCollection new.
	cls := aClass theNonMetaClass.
	cls
				withAllSuperAndSubclassesDo: [:class | (class includesSelector: aSelector)
						ifTrue: [aCollection
								add: (MethodReference new setStandardClass: class methodSymbol: aSelector)]].
			cls class
				withAllSuperAndSubclassesDo: [:class | (class includesSelector: aSelector)
						ifTrue: [aCollection
								add: (MethodReference new setStandardClass: class methodSymbol: aSelector)]].
	^ aCollection! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 8/30/2010 10:33'!
allImplementorsOf: aSelector  localToPackage: packageNameOrInfo
	"Answer a SortedCollection of all the methods that implement the message 
	 aSelector in the given package."

	| aSet |
	aSet := Set new.
	Cursor wait showWhile:
		[(self packageInfoFor: packageNameOrInfo) actualMethodsDo:
			[:m |
			(m selector = aSelector) ifTrue:
				[aSet add: m methodReference]]].
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: 'StephaneDucasse 3/20/2010 22:07'!
allLocalCallsOn: aSymbol ofClass: aClass
	"Answer a SortedCollection of all the methods that call on aSymbol, anywhere in my class hierarchy."

	| aSet special byte cls |
	aSet := Set new.
	cls := aClass theNonMetaClass.
	special := Smalltalk hasSpecialSelector: aSymbol
					ifTrueSetByte: [:b | byte := b ].
	cls withAllSuperAndSubclassesDo: [ :class |
		(class whichSelectorsReferTo: aSymbol special: special byte: byte)
			do: [:sel |
				aSet add: (MethodReference new
							setClass: class  			
							methodSymbol: sel 
							stringVersion: (class name asString , ' ', sel asString)).]].
		
	cls class withAllSuperAndSubclassesDo: [ :class |
		(class whichSelectorsReferTo: aSymbol special: special byte: byte)
			do: [:sel |
				aSet add: (MethodReference new
							setClass: class  			
							methodSymbol: sel 
							stringVersion: (class name asString , ' ', sel asString)).]].
	^aSet! !

!SystemNavigation methodsFor: 'accessing' stamp: 'marcus.denker 7/24/2009 16:47'!
allMethodsInCategory: category 
	| aCollection |
	aCollection := Set new.
	Cursor wait showWhile: [
			self allBehaviorsDo: [:x | ((category = ClassOrganizer allCategory
					ifTrue: [x organization allMethodSelectors]
					ifFalse: [x organization listAtCategoryNamed: category])) do:
				[:sel | aCollection add: (MethodReference new setStandardClass: x 
methodSymbol: sel)]]].
	^aCollection.
! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:10'!
allMethodsSelect: aBlock 
	"Answer a SortedCollection of each method that, when used as the block  
	argument to aBlock, gives a true result."
	| aCollection |
	aCollection := SortedCollection new.
	self allBehaviorsDo: [:class | class
						selectorsAndMethodsDo: [:sel :m | (aBlock value: m)
								ifTrue: [aCollection
										add: (MethodReference new setStandardClass: class methodSymbol: sel)]]].
	^ aCollection! !

!SystemNavigation methodsFor: 'query' stamp: 'eem 8/30/2010 10:38'!
allMethodsSelect: aBlock localToPackage: packageNameOrInfo
	"Answer a SortedCollection of each method in the given package
	 for which the evaluation of aBlock with the metnod answers true."

	| aSet |
	aSet := Set new.
	Cursor wait showWhile:
		[(self packageInfoFor: packageNameOrInfo) actualMethodsDo:
			[:aMethod |
			(aBlock value: aMethod) ifTrue:
				[aSet add: aMethod methodReference]]].
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: 'MarcusDenker 3/26/2010 16:59'!
allMethodsWithSourceString: aString matchCase: caseSensitive
	"Answer a SortedCollection of all the methods that contain, in source code, aString as a substring.  Search the class comments also"

	| list adder |
	list := Set new.
	adder := [ :mrClass :mrSel | list add: ( MethodReference new
											setStandardClass: mrClass
											methodSymbol: mrSel)].
	'Searching all source code...'
		displayProgressAt: Sensor cursorPoint
		from: 0 to: Smalltalk globals classNames size
		during: [:bar | | count |
			count := 0.
			SystemNavigation default allBehaviorsDo: [:each |
				bar value: (count := count + 1).
					each selectorsDo: [:sel | 
						((each sourceCodeAt: sel) findString: aString 
							startingAt: 1 caseSensitive: caseSensitive) > 0 ifTrue: [
								sel isDoIt ifFalse: [adder value: each value: sel]]].
					(each organization classComment asString findString: aString 
							startingAt: 1 caseSensitive: caseSensitive) > 0 ifTrue: [
								adder value: each value: #Comment]	]].
			^ list asSortedCollection! !

!SystemNavigation methodsFor: 'query' stamp: 'ar 8/18/2008 18:08'!
allObjectsDo: aBlock 
	"Evaluate the argument, aBlock, for each object in the system 
	excluding SmallIntegers."
	| object endMarker |
	object := self someObject.
	endMarker := Object new.
	[endMarker == object]
		whileFalse: [aBlock value: object.
			object := object nextObject]! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 20:53'!
allObjectsSelect: aBlock 
	"Evaluate the argument, aBlock, for each object in the system excluding 
	SmallIntegers. Return a collection af all objects for whom the value is 
	true. "
	^ Array
		streamContents: [:s | self
				allObjectsDo: [:object | (aBlock value: object)
						ifTrue: [s nextPut: object]]]! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:10'!
allPrimitiveMethods
	"Answer an OrderedCollection of all the methods that are implemented by primitives."
	| aColl |
	aColl := OrderedCollection new: 200.
	self allBehaviorsDo: [:class | class
		selectorsAndMethodsDo: [:sel :method | 
			method primitive ~= 0
				ifTrue: [aColl addLast: class name , ' ' , sel , ' ' , method primitive printString]]].
	^ aColl! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:11'!
allPrimitiveMethodsInCategories: aList 
	"Answer an OrderedCollection of all the methods that are implemented by 
	primitives in the given categories. 1/26/96 sw"
	"SystemNavigation new allPrimitiveMethodsInCategories:  
	#('Collections-Streams' 'Files-Streams' 'Files-Abstract' 'Files-Macintosh')"

	| aColl |
	aColl := OrderedCollection new: 200.
	self
		allBehaviorsDo: [:aClass | (aList includes: (SystemOrganization categoryOfElement: aClass theNonMetaClass name asString) asString)
				ifTrue: [aClass
						selectorsAndMethodsDo: [:sel :method | 
							method primitive ~= 0
								ifTrue: [aColl addLast: aClass name , ' ' , sel , ' ' , method primitive printString]]]].
	^ aColl! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:06'!
allReferencesTo: aLiteral 
	"Answer a Collection of all the methods that refer to aLiteral even deeply embedded in literal array."

	"self new allReferencesTo: #+"
	"self new allReferencesTo: Point binding"
	
	| result special byte symbol |
	result := OrderedCollection new.
	special := Smalltalk
		hasSpecialSelector: aLiteral
		ifTrueSetByte: [ :value | byte := value ].
	symbol := aLiteral isSymbol.
	self allBehaviorsDo: [ :class |
		| selectors |
		selectors := symbol
			ifFalse: [ class whichSelectorsReferTo: aLiteral special: special byte: byte ]
			ifTrue: [ class thoroughWhichSelectorsReferTo: aLiteral special: special byte: byte ].
		selectors do: [ :selector |
			result addLast: (MethodReference class: class selector: selector) ] ].
	^ result! !

!SystemNavigation methodsFor: 'query' stamp: 'nice 1/5/2010 15:59'!
allReferencesToPool: aPool from: aClass
	"Answer all the references to variables from aPool"
	| list |
	list := OrderedCollection new.
	aClass withAllSubclassesDo:[:cls|
		cls selectorsAndMethodsDo:[:sel :meth| | ref |
			ref := meth literals detect:[:lit|
				lit isVariableBinding and:[(aPool bindingOf: lit key) notNil]
			] ifNone:[nil].
			ref ifNotNil:[
				list add:(MethodReference new setStandardClass: cls methodSymbol: sel)
			].
		].
	].
	^list! !

!SystemNavigation methodsFor: 'query' stamp: 'MarcusDenker 3/5/2010 14:36'!
allSelectorsWithAnyImplementorsIn: selectorList 
	"Answer the subset of the given list which represent method selectors 
	which have at least one implementor in the system."
	| good |
	good := OrderedCollection new.
	self allBehaviorsDo: [:class | 
			selectorList do: [:aSelector | (class includesSelector: aSelector) ifTrue: [good add: aSelector]]].
	^ good asSet asArray sort! !

!SystemNavigation methodsFor: 'message sends' stamp: 'LaurentLaffont 6/27/2010 10:22'!
allSendersOf: selector 
	|  sortedSenders special byte |
	sortedSenders := SortedCollection new.
	special := Smalltalk 
				hasSpecialSelector: selector
				ifTrueSetByte: [ :b | byte := b ].
	self allBehaviorsDo: 
		[ :behavior | 
		self 
			addSelectorsReferingTo: selector
			in: behavior
			to: sortedSenders
			special: special
			byte: byte ].
	^ sortedSenders! !

!SystemNavigation methodsFor: 'message sends' stamp: 'StephaneDucasse 3/20/2010 22:07'!
allSendersOf: selector inClass: aClass 
	| sortedSenders special byte |
	sortedSenders := SortedCollection new.
	special := Smalltalk hasSpecialSelector: selector ifTrueSetByte: [:b | byte := b ].
	self 
		addSelectorsReferingTo: selector
		in: aClass
		to: sortedSenders
		special: special
		byte: byte.
	^sortedSenders! !

!SystemNavigation methodsFor: 'message sends' stamp: 'StephaneDucasse 3/20/2010 22:07'!
allSendersOf: selector inClassCategory: category 
	| classes sortedSenders special byte |
	classes := SystemOrganization classesInCategory: category.
	sortedSenders := SortedCollection new.
	classes ifEmpty: [ ^ sortedSenders ].
	special := Smalltalk
				hasSpecialSelector: selector
				ifTrueSetByte: [ :b | byte := b ].
	classes do: 
		[ :class | 
		self 
			addSelectorsReferingTo: selector
			in: class
			to: sortedSenders
			special: special
			byte: byte ].
	^ sortedSenders! !

!SystemNavigation methodsFor: 'message sends' stamp: 'Noury 10/26/2008 17:12'!
allSendersOf: selector inPackageNamed: packageName
	^(self categoriesInPackageNamed: packageName) inject: SortedCollection new into: [:sortedSenders :category|
		sortedSenders, (self allSendersOf: selector inClassCategory: category)]! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 20:12'!
allSentMessages
	"Answer the set of selectors which are sent somewhere in the system."
	^ self  allSentMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:11'!
allSentMessagesWithout: classesAndMessagesPair 
	"Answer the set of selectors which are sent somewhere in the system,  
	computed in the absence of the supplied classes and messages."
	| sent absentClasses absentSelectors |
	sent := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	absentSelectors := classesAndMessagesPair second.
	self allBehaviorsDo: [:class |
		((absentClasses includes: class)
			ifTrue: [{}]
			ifFalse: [{class. class classSide}]) do: [:each | 
					(absentSelectors isEmpty ifTrue: [each selectors] ifFalse: [each selectors copyWithoutAll: absentSelectors])
					do: [:sel | "Include all sels, but not if sent by self"
						(each compiledMethodAt: sel) literalsDo: [:m | 
								(m isSymbol) ifTrue: ["might be sent"
										m == sel ifFalse: [sent add: m]].
								(m isMemberOf: Array) ifTrue: ["might be performed"
										m do: [:x | (x isSymbol) ifTrue: [x == sel ifFalse: [sent add: x]]]]]]]].
		"The following may be sent without being in any literal frame"
		1
			to: Smalltalk specialSelectorSize
			do: [:index | sent
					add: (Smalltalk specialSelectorAt: index)].
	^ sent! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/29/2003 19:19'!
allUnSentMessagesWithout: classesAndMessagesPair 
	"Answer the set of selectors that are implemented but not sent, computed  
	in the absence of the supplied classes and messages."
	^ (self  allImplementedMessagesWithout: classesAndMessagesPair)
		copyWithoutAll: (self  allSentMessagesWithout: classesAndMessagesPair)! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:11'!
allUnimplementedCalls
	"Answer an Array of each message that is sent by an expression in a  
	method but is not implemented by any object in the system."
	| aStream all |
	all := self allImplementedMessages.
	aStream := (Array new: 50) writeStream.
	self
		allBehaviorsDo: [:cl | cl
				selectorsAndMethodsDo: [:sel :meth |
					| secondStream | 
					secondStream := (String new: 5) writeStream.
					meth messages
						do: [:m | (all includes: m)
								ifFalse: [secondStream nextPutAll: m;
										 space]].
					secondStream position = 0
						ifFalse: [aStream nextPut: cl name , ' ' , sel , ' calls: ' , secondStream contents]]].
	^ aStream contents! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:11'!
allUnimplementedNonPrimitiveCalls
	"Answer an Array of each message that is sent by an expression in a  
	method but is not implemented by any object in the system."
	| aStream all |
	all := self systemNavigation allImplementedMessages.
	aStream := (Array new: 50) writeStream.
	self systemNavigation
		allBehaviorsDo: [:cl | cl
				selectorsAndMethodsDo: [:sel :meth |
					| secondStream | 
					secondStream := (String new: 5) writeStream.
					meth primitive = 0 ifTrue: [
						meth messages
							do: [:m | (all includes: m)
									ifFalse: [secondStream nextPutAll: m;
											 space]].
						secondStream position = 0
							ifFalse: [aStream nextPut: cl name , ' ' , sel , ' calls: ' , secondStream contents]]]].
	^ aStream contents! !

!SystemNavigation methodsFor: 'query' stamp: 'Alexandre Bergel 11/21/2009 12:02'!
allUnreferencedClassVariablesOf: aClass
	"Answer a list of the names of all the receiver's unreferenced class  
	vars, including those defined in superclasses"
	| aList |
	aList := OrderedCollection new.
	aClass withAllSuperclasses
		reverseDo: [:aSuperClass | aSuperClass classVarNames
				do: [:var | (self allReferencesTo: (aSuperClass classPool associationAt: var)) isEmpty
						ifTrue: [aList add: var]]].
	^ aList! !

!SystemNavigation methodsFor: 'query' stamp: 'marcus.denker 10/22/2008 14:38'!
allUnsentMessages
	"SystemNavigation new allUnSentMessages"
	"Answer the set of selectors that are implemented by some object in the  
	system but not sent by any."
	^ self allUnSentMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'MarcusDenker 2/16/2010 22:43'!
allUnsentMessagesIn: selectorSet 
	"Answer the subset of selectorSet which are not sent anywhere in the 
	system. As allSentMessages does not recurse into nested Arrays, filter
	with #allCallsOn, which does"
	
	^ (selectorSet copyWithoutAll: self allSentMessages) select: [:each | 			(SystemNavigation default allCallsOn: each) isEmpty]   ! !

!SystemNavigation methodsFor: 'query' stamp: 'MarcusDenker 3/26/2010 16:58'!
allUnusedClassesWithout: classesAndMessagesPair
	"Enumerates all classes in the system and returns a list of those that are 
	apparently unused. A class is considered in use if it (a) has subclasses  
	or (b) is referred to by some method or (c) has its name in use as a  
	literal."

	"SystemNavigation new unusedClasses"

	| unused |
	self flag: #rewriteMePlease.
	unused := Smalltalk globals classNames asIdentitySet copyWithoutAll: (self allGlobalRefsWithout: classesAndMessagesPair).
	^ unused
		reject: [ :cName | 
			| cl |
			cl := Smalltalk globals at: cName.
			cl subclasses isEmpty not or: [ cl inheritsFrom: FileDirectory ] ]! !

!SystemNavigation methodsFor: 'accessing' stamp: 'Noury 10/26/2008 17:12'!
categoriesInPackageNamed: packageName
	^(SystemOrganization categoriesMatching: packageName), (SystemOrganization categoriesMatching: packageName, '*')! !

!SystemNavigation methodsFor: 'author' stamp: 'nice 1/5/2010 15:59'!
contributionsOf: aString
	"SystemNavigation default contributionsOf: 'alain.plantec'"
	
	| answer |
	answer := OrderedCollection new.
	self allBehaviorsDo: [ :behavior |
		behavior methodsDo: [ :compiledMethod | | stamp initials |
			stamp := compiledMethod timeStamp.
			stamp notEmpty ifTrue: [
				initials := compiledMethod timeStamp substrings first.
				aString = initials ifTrue: [
					answer add: (compiledMethod selector -> compiledMethod methodClass)]]]].
	^answer! !

!SystemNavigation methodsFor: 'author' stamp: 'nice 1/5/2010 15:59'!
contributorsNotSignatories
	"SystemNavigation default contributorsNotSignatories"
	
	| signatories answer |
	answer := Dictionary new.
	signatories := self signatories.
	self allBehaviorsDo: [ :behavior |
		behavior methodsDo: [ :compiledMethod | | stamp initials |
			stamp := compiledMethod timeStamp.
			stamp notEmpty ifTrue: [
				initials :=  compiledMethod timeStamp substrings first.
				(signatories includes: initials) ifFalse: [
					(answer at: initials ifAbsentPut: [OrderedCollection new])
						add: (compiledMethod selector -> compiledMethod methodClass)]]]].
	^answer! !

!SystemNavigation methodsFor: 'private' stamp: 'eem 8/30/2010 11:02'!
headingAndAutoselectForLiteral: aLiteral do: binaryBlock
	"Evaluate aBlock with either Users of ... or Senders of ... plus the auto-select string
	 for the given literal.  aLiteral can be a Symbol, a VariableBinding or an arbitrary object."

	| autoSelect |
	^aLiteral isSymbol
		ifTrue: [binaryBlock value: 'Senders of ', aLiteral value: aLiteral keywords first]
		ifFalse:
			[autoSelect := aLiteral isVariableBinding
							ifTrue: [aLiteral key]
							ifFalse: [aLiteral printString].
			binaryBlock value: 'Users of ', autoSelect value: autoSelect]! !

!SystemNavigation methodsFor: 'query' stamp: 'nice 1/5/2010 15:59'!
hierarchyOfClassesSurrounding: aClass
	"Answer a list of classes in the hierarchy both above and below the given class"
	"SystemNavigation default hierarchyOfClassesSurrounding: StringHolder"
	
	| list aClassNonMeta isMeta |
	aClass ifNil: [^ OrderedCollection new].
	aClass ifNil: [^ self].
	aClassNonMeta := aClass theNonMetaClass.
	isMeta := aClassNonMeta ~~ aClass.
	list := OrderedCollection new.
	aClass allSuperclasses reverseDo:
		[:cl | list addLast: cl].
	aClassNonMeta allSubclassesWithLevelDo:
		[:cl :level | | theClassOrMeta |
		theClassOrMeta := isMeta ifTrue: [cl class] ifFalse: [cl].
		list addLast: theClassOrMeta]
	 	startingLevel: 0.
	^ list

! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 1/16/2004 21:03'!
hierarchyOfImplementorsOf: aSelector forClass: aClass
	"Answer a list of classes in the hierarchy both above and below the given class which implement the given selector."
	"SystemNavigation default hierarchyOfImplementorsOf: #contents forClass: StringHolder"

	^ (self hierarchyOfClassesSurrounding: aClass) select:
		[:cl | cl includesSelector: aSelector]
! !

!SystemNavigation methodsFor: 'message sends' stamp: 'Noury 10/26/2008 15:30'!
isMessage: selector sentInClass: aClass 
	^(self allSendersOf: selector inClass: aClass) notEmpty ! !

!SystemNavigation methodsFor: 'message sends' stamp: 'Noury 10/26/2008 15:57'!
isMessage: selector sentInClassCategory: category 
	^(self allSendersOf: selector inClassCategory: category) notEmpty ! !

!SystemNavigation methodsFor: 'message sends' stamp: 'Noury 10/26/2008 16:00'!
isMessage: selector sentInPackageNamed: packageName 
	^(self allSendersOf: selector inPackageNamed: packageName) notEmpty ! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/18/2003 10:44'!
isThereAnImplementorOf: aSelector 
	"Answer true if there is at least one implementor of the selector found  
	in the system, false if there are no implementors"
	"self new isThereAnImplementorOf: #contents.  
	self new isThereAnImplementorOf: #nobodyImplementsThis."
	self 
		allBehaviorsDo: [:class | (class includesSelector: aSelector)
				ifTrue: [^ true]].
	^ false! !

!SystemNavigation methodsFor: 'message sends' stamp: 'HenrikSperreJohansen 2/5/2010 14:07'!
isUnsentMessage: selector

	| special byte |
	special := Smalltalk 
				hasSpecialSelector: selector
				ifTrueSetByte: [ :b | byte := b ].
	self allBehaviorsDo: 
		[ :behavior | 
			(behavior 
				thoroughHasSelectorReferringTo: selector
				special: special
				byte: byte) ifTrue: [^false] ].
	^true! !

!SystemNavigation methodsFor: 'utilities' stamp: 'MarcusDenker 2/20/2010 03:17'!
methodsWithInitials: targetInitials
	"Based on a do-it contributed to the Squeak mailing list by Goran Hultgen:
 Browse methods whose initials (in the time-stamp, as logged to disk) match the given initials.
 Print out the complete time-stamp table to the Transcript.
 Answer a list of (initials -> count) associations.

CAUTION: It may take several minutes for this to complete."

	"Time millisecondsToRun: [SystemNavigation default methodsWithInitials: 'bf'] "

	| initials timeStamp  allSubmitters |
	initials := ''.
	timeStamp := ''.
	allSubmitters := Bag new.
	self browseAllSelect:
			[:cm |
				timeStamp := cm timeStamp.
				initials := timeStamp isEmpty
					ifTrue:
						['']
					ifFalse:
						[timeStamp substrings first].
				initials := initials isEmpty
					ifTrue:
						['<no time stamp>']
					ifFalse:
						[initials first isDigit
							ifTrue:
								['<date>']
							ifFalse:
								[initials]].
				allSubmitters add: initials.
				(initials = targetInitials)]
		name: ('Methods with initials ', targetInitials)
		autoSelect: nil.

	allSubmitters sortedCounts do: [:elem | Transcript cr; show: elem asString].
	^ allSubmitters
! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 4/20/2003 14:27'!
numberOfImplementorsOf: aSelector 
	"Answer a count of the implementors of the given selector found in the  
	system"
	"self new numberOfImplementorsOf: #contents.  
	self new numberOfImplementorsOf: #nobodyImplementsThis.  
	self new numberOfimplementorsOf: #numberOfImplementorsOf:."
	| aCount |
	aCount := 0.
	self
		allBehaviorsDo: [:class | (class includesSelector: aSelector)
				ifTrue: [aCount := aCount + 1]].
	^ aCount! !

!SystemNavigation methodsFor: 'query' stamp: 'sd 9/23/2004 22:03'!
obsoleteBehaviors
	"SystemNavigation default obsoleteBehaviors inspect"
	"Find all obsolete behaviors including meta classes"

	| obs |
	obs := OrderedCollection new.
	Smalltalk garbageCollect.
	self 
		allObjectsDo: [:cl | (cl isBehavior
					and: [cl isObsolete])
				ifTrue: [obs add: cl]].
	^ obs asArray! !

!SystemNavigation methodsFor: 'query' stamp: 'nice 1/5/2010 15:59'!
obsoleteClasses   

	"SystemNavigation default obsoleteClasses inspect"
	"NOTE:  Also try inspecting comments below"
	| obs |
	obs := OrderedCollection new.  Smalltalk garbageCollect.
	Metaclass allInstancesDo:
		[:m | | c | c := m soleInstance.
		(c ~~ nil and: ['AnOb*' match: c name asString])
			ifTrue: [obs add: c]].
	^ obs asArray

"Likely in a ClassDict or Pool...
(Association allInstances select: [:a | (a value isKindOf: Class) and: ['AnOb*' match: a value name]]) asArray
"
"Obsolete class refs or super pointer in last lit of a method...
| n l found |
Smalltalk browseAllSelect:
	[:m | found := false.
	1 to: m numLiterals do:
		[:i | (((l := m literalAt: i) isMemberOf: Association)
				and: [(l value isKindOf: Behavior)
				and: ['AnOb*' match: l value name]])
			ifTrue: [found := true]].
	found]
"! !

!SystemNavigation methodsFor: 'query' stamp: 'AndyKellens 6/11/2010 14:17'!
obsoleteMethodReferences
	"SystemNavigation default obsoleteMethodReferences"

	"Open a browser on all referenced behaviors that are obsolete"

	| obsClasses references |
	references := Array new writeStream.
	obsClasses := self obsoleteBehaviors.
	'Scanning for methods referencing obsolete classes' 
		displayProgressAt: Sensor cursorPoint
		from: 1
		to: obsClasses size
		during: 
			[:bar | 
			obsClasses keysAndValuesDo: 
					[:index :each | | obsRefs | 
					bar value: index.
					obsRefs := each pointersToExcept: obsClasses.
					obsRefs do: 
							[:ref | 
							"Figure out if it may be a global"

							(ref isVariableBinding and: [ref key isString	"or Symbol"]) 
								ifTrue: 
									[(ref pointersTo) do: 
											[:meth | 
											(meth isKindOf: CompiledMethod) 
												ifTrue: [meth methodReference ifNotNil: [:mref | references nextPut: mref]]]]]]].
	^references contents! !

!SystemNavigation methodsFor: 'private' stamp: 'eem 8/30/2010 10:34'!
packageInfoFor: packageInfoOrString
	^packageInfoOrString isString
		ifTrue: [PackageInfo named: packageInfoOrString]
		ifFalse: [packageInfoOrString]! !

!SystemNavigation methodsFor: 'packages' stamp: 'StephaneDucasse 10/17/2010 20:53'!
removeAllButPackages: packagesList
	"Remove all classes and methods that do not belong to some of listed packages. Skip #unload. Do not use this method if you can use Monticello!! This method is used by some low-level shrinking process. Do not use it. This is quite hackish"

	| keep remove oldRemove extensionCategories log |
	keep := Set new.      
	packagesList do: [:pkg | keep addAll: (self allClassesInPackageNamed: pkg)].
	oldRemove := Set new.
 	[ 
		remove := Smalltalk globals allClassesAndTraits reject: [:c | keep includes: c].
		remove := remove collect: [:c | c name].
		(remove asSortedCollection: [:a :b | a  <= b ]) do: [:className |
			| class |
			class := (Smalltalk globals classOrTraitNamed: className).
			class allSubclasses ifEmpty: [
				SystemChangeNotifier uniqueInstance noMoreNotificationsFor: class.
				class isTrait ifFalse: [ 
					(class class includesSelector: #unload) ifTrue: [ class removeSelector: #unload ] ].
				class removeFromSystem. 
				] ].

		
		oldRemove ifNotEmpty: [
			oldRemove = remove ifTrue: [					
				 self error: 'Some classes still have subclasses' ] ].
		oldRemove := remove.
		remove notEmpty] whileTrue.

	keep do: [: class | 
		extensionCategories := (class organization categories select: [:category | 
			category first = $*]) asOrderedCollection.
		extensionCategories removeAllSuchThat: [:category |
			packagesList anySatisfy: [:package |
				((category asLowercase allButFirst = package asLowercase) or: [package asLowercase, '-*' match: category asLowercase allButFirst])]].
		extensionCategories do: [:category | 
			class removeCategory: category ]. 
		class isTrait ifFalse: [
			extensionCategories := (class class organization categories select: [:category | 
				category first = $*]) asOrderedCollection.
			extensionCategories removeAllSuchThat: [:category |
				packagesList anySatisfy: [:package |
					((category asLowercase allButFirst = package asLowercase) or: [package asLowercase, '-*' match: category asLowercase allButFirst])]].
			extensionCategories do: [:category | 
				class class removeCategory: category ]]].
			
! !

!SystemNavigation methodsFor: 'query' stamp: 'nice 1/5/2010 15:59'!
reportSenderCountsFor: selectorList 
	"Produce a report on the number of senders of each of the selectors in 
	the list. 1/27/96 sw"
	"SystemNavigation default reportSenderCountsFor: #(foo: printString)"
	| total report |
	total := 0.
	report := '
'.
	selectorList
		do: [:selector | | thisSize | 
			thisSize := (self allCallsOn: selector) size.
			report := report , thisSize printString , String tab , selector printString , String cr.
			total := total + thisSize].
	report := report , '--- ------------------
'.
	report := report , total printString , String tab , 'TOTAL
'.
	^ report! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:12'!
selectAllMethods: aBlock 
	"Answer a SortedCollection of each method that, when used as the block  
	argument to aBlock, gives a true result."
	| aCollection |
	aCollection := SortedCollection new.
	self
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:sel :m | (aBlock value: m)
						ifTrue: [aCollection
								add: (MethodReference new setStandardClass: class methodSymbol: sel)]]].
	^ aCollection! !

!SystemNavigation methodsFor: 'author' stamp: 'stephane.ducasse 8/8/2009 15:08'!
signatories
	^(self signatoriesString subStrings: {Character cr})
		collect: [ :each | each substrings first]! !

!SystemNavigation methodsFor: 'author' stamp: 'stephane.ducasse 8/8/2009 15:08'!
signatoriesString
	^'AB               Alexandre Bergel
ack              Alan Kay
ads              Adam Spitz
ajh              Anthony Hannan
aka              Mark Guzdial
al               Adrian Lienhard
aoy              Andres Otaduy
apb              Andrew P. Black
apl              Alain Plantec
ar               Andreas Raab
asm              Alejandro Magistrello
avi              Avi Bryant
bf               Bert Freudenberg
BG               Boris Gaertner
bh               Bob Hartwig
BJP              Bijan Parsia
bkv              Brent Vukmer
bolot            Bolot Kerimbaev
bootstrap        Pavel Krivanek
BP               Bijan Parsia
brp              Brent Pinkney
btr              Brian Rice
cbc              Chris Cunningham
ccn              Chris Norton
CdG              Cees de Groot
cds              C. David Shaffer
cmm              Chris Muller
crl              Craig Latta
cwp              Colin Putney
daf              Dave Faught
dao              danil osipchuk
DAS              David A Smith
dc               Damien Cassou
dd               Dominique Dutoit
dew              Doug Way
DF               Diego Fernandez
dgd              Diego Gomez Deck
dhhi             Dan Ingalls
di               Dan Ingalls
DM               Duane Maxwell
DSM              Duane Maxwell
dtl              Dave Lewis
dvf              Daniel Vainsencher
edc              Edgar DeCleene
efc              Eddie Cottongim
efo              Emilio Oca
em               Ernest Micklei?
emm              Ernest Micklei
es               Enrico Spinielli
FBS              Frank Shearar
fc               Frank Caggiano
fcs              Frank Sergeant
fm               Florin Mateoc
gh               Goran Krampe (nee Hultgren)
gk               Goran Krampe (nee Hultgren)
gm               German Morales
go               Georg Gollmann
gsa              German Arduino
HEG              Henrik Ekenberg
HilaireFernandes Hilaire Fernandes
HK               Herbert Konig
hmm              Hans-Martin Mosner
hpt              Hernan Tylim
huma             Lyndon Tremblay
ich.             Yuji Ichikawa
ikp              Ian Piumarta
jaf              Jan Fietz
jam              Javier Musa
jcg              Joshua Gargus
jdf              David Farber
jdr              Javier Diaz-Reinoso
je               Jorn Eyrich
je77             Jochen Rick
JF               Julian Fitzell
jhm              John Maloney
jlb              Jim Benson
jmb              Hans Baveco
JMM              John McIntosh
JMV              Juan Manuel Vuletich
jon              Jon Hylands
jrm              John-Reed Maffeo
jrp              John Pierce
jsp              Jeff Pierce
ka               Kazuhiro Abe
kfr              Karl Ramberg
KLC              Ken Causey
KR               korakurider
KTT              Kurt Thams
kwl              Klaus D. Witzel
ky               Koji Yokokawa
laza             Alexander Lazarevic
LB               Leo Burd
LC               Leandro Caniglia
LEG              Gerald Leeb
len              Luciano Esteban Notarfrancesco
lr               Lukas Renggli
lrs              Lorenzo Schiavina
ls               Lex Spoon
m3r              Maurice Rabb
mas              Mark Schwenk
MD               Markus Denker
mdr              Mike Rutenberg
mga              Markus Galli
miki             Mikael Kindborg
mikki            Mikael Kindborg
mir              Michael Rueger
mist             Michal Starke
mjg              Mark Guzdial
mjr              Mike Roberts
mjt              Mike Thomas
mk               Matej Kosik
MPW              Marcel Weiher
mrm              Martin McClure
ms               math
MU               Masashi Umezawa
mw               Martin Wirblat
nb               Naala Brewer
nice             Nicolas Cellier
nk               Ned Konz
Noury            Noury Bouraqadi
NS               Nathanael Schaerli
PHK              Peter Keeler
pk               Pavel Krivanek
pm               Patrick Mauritz
pnm              Paul McDonough
RAA              Bob Arning
raok             Richard A. O"Keefe
rbb              Brian Brown
rca              Russell Allen
reThink          Paul McDonough
rew              Roger Whitney
rhi              Robert Hirschfeld
Rik              Rik Fischer SmOOdy
RJT              Ron Teitelbaum
rr               Romain Robbes
rw               Roel Wuyts
rww              Robert Withers
sbw              Stephan B. Wessels
SD               Stephane Ducasse
sge              Steve Elkins
shrink           Pavel Krivanek
slg              Steve Gilbert
sm               Simon Michael
sma              Stefan Matthias Aust
sn               Suslov Nikolay
spfa             Stephane Rollandin
sps              Steven Swerling
SqR              Andres Valloud
SqR!!!!            Andres Valloud
SqR!!!!!!!!          Andres Valloud
sr               Stephan Rudlof
ssa              Sam S. Adams
st               Samuel Tardieu
stephaneducasse  Stephane Ducasse
stp              Stephen Travis Pope
sumim            Masato Sumi
svp              Stephen Vincent Pair
sw               Scott Wallace
T2               Toshiyuki Takeda
tak              Takashi Yamamiya
tao              Tim Olson
tb               Todd Blanchard
TBn              Torsten Bergmann
tetha            Tetsuya Hayashi
tfei             The Fourth Estate, Inc.
th               Torge Husfeldt
ti               Tobias Isenberg
TJ               TJ Leone
tk               Ted Kaehler
tk               Thomas Kowark
tlk              Tom Koenig
TN               korakurider
TPR              Tim Rowledge
TRee             Trygve Reenskaug
Tsutomu          Tsutomu Hiroshima
tween            Andy Tween
vb               Vassili Bykov
vj               Vladimir Janousek
wiz              Jerome Peace
wod              Bill Dargel
ykoubo           Koji Yokokawa
yo               Yoshiki Ohshima
zz               Serge Stinckwich
jmv
jm
jmm
md
sd
tpr'! !

!SystemNavigation methodsFor: 'query' stamp: 'LukasRenggli 5/12/2010 10:12'!
unimplemented
	"Answer an Array of each message that is sent by an expression in a method but is not implemented by any object in the system."
	
	| all unimplemented |
	all := IdentitySet new: Symbol instanceCount * 2.
	self allBehaviorsDo: [:cl | cl selectorsDo: [:aSelector | all add: aSelector]].

	unimplemented := IdentityDictionary new.
	self allBehaviorsDo: [:cl |
		 cl selectorsAndMethodsDo: [:sel :meth |
			meth messages do: [:m | | entry |
				(all includes: m) ifFalse: [
					entry := unimplemented at: m ifAbsent: [Array new].
					entry := entry copyWith: (cl name, '>', sel).
					unimplemented at: m put: entry]]]].
	^ unimplemented
! !

!SystemNavigation methodsFor: 'message sends' stamp: 'Noury 10/26/2008 17:22'!
unsentMessagesInCategory: category
	^(SystemOrganization classesInCategory: category) inject: Set new into: [:unsentMessages :class|
		unsentMessages, (self unsentMessagesInClass: class)]
	! !

!SystemNavigation methodsFor: 'message sends' stamp: 'StephaneDucasse 12/5/2009 21:31'!
unsentMessagesInClass: aClass 
	|methReferences|
	methReferences := Set new.
	aClass selectorsDo: [:selector|
		(self isUnsentMessage: selector) ifTrue: [
			methReferences add: (MethodReference class: aClass selector: selector)]].
	^methReferences! !

!SystemNavigation methodsFor: 'message sends' stamp: 'Noury 10/26/2008 17:26'!
unsentMessagesInPackageNamed: packageName
	| unsentMessages |
	unsentMessages := self unsentMessagesInCategory: packageName.	
	(SystemOrganization categoriesMatching: packageName, '-*') do: [:category|
		unsentMessages addAll: (self unsentMessagesInCategory: category)].
	^unsentMessages 
	! !

Object subclass: #SystemVersion
	instanceVariableNames: 'version date highestUpdate updates type major minor suffix'
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'System-Support'!

!SystemVersion commentStamp: 'tlk 11/14/2004 10:44' prior: 0!
I am responsible for maintaining what version of Squeak and the VM is running.  I also track all of the update items that have been included in the image.

I'm invoked at auto start to get the latest plugins, etc.!

!SystemVersion class methodsFor: 'updating' stamp: 'simon.denier 6/11/2010 17:34'!
check: pluginVersion andRequestPluginUpdate: updateURL
	"SystemVersion check: 'zzz' andRequestPluginUpdate: 'http://www.squeakland.org/installers/update.html' "

	"We don't have a decent versioning scheme yet, so we are basically checking for a nil VM version on the mac."
	(self pluginVersion: pluginVersion newerThan: self currentPluginVersion)
		ifFalse: [^true].
	(self confirm: 'There is a newer plugin version available. Do you want to install it now?')
		ifFalse: [^false].
	HTTPClient
		requestURL: updateURL , (OSPlatform platformName copyWithout: Character space) asLowercase , '.html'
		target: '_top'.
	^false! !

!SystemVersion class methodsFor: 'updating' stamp: 'StephaneDucasse 3/17/2010 21:26'!
checkAndApplyUpdates: availableUpdate
	"SystemVersion checkAndApplyUpdates: nil"

	^(availableUpdate isNil
		or: [availableUpdate > SystemVersion current highestUpdate])
		ifTrue: [
			(self confirm: 'There are updates available. Do you want to install them now?')
				ifFalse: [^false].
			Utilities
				readServerUpdatesThrough: availableUpdate
				saveLocally: false
				updateImage: true.
			Smalltalk snapshot: true andQuit: false.
			true]
		ifFalse: [false]! !

!SystemVersion class methodsFor: 'accessing' stamp: ''!
current
	Current ifNil: [Current := SystemVersion new].
	^Current! !

!SystemVersion class methodsFor: 'updating' stamp: 'StephaneDucasse 3/17/2010 21:26'!
currentPluginVersion
	^Smalltalk vmVersion! !

!SystemVersion class methodsFor: 'instance creation' stamp: 'MikeRoberts 9/28/2010 22:41'!
fromShortVersionString: aString
	"Create a new SystemVersion from the given short version string. Useful for testing."

	| parts |
	parts := self parseShortVersionString: aString.
	^self new
		type: parts first;
		major: parts second;
		minor: parts third;
		suffix: parts fourth;
		yourself! !

!SystemVersion class methodsFor: 'instance creation' stamp: 'MikeRoberts 9/28/2010 22:46'!
newVersion: versionName
	"Install a new SystemVersion as the current version. Only preserve the higest update recorded.
	This method is a candidate for being removed, I am not sure it ever makes sense to do this."
	
	| newVersion |
	newVersion := self fromShortVersionString: versionName.
	newVersion
		highestUpdate: self current highestUpdate.
	Current := newVersion
! !

!SystemVersion class methodsFor: 'accessing' stamp: 'MikeRoberts 11/19/2009 21:50'!
parseShortVersionString: aString 
	"Answer the parts of aString, expected in short format <type><major>.<minor><suffix>."

	| stream length type major minor suffix |

	stream := aString readStream.
	[stream peek isDigit] whileFalse: [stream next].
	length := stream position.
	stream reset.
	type := (stream next: length).
	major := (stream upTo: $.) asInteger.
	stream := stream upToEnd readStream.
	[stream peek notNil and: [stream peek isDigit]] whileTrue: [stream next].
	length := stream position.
	stream reset.
	minor := (stream next: length) asInteger.
	suffix := stream upToEnd.
	^{type. major. minor. suffix}! !

!SystemVersion class methodsFor: 'accessing' stamp: 'nice 1/5/2010 15:59'!
parseVersionString: versionString 
	"Answer the version of this release as version, date, update."
	"SystemVersion parseVersionString: 'Squeak3.1alpha of 28 February 2001 [latest update: #3966]' "
	| version date update |
	
	[ | stream | stream := versionString readStream.
	version := stream upToAll: ' of '.
	date := Date readFrom: stream.
	stream upToAll: ' #'.
	update := Number readFrom: stream ] 
		on: Error
		do: [ ^ nil ].
	^ {  version. date. update  }! !

!SystemVersion class methodsFor: 'accessing' stamp: 'mir 8/10/2001 11:53'!
pluginVersion: availableVersionString newerThan: currentVersionString
	| currentVersion availableVersion |
	(currentVersionString isEmptyOrNil
		or: [availableVersionString isEmptyOrNil])
		ifTrue: [^true].
	currentVersion := self parseVersionString: currentVersionString.
	availableVersion := self parseVersionString: availableVersionString.
	(currentVersion isNil
		or: [availableVersion isNil])
		ifTrue: [^false].
	^(currentVersion at: 2) < (availableVersion at: 2)! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
date
	^date! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
date: newDate
	date := newDate! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 5/1/2001 18:19'!
datedVersion
	"Answer the version of this release."

	^ self version asString , ' of ' , self date printString! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 3/29/2001 18:03'!
highestUpdate
	| sortedUpdates |
	highestUpdate ifNil: [
		sortedUpdates := self updates asSortedCollection.
		highestUpdate := (sortedUpdates isEmpty
			ifTrue: [0]
			ifFalse: [sortedUpdates last])].
	^highestUpdate! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
highestUpdate: anInteger
	highestUpdate := anInteger! !

!SystemVersion methodsFor: 'printing' stamp: 'MikeRoberts 11/15/2009 13:24'!
imageVersionString
	"Print the version compatibile with naming the image."
	
	^String streamContents: [:str |
		str nextPutAll: self type;
			nextPut: $-;
			nextPutAll: self major printString;
			nextPut: $.;
			nextPutAll: self minor printString;
			nextPut: $-;
			nextPutAll: self highestUpdate printString;
			nextPut: $-;
			nextPutAll: self suffix]! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
includesUpdate: anUpdate
	^self updates includes: anUpdate! !

!SystemVersion methodsFor: 'initialize' stamp: 'MikeRoberts 9/27/2010 23:56'!
initialize
	super initialize.
	date := Date today.
	updates := Set new.
! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
major
	"Answer the value of major"

	^ major! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
major: anObject
	"Set the value of major"

	major := anObject! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 9/27/2010 23:58'!
majorMinorVersion
	"Return the major/minor version number of the form X.Y, without any 'alpha' or 'beta' or other suffix. 
	If the receiver needs upgrading, do it. Remove the guard in 1.3."
	"SystemVersion current majorMinorVersion"
	
	self upgrade.
	^String streamContents: [:str |
		str 
			nextPutAll: self type;
			nextPutAll: self major printString; nextPut: $.;
			nextPutAll: self minor printString]! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
minor
	"Answer the value of minor"

	^ minor! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
minor: anObject
	"Set the value of minor"

	minor := anObject! !

!SystemVersion methodsFor: 'printing' stamp: 'mir 5/1/2001 18:20'!
printOn: stream
	stream
		nextPutAll: self datedVersion;
		nextPutAll: ' update ' , self highestUpdate printString! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 3/29/2001 18:01'!
registerUpdate: update
	self updates add: update.
	self resetHighestUpdate! !

!SystemVersion methodsFor: 'accessing' stamp: 'mir 3/29/2001 18:01'!
resetHighestUpdate
	highestUpdate := nil! !

!SystemVersion methodsFor: 'printing' stamp: 'MikeRoberts 9/28/2010 22:33'!
shortVersionString
	"Print a short version string.  Compatibile with senders that concatenate this with 
	#lastUpdateString or #highestUpdate. Essentially the old version format. Guard 
	against no state being set."
	
	self type ifNil: [^'No version set'].
	^String streamContents: [:str |
		str 
			nextPutAll: self type;
			nextPutAll: self major printString;
			nextPut: $.;
			nextPutAll: self minor printString;
			nextPutAll: self suffix]! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
suffix
	"Answer the value of suffix"

	^ suffix! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
suffix: anObject
	"Set the value of suffix"

	suffix := anObject! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
type
	"Answer the value of type"

	^ type! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 11/15/2009 13:09'!
type: anObject
	"Set the value of type"

	type := anObject! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
unregisterUpdate: update
	self updates remove: update ifAbsent: []! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
updates
	^updates! !

!SystemVersion methodsFor: 'upgrading' stamp: 'MikeRoberts 9/27/2010 23:40'!
upgrade

	| parts |
	version ifNil: [^self].
	parts := self class parseShortVersionString: version.
	self type: parts first.
	self major: parts second.
	self minor: parts third.
	self suffix: parts fourth.
	version := nil.! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 9/27/2010 23:53'!
version
	"Answer the short version string. If the receiver needs upgrading, do it. Remove the guard in 1.3."
	
	self upgrade.
	^self shortVersionString! !

!SystemVersion methodsFor: 'accessing' stamp: 'MikeRoberts 9/28/2010 22:58'!
version: newVersion

	self deprecated: 'Use individual accessors type, major, minor, suffix. Version is not a string.' on: '28 Sep 2010' in: 'Pharo1.2'.
	
	version := newVersion.
	type := major := minor := suffix := nil! !

Object subclass: #Utilities
	instanceVariableNames: ''
	classVariableNames: 'PromptForUpdateServer UpdateDownloader UpdateRemoveSequenceNum UpdateSavesFile UpdateUrlLists'
	poolDictionaries: ''
	category: 'System-Support'!

!Utilities commentStamp: 'stephaneducasse 8/16/2010 23:11' prior: 0!
An empty class to help people loading their code when it contains Preferences methods.!

!Utilities class methodsFor: 'fetching updates' stamp: 'AlainPlantec 1/7/2010 22:53'!
chooseUpdateList
	"When there is more than one set of update servers, let the user choose which we will update from.  Put it at the front of the list. Return false if the user aborted.  If the preference Utilities class>>#promptForUpdateServer returns false, then suppress that prompt, in effect using the same server choice that was used the previous time (a convenience for those of us who always answer the same thing to the prompt.)"

	| index him |
	((UpdateUrlLists size > 1) and: [self promptForUpdateServer])
		ifTrue:
			[index := UIManager default 
				chooseFrom: (UpdateUrlLists collect: [:each | each first]) 
				lines: #()
				title: 'Choose a group of servers\from which to fetch updates.' translated withCRs.
			index > 0 ifTrue:
				[him := UpdateUrlLists at: index.
				UpdateUrlLists removeAt: index.
				UpdateUrlLists addFirst: him].
			^ index > 0].
	^ true! !

!Utilities class methodsFor: 'fetching updates' stamp: 'di 4/29/2001 22:26'!
extractThisVersion: list
	"Pull out the part of the list that applies to this version."

	| listContents version versIndex |
	listContents := self parseListContents: list.
	version := SystemVersion current version.
	versIndex := (listContents collect: [:pair | pair first]) indexOf: version.
	versIndex = 0 ifTrue: [^ Array new].		"abort"
	^ (listContents at: versIndex) last! !

!Utilities class methodsFor: 'fetching updates' stamp: 'nice 1/5/2010 15:59'!
newUpdatesOn: serverList special: indexPrefix throughNumber: aNumber
	"Return a list of fully formed URLs of update files we do not yet have.  Go to the listed servers and look at the file 'updates.list' for the names of the last N update files.  We look backwards for the first one we have, and make the list from there.  tk 9/10/97
	No updates numbered higher than aNumber (if it is not nil) are returned " 

	| existing out maxNumber |
	maxNumber := aNumber ifNil: [99999].
	out := OrderedCollection new.
	existing := SystemVersion current updates.
	serverList do: [:server | | list doc raw char |
		doc := HTTPClient httpGet: 'http://' , server,indexPrefix,'updates.list'.
		"test here for server being up"
		doc class == RWBinaryOrTextStream ifTrue:
			[raw := doc reset; contents.	"one file name per line"
			list := self extractThisVersion: raw.
			list reverseDo: [:fileName | | ff itsNumber |
				ff := (fileName findTokens: '/') last.	"allow subdirectories"
				itsNumber := ff initialIntegerOrNil. 
				(existing includes: itsNumber)
					ifFalse:
						[
						(itsNumber isNil or: [itsNumber <= maxNumber])
							ifTrue:
								[out addFirst: 'http://' , server, fileName]]
					ifTrue: [^ out]].
			((out size > 0) or: [char := doc reset; skipSeparators; next.
				(char == $*) | (char == $#)]) ifTrue:
					[^ out "we have our list"]].	"else got error msg instead of file"
		"Server was down, try next one"].
	self inform: 'All code update servers seem to be unavailable'.
	^ out! !

!Utilities class methodsFor: 'fetching updates' stamp: 'HenrikSperreJohansen 6/12/2010 02:40'!
parseListContents: listContents 
	"Parse the contents of updates.list into {{vers. {fileNames*}}*}, and return it."
	| sections vers strm line fileNames |
	sections := OrderedCollection new.
	fileNames := OrderedCollection new: 1000.
	vers := nil.
	strm := listContents readStream.
	[ strm atEnd ] whileFalse: 
		[ line := strm nextLine.
		line size > 0 ifTrue: 
			[ line first = $# 
				ifTrue: 
					[ vers ifNotNil: [ sections addLast: {  vers. (fileNames asArray)  } ].
					"Start a new section"
					vers := line allButFirst.
					fileNames resetTo: 1 ]
				ifFalse: [ line first = $* ifFalse: [ fileNames addLast: line ] ] ] ].
	vers ifNotNil: [ sections addLast: {  vers. (fileNames asArray)  } ].
	^ sections asArray
	" TEST:
 | list |
list := Utilities parseListContents: (FileStream oldFileNamed: 'updates.list') contentsOfEntireFile.
list = (Utilities parseListContents: (String streamContents: [:s | Utilities writeList: list toStream: s]))
	ifFalse: [self error: 'test failed']
	ifTrue: [self inform: 'test OK']
"! !

!Utilities class methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 22:43'!
promptForUpdateServer
	^ PromptForUpdateServer ifNil: [PromptForUpdateServer := true]! !

!Utilities class methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 22:44'!
promptForUpdateServer: aBoolean
	PromptForUpdateServer := aBoolean! !

!Utilities class methodsFor: 'fetching updates' stamp: 'AdrianLienhard 6/7/2010 10:55'!
readServer: serverList special: indexPrefix updatesThrough: maxNumber saveLocally: saveLocally updateImage: updateImage
	"Scan the update server(s) for unassimilated updates. If maxNumber is not nil, it represents the highest-numbered update to load.  This makes it possible to update only up to a particular point.   If saveLocally is true, then save local copies of the update files on disc.  If updateImage is true, then absorb the updates into the current image."

"Utilities readServer: Utilities serverUrls updatesThrough: 828 saveLocally: true updateImage: true"

	| str urls failed loaded this nextDoc updateName |
	(Smalltalk globals includesKey: #ScriptLoader) ifFalse: [
		^ self inform: 'ScriptLoader is not present (this may happen if you have run cleanUpForProduction). ' ].
	Cursor wait showWhile: 
		[ | docQueue docQueueSema |
		urls := self newUpdatesOn: (serverList collect: [:url | url, 'updates/','pharo', (Smalltalk globals at: #ScriptLoader)   currentMajorVersionNumber asString, '/']) 
				special: indexPrefix
				throughNumber: maxNumber.
		loaded := 0.
		failed := nil.
		"send downloaded documents throuh this queue"
		docQueue := SharedQueue new.
		"this semaphore keeps too many documents from beeing queueed up at a time"
		docQueueSema := Semaphore new.
		5 timesRepeat: [ docQueueSema signal ].

		"fork a process to download the updates"
		self retrieveUrls: urls ontoQueue: docQueue withWaitSema: docQueueSema.

		"process downloaded updates in the foreground"

		[ this := docQueue next.
		nextDoc := docQueue next.  
		nextDoc = #failed ifTrue: [ failed := this ].
		(failed isNil and: [ nextDoc ~= #finished ])
		] whileTrue: [
			failed ifNil: [
				nextDoc reset; text.
				nextDoc size = 0 ifTrue: [ failed := this ]. ].
			failed ifNil: [
				nextDoc peek asciiValue = 4	"pure object file"
					ifTrue: [failed := this]].	"Must be fileIn, not pure object file"
			failed ifNil: [
				updateImage
					ifTrue: [
						updateName := (this findTokens: '/') last.
						ChangeSet newChangesFromStream: nextDoc named: updateName.
						SystemVersion current registerUpdate: updateName initialIntegerOrNil].
			saveLocally ifTrue:
				[self saveUpdate: nextDoc onFile: (this findTokens: '/') last].
			].
		loaded := loaded + 1.
		docQueueSema signal].
	].

	failed ~~ nil & (urls size - loaded > 0) ifTrue: [
		str := loaded printString ,' new update file(s) processed.'.
		str := str, '\Could not load ' withCRs, 
			(urls size - loaded) printString ,' update file(s).',
			'\Starting with "' withCRs, failed, '".'.
		self inform: str].
	^ Array with: failed with: loaded
! !

!Utilities class methodsFor: 'fetching updates' stamp: 'sw 1/10/1999 02:02'!
readServerUpdatesSaveLocally: saveLocally updateImage: updateImage
	^ self readServerUpdatesThrough: nil saveLocally: saveLocally updateImage: updateImage! !

!Utilities class methodsFor: 'fetching updates' stamp: 'rbb 2/18/2005 13:16'!
readServerUpdatesThrough: maxNumber saveLocally: saveLocally updateImage: updateImage
	"Scan the update server(s) for unassimilated updates. If maxNumber is not nil, it represents the highest-numbered update to load.  This makes it possible to update only up to a particular point.   If saveLocally is true, then save local copies of the update files on disc.  If updateImage is true, then absorb the updates into the current image.

A file on the server called updates.list has the names of the last N update files.  We look backwards for the first one we do not have, and start there"
"* To add a new update:  Name it starting with a new two-digit code.  
* Do not use %, /, *, space, or more than one period in the name of an update file.
* The update name does not need to have any relation to the version name.
* Figure out which versions of the system the update makes sense for.
* Add the name of the file to each version's category below.
* Put this file and the update file on all of the servers.
*
* To make a new version of the system:  Pick a name for it (no restrictions)
* Put # and exactly that name on a new line at the end of this file.
* During the release process, fill in exactly that name in the dialog box.
* Put this file on the server."
"When two sets of updates need to use the same directory, one of them has a * in its 
serverUrls description.  When that is true, the first word of the description is put on
the front of 'updates.list', and that is the index file used."

"Utilities readServerUpdatesThrough: 3922 saveLocally: true updateImage: true"

	| failed loaded str res servers triple tryAgain indexPrefix |
	Utilities chooseUpdateList ifFalse: [^ self].	"ask the user which kind of updates"

	servers := Utilities serverUrls copy.
	indexPrefix := (Utilities updateUrlLists first first includes: $*) 
		ifTrue: [(Utilities updateUrlLists first first findTokens: ' ') first]
						"special for internal updates"
		ifFalse: ['']. 	"normal"
	[servers isEmpty] whileFalse: [
		triple := self readServer: servers special: indexPrefix 
					updatesThrough: maxNumber 
					saveLocally: saveLocally updateImage: updateImage.

		"report to user"
		failed := triple first.
		loaded := triple second.
		tryAgain := false.
		failed ifNil: ["is OK"
			loaded = 0 ifTrue: ["found no updates"
				servers size > 1 ifTrue: ["not the last server"
					res := UIManager default 
							chooseFrom: #('Stop looking' 'Try next server')
							title: 
'No new updates on the server
', servers first, '
Would you like to try the next server?
(Normally, all servers are identical, but sometimes a
server won''t let us store new files, and gets out of date.)' 
						.
					res = 2 ifFalse: [^ self]
						 ifTrue: [servers := servers allButFirst.	"try the next server"
							tryAgain := true]]]].
		tryAgain ifFalse: [
			str := loaded printString ,' new update file(s) processed.'.
			^ self inform: str].
	].! !

!Utilities class methodsFor: 'fetching updates' stamp: 'nice 1/5/2010 15:59'!
retrieveUrls: urls ontoQueue: queue withWaitSema: waitSema 
	"download the given list of URLs. The queue will be loaded alternately  
	with url's and with the retrieved contents. If a download fails, the  
	contents will be #failed. If all goes well, a special pair with an empty  
	URL and the contents #finished will be put on the queue. waitSema is  
	waited on every time before a new document is downloaded; this keeps 
	the downloader from getting too far  ahead of the main process"
	"kill the existing downloader if there is one"
	| updateCounter |
	UpdateDownloader
		ifNotNil: [UpdateDownloader terminate].
	updateCounter := 0.
	"fork a new downloading process"
	UpdateDownloader := [
		'Downloading updates' displayProgressAt: Sensor cursorPoint from: 0 to: urls size during: [:bar |
			urls
				do: [:url | | canPeek front doc | 
					waitSema wait.
					queue nextPut: url.
					doc := HTTPClient httpGet: url.
					doc isString
						ifTrue: [queue nextPut: #failed.
							UpdateDownloader := nil.
							Processor activeProcess terminate]
						ifFalse: [canPeek := 120 min: doc size.
							front := doc next: canPeek.  doc skip: -1 * canPeek.
							(front beginsWith: '<!!DOCTYPE') ifTrue: [
								(front includesSubString: 'Not Found') ifTrue: [
									queue nextPut: #failed.
									UpdateDownloader := nil.
									Processor activeProcess terminate]]].
						UpdateDownloader ifNotNil: [queue nextPut: doc. updateCounter := updateCounter + 1. bar value: updateCounter]]].
			queue nextPut: ''.
			queue nextPut: #finished.
			UpdateDownloader := nil] newProcess.
	UpdateDownloader priority: Processor userInterruptPriority.
	"start the process running"
	UpdateDownloader resume! !

!Utilities class methodsFor: 'fetching updates' stamp: 'AlainPlantec 1/7/2010 22:53'!
saveUpdate: doc onFile: fileName
	"Save the update on a local file.  With or without the update number on the front, depending on #updateRemoveSequenceNum"

	| file fName pos updateDirectory |

	(FileDirectory default directoryNames includes: 'updates') ifFalse:
		[FileDirectory default createDirectory: 'updates'].
	updateDirectory := FileDirectory default directoryNamed: 'updates'.

	fName := fileName.
	(self updateRemoveSequenceNum) ifTrue:
		[pos := fName findFirst: [:c | c isDigit not].
		fName := fName copyFrom: pos to: fName size].
	doc reset; ascii.
	(updateDirectory fileExists: fName) ifFalse:
		[file := updateDirectory newFileNamed: fName.
		file nextPutAll: doc contents.
		file close].
! !

!Utilities class methodsFor: 'fetching updates' stamp: 'pmm 3/13/2010 11:34'!
serverUrls 
	"Return the current list of server URLs.  For code updates.  Format of UpdateUrlLists is 
#( ('squeak updates' ('url1' 'url2'))
    ('some other updates' ('url3' 'url4')))"

	| list |
	list := UpdateUrlLists first last.

	"If there is a dead server, return a copy with that server last" 
	Socket deadServer ifNotNil: [
		list shallowCopy withIndexDo: [:aName :ind |
		(aName beginsWith: Socket deadServer) ifTrue: [
			list := list asOrderedCollection.	"and it's a copy"
			list removeAt: ind.
			list addLast: aName]]
	].

	^ list asArray! !

!Utilities class methodsFor: 'fetching updates' stamp: 'mir 8/10/2001 12:30'!
setUpdateServer: groupName
	"Utilities setUpdateServer: 'Squeakland' "
	| entry index |


	entry := UpdateUrlLists detect: [:each | each first = groupName] ifNone: [^self].
	index := UpdateUrlLists indexOf: entry.
	UpdateUrlLists removeAt: index.
	UpdateUrlLists addFirst: entry! !

!Utilities class methodsFor: 'fetching updates' stamp: 'AlainPlantec 1/7/2010 22:53'!
updateFromServer
	"Update the image by loading all pending updates from the server.  Also save local copies of the update files if self updateSavesFile returns true"

	self readServerUpdatesSaveLocally: self updateSavesFile updateImage: true! !

!Utilities class methodsFor: 'fetching updates' stamp: 'AlainPlantec 12/20/2009 00:33'!
updateRemoveSequenceNum
	^ UpdateRemoveSequenceNum ifNil: [UpdateRemoveSequenceNum := false]! !

!Utilities class methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 22:45'!
updateSavesFile
	^ UpdateSavesFile ifNil: [UpdateSavesFile := false]! !

!Utilities class methodsFor: 'settings' stamp: 'AlainPlantec 12/17/2009 22:45'!
updateSavesFile: aBoolean
	UpdateSavesFile := aBoolean! !

!Utilities class methodsFor: 'fetching updates' stamp: 'tk 5/7/1998 17:03'!
updateUrlLists

	UpdateUrlLists ifNil: [UpdateUrlLists := OrderedCollection new].
	^ UpdateUrlLists! !

!Utilities class methodsFor: 'fetching updates' stamp: 'nice 1/5/2010 15:59'!
writeList: listContents toStream: strm
	"Write a parsed updates.list out as text.
	This is the inverse of parseListContents:"

	
	strm reset.
	listContents do:
		[:pair | | fileNames version | version := pair first.  fileNames := pair last.
		strm nextPut: $#; nextPutAll: version; cr.
		fileNames do: [:fileName | strm nextPutAll: fileName; cr]].
	strm close! !

!Utilities class methodsFor: 'fetching updates' stamp: 'RAA 12/17/2000 16:19'!
zapUpdateDownloader

	UpdateDownloader ifNotNil: [UpdateDownloader terminate].
	UpdateDownloader := nil.! !

!Utilities methodsFor: 'look in class' stamp: 'sw 10/13/1998 13:14'!
seeClassSide
	"All the code for Utilitites is on the class side"! !

!ClassDescription methodsFor: '*System-Support' stamp: 'dvf 8/23/2003 12:45'!
allUnreferencedClassVariables
	"Answer a list of the names of all the receiver's unreferenced class vars, including those defined in superclasses"

	^ self systemNavigation allUnreferencedClassVariablesOf: self! !
SmalltalkImage initialize!
ExternalSemaphoreTable initialize!
ExternalSettings initialize!
DefaultExternalDropHandler initialize!
SecurityManager initialize!
AutoStart initialize!
ExternalDropHandler initialize!
